/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "Client.h"
#include "RankingDialog.h"

#include "PlayersPage/PlayersPage.h"
#include "PlayersPage/MarkersWidget.h"

#include "shared/defaultValues.h"
#include "shared/SharedObjects.h"
#include "shared/StringTools.h"

// Qt includes
#include <QDebug>
#include <QTcpSocket>
#include <QTimer>
#include <QByteArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QDataStream>
#include <QJsonArray>
#include <QColor>
#include <QApplication>

using namespace ServerProtocol;

Client::Client(QObject *parent, SharedObjects *sharedObjects, PlayersPage *playersPage)
    : AbstractNetworkEngine(parent),
      m_db(sharedObjects->database()),
      m_stringTools(sharedObjects->stringTools()),
      m_playersPage(playersPage),
      m_markersWidget(playersPage->markersWidget())
{
    connect(socket(), QOverload<QAbstractSocket::SocketError>::of(&QAbstractSocket::error),
            this, &Client::error);
    connect(socket(), &QAbstractSocket::connected, this, &Client::socketConnected);

    m_handshakeTimer = new QTimer(this);
    m_handshakeTimer->setInterval(NETWORK_CONNECT_TIMEOUT);
    m_handshakeTimer->setSingleShot(true);
    connect(m_handshakeTimer, &QTimer::timeout, this, &Client::handshakeTimeout);

    m_requestTimer = new QTimer(this);
    m_requestTimer->setInterval(NETWORK_REQUEST_TIMEOUT);
    m_requestTimer->setSingleShot(true);
    connect(m_requestTimer, &QTimer::timeout, this, &Client::requestTimeout);

    m_connectionPending = false;
    m_connected = false;
}

void Client::connectToServer(const QHostAddress &address, quint16 port)
{
    emit message(tr("Verbinde mit Server %1").arg(address.toString()));
    socket()->connectToHost(address, port);
}

QString Client::serverIp() const
{
    return socket()->peerAddress().toString();
}

int Client::serverPort() const
{
    return socket()->peerPort();
}

void Client::socketConnected()
{
    m_connectionPending = true;
    m_handshakeTimer->start();
}

void Client::abortHandshake()
{
    closeConnection();
    m_handshakeTimer->stop();
    m_connectionPending = false;
    m_stringTools->resetLocale();
}

void Client::error(QAbstractSocket::SocketError)
{
    closeConnection();
    if (m_connectionPending) {
        emit handshakeResult(false, socket()->errorString());
        m_handshakeTimer->stop();
        m_connectionPending = false;
    } else {
        emit connectionError(socket()->errorString());
    }
}

void Client::handshakeTimeout()
{
    closeConnection();
    emit message(tr("Timeout: Keine Antwort nach %1 Sekunden").arg(NETWORK_CONNECT_TIMEOUT / 1000));
    emit handshakeResult(false, tr("Der Server hat zu lange keine Antwort geschickt."));
}

void Client::closeConnection()
{
    socket()->abort();
    m_connected = false;
}

void Client::sendRequest(QJsonObject &jsonObject, ProcessingTarget target,
                         const QString &request, const QString &scrollToTarget)
{
    emit message(tr("Frage %1 an").arg(request));
    emit statusUpdate(tr("Frage %1 beim Server an …").arg(request));

    switch (target) {
    case ProcessingTarget::Players:
        m_playersPage->setProcessingRequest(true, scrollToTarget);
        break;
    case ProcessingTarget::Markers:
        m_markersWidget->setProcessingRequest(true);
        break;
    case ProcessingTarget::Ranking:
        break;
    }

    jsonObject.insert(KEY_CHECKSUM,  m_playersPage->checksum());
    jsonObject.insert(KEY_CLIENT_ID, m_id);
    sendData(jsonObject);

    m_processingTarget.append(target);
    m_requestTimer->start();
}

int Client::id() const
{
    return m_id;
}

void Client::requestTimeout()
{
    m_playersPage->setProcessingRequest(false);
    m_markersWidget->setProcessingRequest(false);
    emit connectionError(tr("Der Server hat zu lange nicht auf eine Anfrage reagiert. Bitte die "
                            "Netzwerkverbindung überprüfen und neu verbinden."));
    closeConnection();
}

bool Client::connected() const
{
    return m_connected;
}

void Client::stopProcessingRequest()
{
    m_requestTimer->stop();
    switch (m_processingTarget.takeFirst()) {
    case Players:
        m_playersPage->setProcessingRequest(false);
        break;
    case Markers:
        m_markersWidget->setProcessingRequest(false);
        break;
    case Ranking:
        break;
    }
}

void Client::verifyChecksum(const QString &checksum)
{
    if (checksum != m_playersPage->checksum()) {
        // This should not happen. But who knows?!
        emit connectionError(tr("Die Synchronisation mit dem Server ist fehlgeschlagen, die Daten "
                                "weichen ab. Die Verbindung wird beendet. Bitte neu verbinden, um "
                                "den Datenbestand wieder mit dem Server abzugleichen!"));
    }
}

// ====================
// Handshake processing
// ====================

void Client::processHandshake(const QJsonObject &jsonObject)
{
    const QJsonValue phaseValue = jsonObject.value(HANDSHAKE_KEY_PHASE);
    if (phaseValue.type() != QJsonValue::String) {
        emit handshakeResult(false, tr("Der Server hat eine ungültige Begrüßung geschickt"));
        return;
    }

    const QString handshakePhase = phaseValue.toString();

    // Handshake Initialization
    // ========================

    if (handshakePhase == HANDSHAKE_PHASE_INITIALIZATION) {

        const QJsonValue serverSignatureValue = jsonObject.value(HANDSHAKE_KEY_SERVER_SIGNATURE);
        if (serverSignatureValue.type() != QJsonValue::String
            || serverSignatureValue.toString() != SERVER_SIGNATURE) {

            emit message(tr("Der Server hat eine ungültige Begrüßung geschickt."));
            emit handshakeResult(false, tr("Der Server hat eine ungültige Begrüßung geschickt."));
            return;
        }

        const QJsonValue protocolVersionValue = jsonObject.value(HANDSHAKE_KEY_PROTOCOL_VERSION);
        if (protocolVersionValue.type() != QJsonValue::Double) {
            emit message(tr("Der Server hat eine ungültige Begrüßung geschickt."));
            emit handshakeResult(false, tr("Der Server hat eine ungültige Begrüßung geschickt."));
            return;
        } else {
            const int protocolVersion = protocolVersionValue.toInt();
            if (protocolVersion != PROTOCOL_VERSION) {
                emit message(tr("Falsche Protokollrevision (ist %1, müsste %2 sein)").arg(
                                QString::number(protocolVersion),
                                QString::number(PROTOCOL_VERSION)));
                emit handshakeResult(false, tr(
                    "Der Server verwendet eine andere Protokollrevision (ist %1, müsste %2 sein). "
                    "Es kann keine Verbindung hergestellt werden.").arg(
                    QString::number(protocolVersion), QString::number(PROTOCOL_VERSION)));
                return;
            }
        }

        // The counterpart seems to be a Muckturnier server and we speak the same protocol revision
        emit message(tr("Begrüßung empfangen, frage Synchronisation an"));
        sendData(HANDSHAKE_MESSAGE_SYNCHRONIZE);

    // Data Synchronization
    // ===================

    } else if (handshakePhase == HANDSHAKE_PHASE_SYNCHRONIZE) {

        // Check the settings and state
        // ----------------------------

        const QJsonObject settings = jsonObject.value(HANDSHAKE_KEY_SETTINGS).toObject();
        Database::TournamentMode tournamentMode =
            settings.value(HANDSHAKE_KEY_TOURNAMENT_MODE).toString() == HANDSHAKE_FIXED_PAIRS
                ? Database::TournamentMode::FixedPairs
                : Database::TournamentMode::SinglePlayers;
        const int boogerScore = settings.value(HANDSHAKE_KEY_BOOGER_SCORE).toInt();
        const int boogersPerRound = settings.value(HANDSHAKE_KEY_BOOGERS_PER_ROUND).toInt();
        const bool networkTournamentStarted
                       = settings.value(HANDSHAKE_KEY_TOURNAMENT_STARTED).toBool();

        // Tournament settings
        if (m_db->tournamentMode() != tournamentMode
            || m_db->boogerScore() != boogerScore
            || m_db->boogersPerRound() != boogersPerRound) {

            emit message(tr("Die Datenbank des Servers hat andere Einstellungen als die lokale"));
            emit handshakeResult(false, tr(
                "Die Datenbank des Servers hat andere Einstellungen als die lokale:</p>"
                "<table>"
                "<tr>"
                    "<td></td>"
                    "<td>Server</td>"
                    "<td>Hier</td>"
                "</tr>"
                "<tr>"
                    "<td>Turniermodus:</td>"
                    "<td><i>%1</i>&nbsp;&nbsp;</td>"
                    "<td><i>%2</i></td>"
                "</tr>"
                "<tr>"
                    "<td>Punkte pro Bobbl:&nbsp;&nbsp;</td>"
                    "<td><i>%3</i></td>"
                    "<td><i>%4</i></td>"
                "</tr>"
                "<tr>"
                    "<td>Bobbl pro Runde:</td>"
                    "<td><i>%5</i></td>"
                    "<td><i>%6</i></td>"
                "</tr>"
                "</table>"
                "<p>Es kann keine Verbindung hergestellt werden.").arg(
                tournamentMode == Database::TournamentMode::FixedPairs
                    ? tr("Feste Paare") : tr("Einzelne Spieler"),
                m_db->isPairMode() ? tr("Feste Paare") : tr("Einzelne Spieler"),
                QString::number(boogerScore), QString::number(m_db->boogerScore()),
                QString::number(boogersPerRound), QString::number(m_db->boogersPerRound())));
            return;
        }

        // Registration phase state
        if (m_db->localTournamentStarted() && ! networkTournamentStarted) {
            emit message(tr("Es wurden bereits Ergebnisse eingegeben, aber nicht auf dem Server"));
            emit handshakeResult(false,
                     tr("Es wurden bereits Ergebnisse eingegeben, aber der Server hat die "
                        "Anmeldung noch nicht beendet.</p>"
                        "<p>Bitte nach Beendigung der Anmeldung erneut verbinden, oder die bisher "
                        "eingegebenen Ergebnisse löschen und erneut verbinden!"));
            return;
        }

        // Process the Server's data
        // -------------------------

        // Extract the markers
        QList<Marker::Marker> markers;
        for (const QJsonValue &value : jsonObject.value(HANDSHAKE_KEY_MARKERS).toArray()) {
            const QJsonObject entry = value.toObject();
            Marker::Marker marker;
            marker.id = entry.value(HANDSHAKE_KEY_MARKER_ID).toInt();
            marker.name = entry.value(HANDSHAKE_KEY_MARKER_NAME).toString();
            marker.color = colorFromValue(entry.value(HANDSHAKE_KEY_MARKER_COLOR));
            marker.sorting = static_cast<Marker::Sorting>(
                                entry.value(HANDSHAKE_KEY_MARKER_SORTING).toInt());
            markers.append(marker);
        }

        // Extract the players and marks
        QStringList players;
        QMap<QString, int> markedPlayers;
        for (const QJsonValue &value : jsonObject.value(HANDSHAKE_KEY_PLAYERS).toArray()) {
            const QJsonObject entry = value.toObject();
            const QString name = entry.keys().at(0);
            players.append(name);
            const int marker = entry.value(name).toInt();
            if (marker != 0) {
                markedPlayers[name] = marker;
            }
        }

        // Adopt the data
        // --------------

        // Adopt the locale
        const QString serverLocale = jsonObject.value(HANDSHAKE_KEY_LOCALE).toString();
        if (m_stringTools->locale() != serverLocale) {
            emit message(tr("Übernehme abweichendes Server-Gebietsschema „%1“ (die lokale "
                            "Einstellung ist „%2“)").arg(serverLocale, m_stringTools->locale()));
            m_stringTools->setLocale(serverLocale);
        }

        // Adopt the registration pahse state
        m_db->setNetworkTournamentStarted(networkTournamentStarted);

        // If we don't have results yet, we can simply adopt all data
        if (! m_db->localTournamentStarted()) {
            m_db->adoptServerData(players, markers, markedPlayers);
            m_playersPage->reload();

        // Otherwise, the local players list has to be matched and probably updated
        } else {
            QStringList localPlayers(m_db->listPlayers());
            if (localPlayers.count() != players.count()) {
                emit message(tr("Es wurden bereits Ergebnisse eingetragen, und es sind "
                                "unterschiedlich viele %1 angemeldet").arg(
                                m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
                emit handshakeResult(false, tr(
                    "Es wurden bereits Ergebnisse eingetragen, und es sind unterschiedlich viele "
                    "%1 angemeldet (am Server %2, hier %3). Es kann keine Verbindung hergestellt "
                    "werden!").arg(m_db->isPairMode() ? tr("Paare") : tr("Spieler"),
                                   QString::number(players.count()),
                                   QString::number(localPlayers.count())));
                return;
            }

            m_stringTools->sortList(localPlayers);
            m_stringTools->sortList(players);
            if (localPlayers == players) {
                // The lists match, we can leave it as-is and adopt the markers
                m_db->adoptServerData(markers, markedPlayers);
                m_playersPage->reload();
            } else {
                // The names differ, so we have to match the lists first
                emit namesDiffer(localPlayers, players);
                emit handshakeResult(false, tr(
                    "Es wurden bereits Ergebnisse eingetragen, und die %1 des Servers "
                    "unterscheidet sich von der lokalen. Bevor eine Verbindung hergestellt werden "
                    "kann, müssen alle Namen übereinstimmen.</p>"
                    "<p>Bitte die Liste überprüfen, die Namen übernehmen bzw. manuell korrigieren "
                    "und erneut verbinden!").arg(m_db->isPairMode() ? tr("Paarliste")
                                                                    : tr("Spielerliste")));
                return;
            }
        }

        // Here, the data should be in sync
        if (jsonObject.value(KEY_CHECKSUM).toString() == m_playersPage->checksum()) {
            // We are in sync with the server and everything seems to be fine, so we want an ID
            emit message(tr("Synchronisation war erfolgreich, frage ID an"));
            sendData(HANDSHAKE_MESSAGE_ALLOCATE_ID);
        } else {
            // This should not happen, but who knows?!
            emit message(tr("Die Übernahme der Daten des Servers ist fehlgeschlagen!"));
            emit handshakeResult(false, tr("Die Übernahme der Daten des Servers ist "
                                           "fehlgeschlagen!"));
            return;
        }

    // Finish the handshake
    // ====================

    } else if (handshakePhase == HANDSHAKE_PHASE_ALLOCATE_ID) {
        m_handshakeTimer->stop();
        m_connectionPending = false;
        m_id = jsonObject.value(KEY_CLIENT_ID).toInt();
        emit message(tr("Verbindung mit %1 Port %2 hergestellt (Client-ID %3 erhalten)").arg(
                        serverIp(), QString::number(serverPort()), QString::number(m_id)));
        emit handshakeResult(true);
        m_connected = true;
    }
}

// ====================================
// Change announcements from the server
// ====================================

void Client::dataReceived(const QString &action, const QJsonObject &jsonObject)
{
    if (action == ACTION_HANDSHAKE) {
        if (! m_connectionPending) {
            return;
        }
        processHandshake(jsonObject);
        return;
    }

    const int clientId = jsonObject.value(KEY_CLIENT_ID).toInt();
    if (clientId == m_id) {
        stopProcessingRequest();
    } else {
        QApplication::setOverrideCursor(Qt::BusyCursor);
    }

    processData(clientId, action, jsonObject);
}

void Client::statusMessage(const QString &message, int clientId)
{
    if (clientId == m_id) {
        emit statusUpdate(message);
    } else {
        emit statusUpdate(tr("Netzwerkänderung: ") + message);
    }
}

void Client::processData(int clientId, const QString &action, const QJsonObject &jsonObject)
{

// General checks
// ==============

    if (action == ACTION_REJECTED) {
        QApplication::restoreOverrideCursor();
        emit statusUpdate(tr("Anfrage vom Server zurückgewiesen"));
        emit requestRejected(static_cast<RejectReason>(
                 jsonObject.value(KEY_REJECT_REASON).toInt()));
        return;

    } else if (action == ACTION_UNKNOWN) {
        // This should not happen
        QApplication::restoreOverrideCursor();
        emit statusUpdate(tr("Anfrage vom Server zurückgewiesen"));
        emit connectionError(tr("Die Anfrage wurde vom Server nicht verstanden."));
        return;

// Players
// =======

    } else if (action == ACTION_PLAYERS_REGISTER) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        const int marker = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Melde „%1“ an").arg(name));
        m_db->registerPlayers(name, marker);
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("%1 „%2“ angemeldet").arg(
                         m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name),
                      clientId);

    } else if (action == ACTION_PLAYERS_DELETE) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        emit message(tr("Lösche „%1“").arg(name));
        m_db->deletePlayers(name);
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("%1 „%2“ gelöscht").arg(
                         m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name),
                      clientId);

    } else if (action == ACTION_PLAYERS_DELETE_MARKED) {
        const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
        const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Lösche alle als „%1“ markierten %2").arg(
                        markerName, m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
        m_db->deleteMarkedPlayers(markerId);
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("Alle als „%1“ markierten %2 gelöscht").arg(
                         markerName, m_db->isPairMode() ? tr("Paare") : tr("Spieler")),
                      clientId);

    } else if (action == ACTION_PLAYERS_DELETE_UNMARKED) {
        emit message(tr("Lösche alle unmarkierten %1").arg(
                        m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
        m_db->deleteUnmarkedPlayers();
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("Alle unmarkierten %1 gelöscht").arg(
                         m_db->isPairMode() ? tr("Paare") : tr("Spieler")),
                      clientId);

    } else if (action == ACTION_PLAYERS_DELETE_ALL) {
        emit message(tr("Lösche alle %1").arg(m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
        m_db->deleteAllPlayers();
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("Alle %1 gelöscht").arg(
                         m_db->isPairMode() ? tr("Paare") : tr("Spieler")),
                      clientId);

    } else if (action == ACTION_PLAYERS_RENAME) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        const QString newName = jsonObject.value(PLAYERS_KEY_NEW_NAME).toString();
        emit message(tr("Benenne „%1“ zu „%2“ um").arg(name, newName));
        m_db->renamePlayers(name, newName);
        m_playersPage->processNetworkChange(clientId);
        emit nameEdited();
        statusMessage(tr("%1 „%2“ zu „%3“ umbenannt").arg(
                         m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name, newName),
                      clientId);

    } else if (action == ACTION_PLAYERS_SET_MARKER) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
        const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Markiere „%1“ als „%2“").arg(name, markerName));
        m_db->setMarker(name, markerId);
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("%1 „%2“ als „%3“ markiert").arg(
                         m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name, markerName),
                      clientId);

    } else if (action == ACTION_PLAYERS_REMOVE_MARKER) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        emit message(tr("Entferne Markierung von „%1“").arg(name));
        m_db->removeMarker(name);
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("Markierung von %1 „%2“ entfernt").arg(
                         m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name),
                      clientId);

    } else if (action == ACTION_PLAYERS_LIST_SET_MARKER) {
        const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
        const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Markiere alle %1 als „%2“").arg(
                        m_db->isPairMode() ? tr("Paar") : tr("Spieler"), markerName));
        m_db->listSetMarker(markerId);
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("Alle %1 als „%2“ markiert").arg(
                         m_db->isPairMode() ? tr("Paare") : tr("Spieler"), markerName),
                      clientId);

    } else if (action == ACTION_PLAYERS_LIST_REMOVE_MARKER) {
        const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
        const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Entferne alle „%1“-Markierungen").arg(markerName));
        m_db->listRemoveMarker(markerId);
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("Alle „%1“-Markierungen entfernt").arg(markerName), clientId);

    } else if (action == ACTION_PLAYERS_LIST_REMOVE_ALL_MARKERS) {
        emit message(tr("Entferne alle Markierungen"));
        m_db->listRemoveAllMarkers();
        m_playersPage->processNetworkChange(clientId);
        statusMessage(tr("Alle Markierungen entfernt"), clientId);

// Markers
// =======

    } else if (action == ACTION_MARKERS_ADD) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        emit message(tr("Füge Markierung „%1“ hinzu").arg(name));
        m_db->addMarker(name, colorFromValue(jsonObject.value(MARKERS_KEY_COLOR)),
                        static_cast<Marker::Sorting>(
                            jsonObject.value(MARKERS_KEY_SORTING).toInt()));
        m_markersWidget->processNetworkChange();
        statusMessage(tr("Markierung „%1“ hinzugefügt").arg(name), clientId);

    } else if (action == ACTION_MARKERS_DELETE) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        emit message(tr("Lösche Markierung „%1“").arg(name));
        m_db->deleteMarker(jsonObject.value(MARKERS_KEY_ID).toInt(),
                           jsonObject.value(MARKERS_KEY_SEQUENCE).toInt());
        m_markersWidget->processNetworkChange();
        statusMessage(tr("Markierung „%1“ gelöscht").arg(name), clientId);

    } else if (action == ACTION_MARKERS_MOVE_UP) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        emit message(tr("Verschiebe Markierung „%1“").arg(name));
        m_db->moveMarkerUp(jsonObject.value(MARKERS_KEY_ID).toInt(),
                           jsonObject.value(MARKERS_KEY_SEQUENCE).toInt());
        m_markersWidget->processNetworkChange();
        statusMessage(tr("Markierung „%1“ verschoben").arg(name), clientId);

    } else if (action == ACTION_MARKERS_MOVE_DOWN) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        emit message(tr("Verschiebe Markierung „%1“").arg(name));
        m_db->moveMarkerDown(jsonObject.value(MARKERS_KEY_ID).toInt(),
                             jsonObject.value(MARKERS_KEY_SEQUENCE).toInt());
        m_markersWidget->processNetworkChange();
        statusMessage(tr("Markierung „%1“ verschoben").arg(name), clientId);

    } else if (action == ACTION_MARKERS_EDIT) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        emit message(tr("Speichere Änderungen an Markierung „%1“").arg(name));
        m_db->editMarker(jsonObject.value(MARKERS_KEY_ID).toInt(),
                         jsonObject.value(MARKERS_KEY_NEW_NAME).toString(),
                         colorFromValue(jsonObject.value(MARKERS_KEY_COLOR)),
                         static_cast<Marker::Sorting>(
                             jsonObject.value(MARKERS_KEY_SORTING).toInt()));
        m_markersWidget->processNetworkChange();
        statusMessage(tr("Änderungen an Markierung „%1“ gespeichert").arg(name), clientId);

// Finish registration
// ===================

    } else if (action == ACTION_FINISH_REGISTRATION) {
        emit message(tr("Beende die Anmeldung"));
        m_playersPage->networkFinishRegistration(clientId);
        statusMessage(tr("Anmeldung beendet"), clientId);

// Ranking
// =======

    } else if (action == ACTION_RANKING_REQUEST) {
        if (! jsonObject.value(RANKING_KEY_AVAILABLE).toBool()) {
            QApplication::restoreOverrideCursor();
            m_rankingDialog->noRanking();
            emit message(tr("Es ist noch keine Rangliste verfügbar"));
            return;
        }

        const QJsonArray names = jsonObject.value(RANKING_KEY_NAMES).toArray();
        const QJsonArray boogers = jsonObject.value(RANKING_KEY_BOOGERS).toArray();
        const QJsonArray goals = jsonObject.value(RANKING_KEY_GOALS).toArray();
        const QJsonArray opponentGoals = jsonObject.value(RANKING_KEY_OPPONENT_GOALS).toArray();
        const QJsonArray playedBoogers = jsonObject.value(RANKING_KEY_PLAYED_BOOGERS).toArray();
        QList<Database::RankData> ranking;
        for (int i = 0; i < names.count(); i++) {
            Database::RankData rank;
            rank.name = names.at(i).toString();
            rank.boogers = boogers.at(i).toInt();
            rank.goals = goals.at(i).toInt();
            rank.opponentGoals = opponentGoals.at(i).toInt();
            rank.playedBoogers = playedBoogers.at(i).toInt();
            ranking.append(rank);
        }
        m_rankingDialog->compareRanking(ranking);

        QApplication::restoreOverrideCursor();
        emit message(tr("Rangliste empfangen"));
        emit statusUpdate(tr("Rangliste empfangen"));
        return;

// Unhandled answer
// ================

    } else {
        // This should not happen
        QApplication::restoreOverrideCursor();
        emit connectionError(tr("Die Antwort des Servers konnte nicht verarbeitet werden."));
        return;
    }

// Checksum comparison
// ===================

    // If we didn't return early
    QApplication::restoreOverrideCursor();
    verifyChecksum(jsonObject.value(KEY_CHECKSUM).toString());
}

// ======================
// Requests to the server
// ======================

// Players
// =======

void Client::requestRegisterPlayers(const QString &name, const int marker)
{
    QJsonObject message { { KEY_ACTION,            ACTION_PLAYERS_REGISTER },
                          { PLAYERS_KEY_NAME,      name },
                          { PLAYERS_KEY_MARKER_ID, marker } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Anmelden von „%1“").arg(name),
                name);
}

void Client::requestRenamePlayers(const QString &name, const QString &newName)
{
    QJsonObject message { { KEY_ACTION,           ACTION_PLAYERS_RENAME },
                          { PLAYERS_KEY_NAME,     name },
                          { PLAYERS_KEY_NEW_NAME, newName } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Umbenennen von „%1“ zu „%2“").arg(name, newName),
                newName);
}

void Client::requestDeletePlayers(const QString &name)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_DELETE },
                          { PLAYERS_KEY_NAME, name } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Löschen von „%1“").arg(name));
}

void Client::requestDeleteMarkedPlayers(const QString &markerName, const int markerId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_DELETE_MARKED },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Löschen aller mit „%1“ markierten %2").arg(
                   markerName, m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Client::requestDeleteUnmarkedPlayers()
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_UNMARKED } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Frage Löschen aller unmarkierten %1").arg(
                   m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Client::requestDeleteAllPlayers()
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_ALL } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Löschen aller %1 an").arg(m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Client::requestSetMarker(const QString &name, const QString &markerName, const int markerId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_SET_MARKER },
                          { PLAYERS_KEY_NAME,        name },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Markieren von „%1“ als „%2“ an").arg(name, markerName),
                name);
}

void Client::requestRemoveMarker(const QString &name)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_REMOVE_MARKER },
                          { PLAYERS_KEY_NAME, name } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Entfernen der Markierung von „%1“ an").arg(name),
                name);
}

void Client::requestListSetMarker(const QString &markerName, const int markerId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_LIST_SET_MARKER },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Markieren aller %1 als „%2“ an").arg(
                   m_db->isPairMode() ? tr("Paare") : tr("Spieler"), markerName));
}

void Client::requestListRemoveMarker(const QString &markerName, const int markerId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_LIST_REMOVE_MARKER },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Entfernen aller „%1“-Markierungen").arg(markerName));
}

void Client::requestListRemoveAllMarkers()
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_LIST_REMOVE_ALL_MARKERS } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Frage Entfernen aller Markierungen"));
}

// Markers
// =======

void Client::requestAddMarker(const QString &name, const QColor &color,
                              const Marker::Sorting sorting)
{
    QJsonObject message { { KEY_ACTION,          ACTION_MARKERS_ADD },
                          { MARKERS_KEY_NAME,    name },
                          { MARKERS_KEY_COLOR,   colorToString(color) },
                          { MARKERS_KEY_SORTING, sorting } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Hinzufügen der Markierung „%1“").arg(name));
}

void Client::requestDeleteMarker(const QString &name, const int id, const int sequence)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_DELETE },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_SEQUENCE, sequence } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Frage Löschen der Markierung „%1“").arg(name));
}

void Client::requestMoveMarkerUp(const QString &name, const int id, const int sequence)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_MOVE_UP },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_SEQUENCE, sequence } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Verschieben der Markierung „%1“").arg(name));
}

void Client::requestMoveMarkerDown(const QString &name, const int id, const int sequence)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_MOVE_DOWN },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_SEQUENCE, sequence } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Verschieben der Markierung „%1“").arg(name));
}

void Client::requestEditMarker(const QString &name, const int id, const QString &newName,
                               const QColor &color, const Marker::Sorting sorting)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_EDIT },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_NEW_NAME, newName },
                          { MARKERS_KEY_COLOR,    colorToString(color) },
                          { MARKERS_KEY_SORTING,  sorting } };
    sendRequest(message, ProcessingTarget::Markers,
                tr("Bearbeiten der Markierung „%1“").arg(name));
}

// Finish registration
// ===================

void Client::requestFinishRegistration()
{
    QJsonObject message { { KEY_ACTION, ACTION_FINISH_REGISTRATION } };
    sendRequest(message, ProcessingTarget::Players,
                tr("Beenden der Anmeldung"));
}

// Ranking
// =======

void Client::requestRanking(RankingDialog *dialog)
{
    m_rankingDialog = dialog;
    QJsonObject message { { KEY_ACTION, ACTION_RANKING_REQUEST } };
    sendRequest(message, ProcessingTarget::Ranking,
                tr("Rangliste"));
}
