/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CLIENT_H
#define CLIENT_H

// Local includes
#include "shared/AbstractNetworkEngine.h"
#include "shared/ServerProtocol.h"
#include "shared/Database.h"
#include "shared/Marker.h"

// Qt includes
#include <QHostAddress>

// Local classes
class PlayersPage;
class SharedObjects;
class StringTools;
class RankingDialog;
class MarkersWidget;

// Qt classes
class QTimer;
class QJsonObject;

class Client : public AbstractNetworkEngine
{
    Q_OBJECT

public:
    enum ProcessingTarget {
        Players,
        Markers,
        Ranking
    };

    explicit Client(QObject *parent, SharedObjects *sharedObjects, PlayersPage *playersPage);
    void connectToServer(const QHostAddress &address, quint16 port);
    void closeConnection();
    QString serverIp() const;
    int serverPort() const;
    int id() const;
    bool connected() const;

    void abortHandshake();
    void requestRegisterPlayers(const QString &name, const int marker);
    void requestRenamePlayers(const QString &name, const QString &newName);
    void requestDeletePlayers(const QString &name);
    void requestDeleteMarkedPlayers(const QString &markerName, const int markerId);
    void requestDeleteUnmarkedPlayers();
    void requestDeleteAllPlayers();
    void requestSetMarker(const QString &name, const QString &markerName, const int markerId);
    void requestRemoveMarker(const QString &name);
    void requestListSetMarker(const QString &markerName, const int markerId);
    void requestListRemoveMarker(const QString &markerName, const int markerId);
    void requestListRemoveAllMarkers();

    void requestAddMarker(const QString &name, const QColor &color,
                          const Marker::Sorting sorting);
    void requestDeleteMarker(const QString &name, const int id, const int sequence);
    void requestMoveMarkerUp(const QString &name, const int id, const int sequence);
    void requestMoveMarkerDown(const QString &name, const int id, const int sequence);
    void requestEditMarker(const QString &name, const int id, const QString &newName,
                           const QColor &color, const Marker::Sorting sorting);

    void requestFinishRegistration();

    void requestRanking(RankingDialog *dialog);

signals:
    void message(const QString &text);
    void handshakeResult(bool success, const QString &error = QString());
    void connectionError(const QString &text);
    void requestRejected(ServerProtocol::RejectReason reason);
    void statusUpdate(const QString &text);
    void namesDiffer(QStringList &localNames, QStringList &remoteNames);
    void nameEdited();

private slots:
    void socketConnected();
    void error(QAbstractSocket::SocketError);
    void handshakeTimeout();
    void requestTimeout();

private: // Functions
    virtual void dataReceived(const QString &action, const QJsonObject &jsonObject) override;
    void processData(int clientId, const QString &action, const QJsonObject &jsonObject);
    void statusMessage(const QString &message, int clientId);
    void sendRequest(QJsonObject &jsonObject, ProcessingTarget target,
                     const QString &request, const QString &scrollToTarget = QString());
    void processHandshake(const QJsonObject &jsonObject);
    void verifyChecksum(const QString &checksum);
    void stopProcessingRequest();

private: // Variables
    Database *m_db;
    StringTools *m_stringTools;
    PlayersPage *m_playersPage;
    int m_id;
    bool m_connected;
    bool m_connectionPending;
    QTimer *m_handshakeTimer;
    QTimer *m_requestTimer;
    RankingDialog *m_rankingDialog;
    MarkersWidget *m_markersWidget;
    QVector<ProcessingTarget> m_processingTarget;

};

#endif // CLIENT_H
