/* Copyright (c) 2012-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "ExportTournamentDialog.h"

#include "shared/version.h"
#include "shared/Characters.h"
#include "shared/SharedObjects.h"
#include "shared/Database.h"
#include "shared/StringTools.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QLabel>
#include <QVBoxLayout>
#include <QCheckBox>
#include <QLineEdit>
#include <QDir>
#include <QApplication>
#include <QTextCodec>
#include <QDateTime>
#include <QTableWidget>
#include <QRadioButton>
#include <QComboBox>

// C++ includes
#include <algorithm> // std::sort

ExportTournamentDialog::ExportTournamentDialog(QWidget *parent, SharedObjects *sharedObjects)
    : FileDialog(parent, sharedObjects->database()),
      m_stringTools(sharedObjects->stringTools())
{
    mainLayout()->setSizeConstraint(QLayout::SetFixedSize);

    setTitle(tr("Turnierdaten exportieren"));
    setOkButtonText(tr("Exportieren"));
    setFileDialogTitle(tr("Bitte einen Dateinamen für die Turnierdaten-Exportdatei auswählen"));

    QGroupBox *optionsBox = new QGroupBox(tr("Exportoptionen"));
    QVBoxLayout *optionsBoxLayout = new QVBoxLayout(optionsBox);
    layout()->addWidget(optionsBox);

    // Title

    QHBoxLayout *titleEntryLayout = new QHBoxLayout;
    optionsBoxLayout->addLayout(titleEntryLayout);
    m_exportTitle = new QCheckBox(tr("Turniertitel ausgeben:"));
    m_exportTitle->setChecked(true);
    titleEntryLayout->addWidget(m_exportTitle);

    m_titleEdit = new QLineEdit;
    titleEntryLayout->addWidget(m_titleEdit);
    connect(m_titleEdit, &QLineEdit::textEdited, this, &ExportTournamentDialog::checkTitle);

    // Show options or not

    QCheckBox *customExport = new QCheckBox(tr("Export anpassen"));
    customExport->setChecked(Qt::Unchecked);
    optionsBoxLayout->addWidget(customExport);
    connect(customExport, &QCheckBox::stateChanged, this, &ExportTournamentDialog::setExportType);

    // Formatting

    m_formatBox = new QGroupBox(tr("Formatierung"));
    QVBoxLayout *formatBoxLayout = new QVBoxLayout(m_formatBox);
    m_formatBox->hide();
    optionsBoxLayout->addWidget(m_formatBox);

    m_exportHeadlines = new QCheckBox(tr("Überschriften der Datensätze anzeigen "
                                         "(z.\u2009B. „Einzelergebnisse“)"));
    m_exportHeadlines->setChecked(true);
    formatBoxLayout->addWidget(m_exportHeadlines);

    m_exportSubHeadlines = new QCheckBox(tr("Unterüberschriften anzeigen (z.\u2009B. "
                                            "„Spielergebnisse der 1. Runde“)"));
    m_exportSubHeadlines->setChecked(true);
    formatBoxLayout->addWidget(m_exportSubHeadlines);

    m_displayBoogers = new QCheckBox(tr("Bobbl-Symbole („%1“) anzeigen").arg(
                                        m_stringTools->boogerSymbol()));
    m_displayBoogers->setChecked(true);
    formatBoxLayout->addWidget(m_displayBoogers);

    // Datasets to export

    m_selectDataBox = new QGroupBox(tr("Zu exportierende Datensätze"));
    QVBoxLayout *selectDataBoxLayout = new QVBoxLayout(m_selectDataBox);
    m_selectDataBox->hide();
    optionsBoxLayout->addWidget(m_selectDataBox);

    m_exportGenericData = new QCheckBox(tr("Allgemeine Daten (Datum, Anzahl Teilnehmer etc.)"));
    m_exportGenericData->setChecked(true);
    selectDataBoxLayout->addWidget(m_exportGenericData);
    connect(m_exportGenericData, &QCheckBox::stateChanged,
            this, &ExportTournamentDialog::checkExportPossible);

    m_exportRanking = new QCheckBox(tr("Rangliste zum Turnierende"));
    m_exportRanking->setChecked(true);
    selectDataBoxLayout->addWidget(m_exportRanking);
    connect(m_exportRanking, &QCheckBox::stateChanged,
            this, &ExportTournamentDialog::checkExportPossible);

    m_exportRoundResults = new QCheckBox;
    m_exportRoundResults->setChecked(true);
    selectDataBoxLayout->addWidget(m_exportRoundResults);
    connect(m_exportRoundResults, &QCheckBox::stateChanged,
            this, &ExportTournamentDialog::checkExportPossible);

    QHBoxLayout *roundSelectLayout = new QHBoxLayout;
    selectDataBoxLayout->addLayout(roundSelectLayout);

    m_roundSelect = new QLineEdit;
    m_defaultTextColor = m_roundSelect->palette().color(QPalette::Foreground);
    roundSelectLayout->addWidget(m_roundSelect);
    connect(m_roundSelect, &QLineEdit::textChanged,
            this, &ExportTournamentDialog::checkRoundSelection);

    m_roundSelectLabel = new QLabel(tr("Format z.\u2009B.: „1, 3, 7-9, 11“"));
    roundSelectLayout->addWidget(m_roundSelectLabel);

    m_errorLabel = new QLabel;
    selectDataBoxLayout->addWidget(m_errorLabel);
    m_errorLabel->hide();

    layout()->addStretch();

    QGroupBox *fileFormatBox = new QGroupBox(tr("Dateiformat"));
    QVBoxLayout *fileFormatBoxLayout = new QVBoxLayout(fileFormatBox);
    layout()->addWidget(fileFormatBox);

    m_exportHTML = new QRadioButton(tr("HTML"));
    m_exportHTML->setChecked(true);
    fileFormatBoxLayout->addWidget(m_exportHTML);

    m_exportCSV = new QRadioButton(tr("CSV (UTF-8 mit Tabulatoren als Trenner)"));
    fileFormatBoxLayout->addWidget(m_exportCSV);

    QGroupBox *actionsBox = new QGroupBox(tr("Aktionen"));
    QVBoxLayout *actionsBoxLayout = new QVBoxLayout(actionsBox);
    layout()->addWidget(actionsBox);

    m_openAfterExport = new QCheckBox(tr("Datei nach dem Export öffnen"));
    m_openAfterExport->setChecked(true);
    actionsBoxLayout->addWidget(m_openAfterExport);

    m_closeAfterExport = new QCheckBox(tr("Dialog nach dem Export schließen"));
    connect(m_closeAfterExport, &QCheckBox::stateChanged,
            this, [this](bool state) { setButtonText(QDialogButtonBox::Cancel,
                                           state ? tr("Abbrechen") : tr("Schließen")); } );

    m_closeAfterExport->setChecked(true);
    actionsBoxLayout->addWidget(m_closeAfterExport);

    // Fill in data

    m_dateFormat = QLocale().dateFormat(QLocale::ShortFormat);
    if (! m_dateFormat.contains(QStringLiteral("yyyy"))) {
        m_dateFormat = m_dateFormat.replace(QStringLiteral("yy"), QStringLiteral("yyyy"));
    }
    m_timeFormat = QLocale().timeFormat(QLocale::ShortFormat);

    m_allRounds = db()->roundsState().count();
    if (m_allRounds == 1) {
        m_exportRoundResults->setText(tr("Einzelergebnisse der folgenden Runden "
                                         "(Es wurde eine Runde gespielt)"));
        m_roundSelect->setText(QStringLiteral("1"));
    } else {
        QString lastRound = QString::number(m_allRounds);
        m_exportRoundResults->setText(tr("Einzelergebnisse der folgenden Runden "
                                         "(Es wurden %1 Runden gespielt)").arg(lastRound));
        m_roundSelect->setText(QStringLiteral("1-") + lastRound);
    }

    // Add the database's filename as the default title and set a reasonable width
    m_titleEdit->setText(QFileInfo(db()->dbFile()).completeBaseName());
    m_titleEdit->setMinimumSize(m_titleEdit->sizeHint().width() * 2,
                                m_titleEdit->sizeHint().height());

    m_titleEdit->setFocus();
    m_titleEdit->selectAll();
}

void ExportTournamentDialog::setExportType(bool custom)
{
    if (custom) {
        m_selectDataBox->show();
        m_formatBox->show();
    } else {
        m_exportTitle->setChecked(true);
        m_exportHeadlines->setChecked(true);
        m_exportSubHeadlines->setChecked(true);
        m_displayBoogers->setChecked(true);
        m_exportGenericData->setChecked(true);
        m_exportRanking->setChecked(true);
        m_exportRoundResults->setChecked(true);

        if (m_allRounds == 1) {
            m_roundSelect->setText(QStringLiteral("1"));
        } else {
            m_roundSelect->setText(QStringLiteral("1-") + QString::number(m_allRounds));
        }
        m_selectDataBox->hide();
        m_formatBox->hide();
    }
}

void ExportTournamentDialog::checkTitle(const QString &title)
{
    m_exportTitle->setChecked(! title.simplified().isEmpty());
    m_exportTitle->setEnabled(! title.simplified().isEmpty());
}

void ExportTournamentDialog::roundSelectionIsInvalid(const QString &message)
{
    if (! message.isEmpty()) {
        m_errorLabel->show();
        m_errorLabel->setText(message);
    }
    setRoundSelectColor(Qt::red);
    m_roundSelectionIsRed = true;
    m_exportRoundResults->setEnabled(false);
    m_exportRoundResults->setChecked(false);
    checkExportPossible();
}

void ExportTournamentDialog::checkRoundSelection(const QString &range)
{
    m_selectedRounds.clear();

    // No input: range is not okay, but set the default color
    if (range.isEmpty()) {
        roundSelectionIsInvalid();
        setRoundSelectColor(m_defaultTextColor);
        m_errorLabel->hide();
        return;
    }

    // Remove all whitespace
    const QString parsedRange = range.simplified().remove(QLatin1String(" "));

    // Search for input other than 0-9, "," and "-"
    const QRegularExpression check(QStringLiteral("([^\\d,-]+)"));
    const QRegularExpressionMatch match = check.match(parsedRange);
    if (match.hasMatch()) {
        roundSelectionIsInvalid(tr("Ungültige Eingabe: „%1“").arg(match.captured()));
        return;
    }

    // Process all parts
    for (const QString &part : parsedRange.split(QLatin1String(","))) {

        // Single round
        if (! part.contains(QLatin1String("-"))) {
            // Skip empty parts
            if (part.isEmpty()) {
                continue;
            }

            // Check the requested round
            const int requestedRound = part.toInt();
            if (requestedRound > 0 && requestedRound <= m_allRounds) {
                m_stringTools->appendIfNotExists(m_selectedRounds, requestedRound);
            } else {
                roundSelectionIsInvalid(tr("Ungültige Rundenzahl: %1").arg(requestedRound));
                return;
            }

        // Range of rounds
        } else {
            const QStringList givenRange = part.split(QLatin1String("-"));

            // We have to find exacly two parts
            if (givenRange.size() != 2) {
                roundSelectionIsInvalid(tr("Ungültige Intervallangabe: „%1“").arg(part));
                return;
            }

            // About to enter, e. g. "1-", handle like a single round entry
            if (givenRange.at(1).isEmpty()) {
                const int requestedRound = givenRange.at(0).toInt();
                if (requestedRound > 0 && requestedRound <= m_allRounds) {
                    m_stringTools->appendIfNotExists(m_selectedRounds, requestedRound);
                } else {
                    roundSelectionIsInvalid(tr("Ungültige Rundenzahl: „%1“").arg(requestedRound));
                    return;
                }

            // True range entered
            } else {
                const int fromRound = givenRange.at(0).toInt();
                const int toRound = givenRange.at(1).toInt();

                // Check the range
                if (fromRound == 0 || fromRound > m_allRounds
                    || toRound == 0 || toRound > m_allRounds
                    || fromRound > toRound) {

                    roundSelectionIsInvalid(tr("Ungültiges Intervall: „%1“").arg(part));
                    return;
                }

                // Add all rounds between
                for (int requestedRound = fromRound; requestedRound <= toRound; requestedRound++) {
                    m_stringTools->appendIfNotExists(m_selectedRounds, requestedRound);
                }
            }
        }
    }

    if (m_roundSelectionIsRed) {
        setRoundSelectColor(m_defaultTextColor);
        m_roundSelectionIsRed = false;
        m_errorLabel->hide();
    }

    m_exportRoundResults->setEnabled(true);
    m_exportRoundResults->setChecked(true);

    std::sort(m_selectedRounds.begin(), m_selectedRounds.end());
    checkExportPossible();
}

void ExportTournamentDialog::checkExportPossible()
{
    setOkButtonEnabled(m_exportGenericData->isChecked()
                       || m_exportRanking->isChecked()
                       || m_exportRoundResults->isChecked());
}

void ExportTournamentDialog::setRoundSelectColor(const QColor &color)
{
    QPalette palette;
    palette.setColor(QPalette::Text, color);
    m_roundSelect->setPalette(palette);
}

void ExportTournamentDialog::accept()
{
    QString title = m_titleEdit->text().simplified();
    if (! title.isEmpty()) {
        setTargetFileBaseName(title);
    } else {
        setTargetFileBaseName(tr("Protokoll"));
    }

    QString message;
    if (m_exportHTML->isChecked()) {
        setTargetFileExtension(QStringLiteral("htm"));
        setTargetFileDescription(tr("HTML-Dateien"));
        message = tr("Exportiere HTML-Daten …");
    } else {
        setTargetFileExtension(QStringLiteral("csv"));
        setTargetFileDescription(tr("CSV-Dateien"));
        message = tr("Exportiere CSV-Daten …");
    }

    QString journalFileName = getTargetFileName(db()->dbPath());
    if (journalFileName.isEmpty()) {
        return;
    }

    QFile journalFile(journalFileName);
    journalFile.open(QIODevice::WriteOnly | QIODevice::Text);

    QTextStream outStream(&journalFile);
    outStream.setCodec(QTextCodec::codecForName("UTF-8"));
    emit updateStatus(message);
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (m_exportHTML->isChecked()) {
        saveHTMLFile(outStream);
    } else {
        saveCSVFile(outStream);
    }

    journalFile.close();
    QApplication::restoreOverrideCursor();
    emit updateStatus(tr("Daten in Datei „%1“ exportiert").arg(
                         QDir::toNativeSeparators(journalFileName)));

    successMessage(tr("Turnierdaten exportieren"),
                   tr("Die Daten wurden erfolgreich in die Datei „%1“ exportiert.").arg(
                      QDir::toNativeSeparators(journalFileName)));

    if (m_openAfterExport->isChecked()) {
        openFile(journalFileName);
    }

    if (m_closeAfterExport->isChecked()) {
        QDialog::accept();
    }
}

void ExportTournamentDialog::saveHTMLFile(QTextStream &outStream)
{
    // Add the HTML header
    outStream << "<!DOCTYPE html>\n\n"
              << "<html lang=\"de\">\n\n"
              << "<head>\n\n"
              << "<meta charset=\"utf-8\"/>\n\n"
              << "<title>" << targetFileBaseName().toHtmlEscaped() << "</title>\n\n"
              << "<style>\n"
              << "table { border-spacing: 0px; }\n"
              << "th, td { padding: 0.2em 1em 0.2em; text-align: left; }\n"
              << "tr.line td { border-top: 1px solid black; }\n"
              << "</style>\n\n"
              << "</head>\n\n"
              << "<body>\n\n";

    if (m_exportTitle->isChecked()) {
        outStream << "<h1>" << targetFileBaseName().toHtmlEscaped() << "</h1>\n\n";
    }

    if (m_exportGenericData->isChecked()) {
        // Add tournament meta data

        outStream << "<section>\n\n";

        if (m_exportHeadlines->isChecked()) {
            outStream << "<h2>" << tr("Allgemeine Daten") << "</h2>\n\n";
        }
        outStream << "<p>" << tr("Protokoll erstellt am %1 um %2 Uhr.").arg(
                                 QDateTime::currentDateTime().toString(m_dateFormat),
                                 QDateTime::currentDateTime().toString(m_timeFormat)) << "</p>\n\n";

        if (m_exportSubHeadlines->isChecked()) {
            outStream << "<h3>Muckturnier</h3>\n\n";
        }

        outStream << "<table>\n";
        outStream << "<tr><td>" << tr("Homepage:") << "</td>"
                  << "<td><a href=\"http://muckturnier.org/\">http://muckturnier.org/</a></td>"
                  << "</tr>\n";
        if (QStringLiteral(MT_VERSION) == db()->muckturnierVersion()) {
            outStream << "<tr class=\"line\">"
                      << "<td>" << tr("Turnierdatenbank und Protokoll erstellt mit:") << "</td>"
                      << "<td>Version " << MT_VERSION << "</td>"
                      << "</tr>\n";
        } else {
            outStream << "<tr class=\"line\"><td>" << tr("Turnierdatenbank erstellt mit:")
                      << "</td><td>" << tr("Version ")
                      << db()->muckturnierVersion() << "</td></tr>\n"
                      << "<tr><td>" << tr("Protokoll erstellt mit:")
                      << "</td><td>" << tr("Version ")
                      << MT_VERSION << "</td></tr>\n";
        }
        outStream << "<tr><td>" << tr("Datenbankrevision:") << "</td>"
                  << "<td>" << tr("Version ") << db()->dbVersion() << "</td></tr>\n"
                  << "</table>\n\n";

        if (m_exportSubHeadlines->isChecked()) {
            outStream << "<h3>" << tr("Turnier") << "</h3>\n\n";
        }

        outStream << "<table>\n"
                  << "<tr><td>" << tr("Turniertyp:") << "</td>"
                  << "<td>" << (db()->isPairMode() ? tr("Feste Paare") : tr("Einzelne Spieler"))
                  << "</td></tr>\n"
                  << "<tr><td>" << tr("Punkte pro Bobbl:") << "</td>"
                  << "<td>" << QString::number(db()->boogerScore()) << "</td></tr>\n"
                  << "<tr><td>" << tr("Bobbl pro Runde:") << "</td>"
                  << "<td>" << QString::number(db()->boogersPerRound()) << "</td></tr>\n"
                  << "<tr class=\"line\"><td>"
                  << (db()->isPairMode() ? tr("Teilnehmende Paare:") : tr("Teilnehmende Spieler:"))
                  << "</td>"
                  << "<td>" << QString::number(db()->playersCount())
                  << "</td></tr>\n"
                  << "<tr><td>" << tr("Gespielte Runden:") << "</td>"
                  << "<td>" << QString::number(m_allRounds) << "</td></tr>\n"
                  << "</table>\n\n";

        outStream << "</section>\n\n";
    }

    if (m_exportRanking->isChecked()) {
        // Add the ranking

        outStream << "<section>\n\n";

        if (m_exportHeadlines->isChecked()) {
            outStream << "<h2>" << tr("Rangliste nach %n Runde(n)", "", m_allRounds) << "</h2>\n\n";
        }
        outStream << "<table>\n"
                  << "<thead>\n"
                  << "<tr>"
                  << "<th>" << tr("Platz") << "</th>"
                  << "<th>" << (db()->isPairMode() ? tr("Paar") : tr("Spieler")) << "</th>"
                  << "<th>" << tr("Bobbl") << "</th>"
                  << "<th>" << tr("Tore") << "</th>";
        if (db()->includeOpponentGoals()) {
            outStream << "<th>" << tr("Gegn. Tore") << "</th>";
        }
        outStream << "</tr>\n</thead>\n"
                  << "<tbody>\n";

        int lastRank = -1;
        QString style;

        for (const Database::RankData &currentRank : db()->ranking()) {
            if (currentRank.rank != lastRank) {
                style = QStringLiteral(" class=\"line\"");
                lastRank = currentRank.rank;
            } else {
                style.clear();
            }

            outStream << "<tr" << style << "><td>" << QString::number(currentRank.rank) << "</td>";
            outStream << "<td>" << currentRank.name << "</td>";
            outStream << "<td>" << currentRank.boogers << "</td>";
            outStream << "<td>" << currentRank.goals << "</td>";
            if (db()->includeOpponentGoals()) {
                outStream << "<td>" << currentRank.opponentGoals << "</td>";
            }
            outStream << "</tr>\n";
        }
        outStream << "</tbody>\n"
                  << "</table>\n\n";

        outStream << "</section>\n\n";
    }

    if (m_exportRoundResults->isChecked()) {
        // Add all requested round data

        outStream << "<section>\n\n";

        if (m_exportHeadlines->isChecked()) {
            outStream << "<h2>" << tr("Einzelergebnisse") << "</h2>\n\n";
        }

        for (const int &round : m_selectedRounds) {
            if (m_exportSubHeadlines->isChecked()) {
                outStream << "<h3>" << tr("Spielergebnisse der %1. Runde").arg(
                                          QString::number(round))
                          << "</h3>\n\n";
            }
            outStream << "<table>\n"
                      << "<thead>\n"
                      << "<tr>";
            if (! db()->noTableNumbers()) {
                outStream << "<th>" << tr("Tisch") << "</th>";
            }
            outStream << "<th>" << tr("Paar") << "</th>";

            for (int booger = 1; booger <= db()->boogersPerRound(); booger++) {
                outStream << "<th>" << tr("%1. Bobbl").arg(QString::number(booger)) << "</th>";
            }
            outStream << "</tr>\n"
                      << "</thead>\n"
                      << "<tbody>\n";

            const QList<Database::Score> allScores = db()->scores(round);
            for (int i = 0; i < allScores.size(); i++) {
                const QString tableNumber = QString::number(allScores.at(i).table);
                const QString pair1Name = allScores.at(i).pairName;
                QStringList pair1Scores;
                for (int j = 0; j < db()->boogersPerRound(); j++) {
                    pair1Scores.append(m_displayBoogers->isChecked()
                        ? m_stringTools->boogify(allScores.at(i).score, db()->boogerScore())
                        : QString::number(allScores.at(i).score));
                    i++;
                }
                const QString pair2Name = allScores.at(i).pairName;
                QStringList pair2Scores;
                for (int j = 0; j < db()->boogersPerRound(); j++) {
                    pair2Scores.append(m_displayBoogers->isChecked()
                        ? m_stringTools->boogify(allScores.at(i).score, db()->boogerScore())
                        : QString::number(allScores.at(i).score));
                    i++;
                }
                i--;

                outStream << "<tr class=\"line\">";
                if (! db()->noTableNumbers()) {
                    outStream << "<td rowspan=\"2\">" << tableNumber << "</td>";
                }
                outStream << "<td>" << pair1Name.toHtmlEscaped() << "</td>";
                for (const QString &score : pair1Scores) {
                    outStream << "<td>" << score << "</td>";
                }
                outStream << "</tr>\n"
                          << "<tr><td>" << pair2Name.toHtmlEscaped() << "</td>";
                for (const QString &score : pair2Scores) {
                    outStream << "<td>" << score << "</td>";
                }
                outStream << "</tr>\n";
            }
            outStream << "</tbody>\n"
                      << "</table>\n\n";
        }

        outStream << "</section>\n\n";
    }

    // Close HTML
    outStream << "</body>\n\n"
              << "</html>\n";
}

void ExportTournamentDialog::saveCSVFile(QTextStream &outStream)
{
    if (m_exportTitle->isChecked()) {
        outStream << m_titleEdit->text().simplified() << "\n\n";
    }

    if (m_exportGenericData->isChecked()) {
        // Add tournament meta data

        if (m_exportHeadlines->isChecked()) {
            outStream << tr("Allgemeine Daten") << "\n\n";
        }

        outStream << tr("Protokoll erstellt am %1 um %2 Uhr.").arg(
                        QDateTime::currentDateTime().toString(m_dateFormat),
                        QDateTime::currentDateTime().toString(m_timeFormat))
                  << "\n\n";

        if (m_exportSubHeadlines->isChecked()) {
            outStream << "Muckturnier\n";
        }

        outStream << tr("Homepage:") << "\t" << "http://muckturnier.org/\n";
        if (QStringLiteral(MT_VERSION) == db()->muckturnierVersion()) {
            outStream << tr("Turnierdatenbank und Protokoll erstellt mit:") << "\t"
                      << "Version " << MT_VERSION << "\n";
        } else {
            outStream << tr("Turnierdatenbank erstellt mit:") << "\tVersion "
                      << db()->muckturnierVersion() << "\n"
                      << tr("Protokoll erstellt mit:") << "\tVersion "
                      << MT_VERSION << "\n";
        }
        outStream << tr("Datenbankrevision:") << "\t"
                  << tr("Version ") << db()->dbVersion() << "\n\n";

        if (m_exportSubHeadlines->isChecked()) {
            outStream << tr("Turnier") << "\n\n";
        }

        outStream << tr("Turniertyp:")
                  << "\t" << (db()->isPairMode() ? tr("Feste Paare") : tr("Einzelne Spieler"))
                  << "\n"
                  << tr("Punkte pro Bobbl:")
                  << "\t" << QString::number(db()->boogerScore()) << "\n"
                  << tr("Bobbl pro Runde:")
                  << "\t" << QString::number(db()->boogersPerRound()) << "\n"
                  << (db()->isPairMode() ? tr("Teilnehmende Paare:") : tr("Teilnehmende Spieler:"))
                  << "\t" << QString::number(db()->playersCount()) << "\n"
                  << tr("Gespielte Runden:") << "\t"
                  << QString::number(m_allRounds) << "\n\n";
    }

    if (m_exportRanking->isChecked()) {
        // Add the ranking

        if (m_exportHeadlines->isChecked()) {
            outStream << tr("Rangliste nach %n Runde(n)", "", m_allRounds) << "\n\n";
        }
        outStream << tr("Platz") << "\t"
                  << (db()->isPairMode() ? tr("Paar") : tr("Spieler")) << "\t"
                  << tr("Bobbl") << "\t"
                  << tr("Tore");
        if (db()->includeOpponentGoals()) {
            outStream << "\t" << tr("Gegn. Tore");
        }
        outStream << "\n";

        for (const Database::RankData &currentRank : db()->ranking()) {
            outStream << QString::number(currentRank.rank);
            outStream << "\t" << currentRank.name;
            outStream << "\t" << QString::number(currentRank.boogers);
            outStream << "\t" << QString::number(currentRank.goals);
            if (db()->includeOpponentGoals()) {
                outStream << "\t" << QString::number(currentRank.opponentGoals);
            }
            outStream << "\n";
        }
        outStream << "\n";
    }

    if (m_exportRoundResults->isChecked()) {
        // Add all requested round data

        if (m_exportHeadlines->isChecked()) {
            outStream << tr("Einzelergebnisse") << "\n\n";
        }

        QString tableString;
        QString tableTabString;
        if (! db()->noTableNumbers()) {
            tableString = tr("Tisch\t");
            tableTabString = QStringLiteral("\t");
        }

        for (const int &round : m_selectedRounds) {
            if (m_exportSubHeadlines->isChecked()) {
                outStream << tr("Spielergebnisse der %1. Runde").arg(QString::number(round))
                          << "\n";
            }
            outStream << tableString << tr("Paar");
            for (int booger = 1; booger <= db()->boogersPerRound(); booger++) {
                outStream << "\t" << tr("%1. Bobbl").arg(QString::number(booger));
            }
            outStream << "\n";

            const QList<Database::Score> allScores = db()->scores(round);
            for (int i = 0; i < allScores.size(); i++) {
                const QString tableNumber = QString::number(allScores.at(i).table);
                const QString pair1Name = allScores.at(i).pairName;
                QStringList pair1Scores;
                for (int j = 0; j < db()->boogersPerRound(); j++) {
                    pair1Scores.append(m_displayBoogers->isChecked()
                        ? m_stringTools->boogify(allScores.at(i).score, db()->boogerScore())
                        : QString::number(allScores.at(i).score));
                    i++;
                }
                const QString pair2Name = allScores.at(i).pairName;
                QStringList pair2Scores;
                for (int j = 0; j < db()->boogersPerRound(); j++) {
                    pair2Scores.append(m_displayBoogers->isChecked()
                        ? m_stringTools->boogify(allScores.at(i).score, db()->boogerScore())
                        : QString::number(allScores.at(i).score));
                    i++;
                }
                i--;

                if (! db()->noTableNumbers()) {
                    outStream << tableNumber << "\t";
                }
                outStream << pair1Name;
                for (const QString &score : pair1Scores) {
                    outStream << "\t" << score;
                }
                outStream << "\n" << tableTabString << pair2Name;
                for (const QString &score : pair2Scores) {
                    outStream << "\t" << score;
                }
                outStream << "\n";
            }
            outStream << "\n";
        }
    }
}
