/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "MarkersWidget.h"
#include "MarkerDialog.h"

#include "ServerPage/Server.h"
#include "ClientPage/Client.h"

#include "shared/SharedObjects.h"
#include "shared/ResourceFinder.h"
#include "shared/TableDelegate.h"
#include "shared/TitleMenu.h"
#include "shared/defaultValues.h"

// Qt includes
#include <QVBoxLayout>
#include <QTableWidget>
#include <QDebug>
#include <QHeaderView>
#include <QGroupBox>
#include <QApplication>
#include <QMessageBox>
#include <QFontMetrics>
#include <QCheckBox>
#include <QComboBox>
#include <QTimer>
#include <QLabel>
#include <QLineEdit>

// C++ includes
#include <functional>

namespace
{
const QString NORMAL_TITLE_WEIGHT = QStringLiteral("QGroupBox { font-weight: normal; }");
}

MarkersWidget::MarkersWidget(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent), m_db(sharedObjects->database())
{
    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    // This eliminates additional space around this widget when displayed inside the players page
    mainLayout->setContentsMargins(0, 0, 0, 0);

    // Markers list

    QGroupBox *markersBox = new QGroupBox(tr("Markierungen"));
    QVBoxLayout *markersBoxLayout = new QVBoxLayout(markersBox);
    mainLayout->addWidget(markersBox);

    m_markers = new QTableWidget;
    m_markers->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_markers->setSelectionMode(QAbstractItemView::NoSelection);
    m_markers->setFocusPolicy(Qt::NoFocus);
    m_markers->setColumnCount(4);
    m_markers->setHorizontalHeaderLabels( { tr("n", "Anzahl"), tr("Name"), tr("Sortierung"),
                                            QString() } );
    m_markers->verticalHeader()->hide();
    const int iconSize = QFontMetrics(QFont(font())).boundingRect(QStringLiteral("M")).height()
#ifdef Q_OS_WIN32
                         * 1.6; // On Windows, the icon looks better if it's a bit bigger
#else
                         * 1.3; // On Linux, a bit smaller looks better
#endif
    m_markers->setIconSize(QSize(iconSize, iconSize));
    m_markers->setShowGrid(false);
    m_markers->setItemDelegate(new TableDelegate(this));
    m_markers->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_markers, &QWidget::customContextMenuRequested, this, &MarkersWidget::showContextMenu);
    connect(m_markers, &QTableWidget::cellDoubleClicked,
            this, &MarkersWidget::checkShowContextMenu);
    connect(m_markers, &QTableWidget::cellClicked, this, &MarkersWidget::changeVisibility);
    markersBoxLayout->addWidget(m_markers);

    // Default markers for new pairs/players

    m_automaticMarkingBox = new QGroupBox(tr("Automatische Markierungen"));
    QVBoxLayout *automaticMarkingBoxLayout = new QVBoxLayout(m_automaticMarkingBox);
    mainLayout->addWidget(m_automaticMarkingBox);

    QGroupBox *defaultMarkerBox = new QGroupBox(tr("Voreinstellung für neue Einträge"));
    defaultMarkerBox->setStyleSheet(NORMAL_TITLE_WEIGHT);
    QHBoxLayout *defaultMarkerBoxLayout = new QHBoxLayout(defaultMarkerBox);
    automaticMarkingBoxLayout->addWidget(defaultMarkerBox);

    QLabel *defaultMarkerLabel = new QLabel(tr("Markierung:"));
    defaultMarkerLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    defaultMarkerBoxLayout->addWidget(defaultMarkerLabel);
    m_defaultMarker = new QComboBox;
    m_defaultMarker->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    defaultMarkerBoxLayout->addWidget(m_defaultMarker);

    m_conditionalMarkerBox = new QGroupBox(tr("Bedingte Markierung"));
    m_conditionalMarkerBox->setStyleSheet(NORMAL_TITLE_WEIGHT);
    m_conditionalMarkerBox->setCheckable(true);
    m_conditionalMarkerBox->setChecked(false);
    QVBoxLayout *conditionalMarkerBoxLayout = new QVBoxLayout(m_conditionalMarkerBox);
    automaticMarkingBoxLayout->addWidget(m_conditionalMarkerBox);

    QWidget *conditionalMarkerWidget = new QWidget;
    QGridLayout *conditionalMarkerWidgetLayout = new QGridLayout(conditionalMarkerWidget);
    conditionalMarkerBoxLayout->addWidget(conditionalMarkerWidget);
    conditionalMarkerWidget->setEnabled(false);
    connect(m_conditionalMarkerBox, &QGroupBox::toggled,
            conditionalMarkerWidget, &QWidget::setEnabled);

    QLabel *nameLabel = new QLabel(tr("Name"));
    nameLabel->setAlignment(Qt::AlignRight);
    conditionalMarkerWidgetLayout->addWidget(nameLabel, 0, 0);
    m_markerCondition = new QComboBox;
    m_markerCondition->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_markerCondition->addItem(tr("enthält nicht"));
    m_markerCondition->addItem(tr("enthält"));
    conditionalMarkerWidgetLayout->addWidget(m_markerCondition, 0, 1);

    m_markerConditionString = new QLineEdit(QStringLiteral(DEFAULT_NAMES_SEPARATOR).simplified());
    m_markerConditionString->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    conditionalMarkerWidgetLayout->addWidget(m_markerConditionString, 0, 2);

    conditionalMarkerWidgetLayout->addWidget(new QLabel(tr("Markierung:")), 1, 0);
    m_conditionMarker = new QComboBox;
    m_conditionMarker->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Maximum);
    conditionalMarkerWidgetLayout->addWidget(m_conditionMarker, 1, 1, 1, 2);

    // Context menu

    m_markerMenu = new TitleMenu(this);

    m_moveUpAction = m_markerMenu->addAction(tr("Nach oben bewegen"));
    connect(m_moveUpAction, &QAction::triggered, this, &MarkersWidget::moveUp);

    m_moveDownAction = m_markerMenu->addAction(tr("Nach unten bewegen"));
    connect(m_moveDownAction, &QAction::triggered, this, &MarkersWidget::moveDown);

    m_markerMenu->addSeparator();

    m_editMarkerAction = m_markerMenu->addAction(tr("Bearbeiten"));
    connect(m_editMarkerAction, &QAction::triggered, this, &MarkersWidget::editMarker);

    m_deleteMarkerAction = m_markerMenu->addAction(tr("Löschen"));
    connect(m_deleteMarkerAction, &QAction::triggered, this, &MarkersWidget::deleteMarker);

    m_markerMenu->addSeparator();

    m_addMarkerAction = m_markerMenu->addAction(tr("Neue Markierung"));
    connect(m_addMarkerAction, &QAction::triggered, this, &MarkersWidget::addMarker);

    m_addMarkerMenu = new QMenu(this);
    m_addMarkerMenu->addAction(m_addMarkerAction);

    // Hide/Show context menu

    m_hideShowMarkedMenu = new QMenu(this);

    m_showOnlyAction = m_hideShowMarkedMenu->addAction(QString());

    m_showAllAction = m_hideShowMarkedMenu->addAction(tr("Alle anzeigen"));
    connect(m_showAllAction, &QAction::triggered, this, &MarkersWidget::showAll);

    // Icons
    m_unused = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("unused.svg")));
    m_visible = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("visible.svg")));
    m_hidden = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("hidden.svg")));
}

void MarkersWidget::reload()
{
    const QVariant lastDefaultMarkerData = m_defaultMarker->currentData();
    const int lastDefaultMarker = lastDefaultMarkerData.isValid()
                                      ? lastDefaultMarkerData.toInt() : -1;
    const QVariant lastConditionMarkerData = m_defaultMarker->currentData();
    const int lastConditionMarker = lastConditionMarkerData.isValid()
                                      ? lastConditionMarkerData.toInt() : -1;

    const QList<Marker::Marker> markers = m_db->markers();

    m_markers->setRowCount(markers.count());
    m_defaultMarker->clear();
    m_conditionMarker->clear();

    QList<int> ids;
    int row = 0;
    for (const Marker::Marker &marker : markers) {
        ids.append(marker.id);
        m_markers->setItem(row, 0, new QTableWidgetItem);

        QTableWidgetItem *item = new QTableWidgetItem(marker.name);
        if (marker.name.isEmpty()) {
            item->setText(tr("(unmarkiert)"));
        }
        item->setForeground(marker.color);
        item->setData(MarkerData::Id, QVariant(marker.id));
        item->setData(MarkerData::Sorting, QVariant(marker.sorting));
        m_markers->setItem(row, 1, item);

        m_defaultMarker->addItem(item->text(), marker.id);
        m_conditionMarker->addItem(item->text(), marker.id);

        switch (marker.sorting) {
        case Marker::Sorting::NewBlock:
            item = new QTableWidgetItem(tr("Neuer Block"));
            break;
        case Marker::Sorting::Merged:
            item = new QTableWidgetItem(tr("Fortlaufend"));
            break;
        }
        m_markers->setItem(row, 2, item);

        item = new QTableWidgetItem;
        m_markers->setItem(row, 3, item);

        row++;
    }

    // Set new markers visible
    for (int id : ids) {
        if (! m_isVisible.contains(id)) {
            m_isVisible[id] = true;
        }
    }

    // Remove visibility entries for deleted markers
    for (int id : m_isVisible.keys()) {
        if (! ids.contains(id)) {
            m_isVisible.remove(id);
        }
    }

    m_markers->horizontalHeaderItem(0)->setWhatsThis(tr(
        "In dieser Spalte wird die Anzahl der mit der entsprechenden Markierung gekennzeichneten "
        "%1 angezeigt.").arg(m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_markers->horizontalHeaderItem(3)->setWhatsThis(tr(
        "Durch Klicken auf das Augensymbol können alle mit der entsprechenden Markierung "
        "gekennzeichneten %1 ein- bzw. ausgeblendet werden.").arg(
        m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_markers->resizeColumnsToContents();
    emit markersChanged();

    const int lastDefaultMarkerIndex = m_defaultMarker->findData(lastDefaultMarker);
    m_defaultMarker->setCurrentIndex(lastDefaultMarkerIndex != -1
                                         ? lastDefaultMarkerIndex
                                         : m_defaultMarker->findData(0));
    const int lastConditionMarkerIndex = m_conditionMarker->findData(lastConditionMarker);
    m_conditionMarker->setCurrentIndex(lastConditionMarkerIndex != -1
                                           ? lastConditionMarkerIndex
                                           : m_conditionMarker->findData(0));
}

void MarkersWidget::showContextMenu(const QPoint &position)
{
    int column = -1;
    QString name;
    int id;
    int row;
    const QPoint target = m_markers->viewport()->mapToGlobal(position);

    // Collect data
    QTableWidgetItem *markerItem = m_markers->itemAt(position);
    if (markerItem != nullptr) {
        column = markerItem->column();
        row = markerItem->row();
        if (column != 1) {
            markerItem = m_markers->item(row, 1);
        }
        name = m_markers->item(row, 1)->text();
        id = m_markers->item(row, 1)->data(MarkerData::Id).toInt();
    }

    // Show no context menu on the counter column
    if (column == 0) {
        return;
    }

    // Show the "Hide/Show marked entries" context menu on the "Hide/Show" column
    if (column == 3) {
        m_showAllAction->setEnabled(false);
        for (bool visible : m_isVisible.values()) {
            if (! visible) {
                m_showAllAction->setEnabled(true);
                break;
            }
        }

        // Set the context menu to handle the selected marker
        m_showOnlyAction->disconnect();
        m_showOnlyAction->setText(tr("Nur „%1“ anzeigen").arg(name));
        connect(m_showOnlyAction, &QAction::triggered,
                this, std::bind(&MarkersWidget::showOnly, this, id));

        m_hideShowMarkedMenu->exec(target);
        return;
    }

    // All other actions require write access, so return here if the database if ro
    if (! m_db->isWritable()) {
        return;
    }

    // Show the "Add marker" context menu if no item was clicked
    if (column == -1) {
        m_addMarkerMenu->exec(target);
        return;
    }

    // Otherwise, show the marker context menu

    m_markerMenu->setTitle(tr("Markierung „%1“").arg(name));
    m_selectedMarker.id = id;
    m_selectedMarker.name = name;
    m_selectedMarker.color = markerItem->foreground().color();
    m_selectedMarker.sorting = static_cast<Marker::Sorting>(
                                   markerItem->data(MarkerData::Sorting).toInt());
    m_selectedMarkerSequence = markerItem->row();
    m_moveUpAction->setEnabled(m_selectedMarkerSequence > 0);
    m_moveDownAction->setEnabled(m_selectedMarkerSequence < m_markers->rowCount() - 1);
    m_editMarkerAction->setEnabled(m_selectedMarker.id != 0);
    m_deleteMarkerAction->setEnabled(m_selectedMarker.id != 0);

    m_markerMenu->exec(target);
}

void MarkersWidget::addMarker()
{
    MarkerDialog *dialog = new MarkerDialog(this, m_markers->palette().color(QPalette::Base),
                                            m_markers);
    connect(dialog, &MarkerDialog::markerSaved, this, &MarkersWidget::saveNewMarker);

    QString name = tr("Neue Markierung");
    int counter = 1;
    QString newName = name;
    while (m_markers->findItems(newName, Qt::MatchFixedString).size() > 0) {
        counter++;
        newName = QStringLiteral("%1 %2").arg(name, QString::number(counter));
    }

    dialog->setName(newName);
    dialog->setTitle(tr("Neue Markierung"));
    dialog->setButtonText(QDialogButtonBox::Ok, tr("Hinzufügen"));

    dialog->exec();
}

void MarkersWidget::saveNewMarker(const QString name, const QColor color,
                                  const Marker::Sorting sorting)
{
    if (m_isClient) {
        m_client->requestAddMarker(name, color, sorting);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    emit statusUpdate(tr("Lege neue Markierung an …"));
    m_db->addMarker(name, color, sorting);
    reload();

    if (m_isServer) {
        m_server->broadcastAddMarker(name, color, sorting);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ hinzugefügt").arg(name));
}

void MarkersWidget::deleteMarker()
{
    if (QMessageBox::question(this,
            tr("Markierung löschen"),
            tr("Soll die Markierung „%1“ wirklich gelöscht werden?").arg(m_selectedMarker.name),
            QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Cancel) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteMarker(m_selectedMarker.name, m_selectedMarker.id,
                                      m_selectedMarkerSequence);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    emit statusUpdate(tr("Lösche Markierung „%1“ …").arg(m_selectedMarker.name));
    m_db->deleteMarker(m_selectedMarker.id, m_selectedMarkerSequence);
    reload();

    if (m_isServer) {
        m_server->broadcastDeleteMarker(m_selectedMarker.name, m_selectedMarker.id,
                                        m_selectedMarkerSequence);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ gelöscht").arg(m_selectedMarker.name));
}

void MarkersWidget::editMarker()
{
    MarkerDialog *dialog = new MarkerDialog(this, m_markers->palette().color(QPalette::Base),
                                            m_markers);
    connect(dialog, &MarkerDialog::markerSaved, this, &MarkersWidget::saveEdit);

    dialog->setTitle(tr("Markierung bearbeiten"));
    dialog->setButtonText(QDialogButtonBox::Ok, tr("Speichern"));
    dialog->setName(m_selectedMarker.name);
    dialog->setEditedName(m_selectedMarker.name);
    dialog->setColor(m_selectedMarker.color);
    dialog->setSorting(m_selectedMarker.sorting);
    if (m_selectedMarkerSequence == 0) {
        dialog->disableSortingSelection();
    }
    dialog->setIsEditing();

    dialog->exec();
}

void MarkersWidget::saveEdit(const QString name, const QColor color, const Marker::Sorting sorting)
{
    if (name == m_selectedMarker.name
        && color == m_selectedMarker.color
        && sorting == m_selectedMarker.sorting) {

        emit statusUpdate(tr("Markierung „%1“ wurde nicht verändert").arg(m_selectedMarker.name));
        return;
    }

    if (m_isClient) {
        m_client->requestEditMarker(m_selectedMarker.name, m_selectedMarker.id,
                                    name, color, sorting);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Speichere Änderungen an Markierung „%1“ …").arg(m_selectedMarker.name));
    m_db->editMarker(m_selectedMarker.id, name, color, sorting);
    reload();

    if (m_isServer) {
        m_server->broadcastEditMarker(m_selectedMarker.name, m_selectedMarker.id,
                                      name, color, sorting);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Änderungen an Markierung „%1“ gespeichert").arg(m_selectedMarker.name));
}

void MarkersWidget::moveDown()
{
    if (m_isClient) {
        m_client->requestMoveMarkerDown(m_selectedMarker.name, m_selectedMarker.id,
                                        m_selectedMarkerSequence);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Verschiebe Markierung „%1“ …").arg(m_selectedMarker.name));
    m_db->moveMarkerDown(m_selectedMarker.id, m_selectedMarkerSequence);
    reload();

    if (m_isServer) {
        m_server->broadcastMoveMarkerDown(m_selectedMarker.name, m_selectedMarker.id,
                                          m_selectedMarkerSequence);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ verschoben").arg(m_selectedMarker.name));
}

void MarkersWidget::moveUp()
{
    if (m_isClient) {
        m_client->requestMoveMarkerUp(m_selectedMarker.name, m_selectedMarker.id,
                                      m_selectedMarkerSequence);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Verschiebe Markierung „%1“ …").arg(m_selectedMarker.name));
    m_db->moveMarkerUp(m_selectedMarker.id, m_selectedMarkerSequence);
    reload();

    if (m_isServer) {
        m_server->broadcastMoveMarkerUp(m_selectedMarker.name, m_selectedMarker.id,
                                        m_selectedMarkerSequence);
    }

    QApplication::restoreOverrideCursor();
    emit statusUpdate(tr("Markierung „%1“ verschoben").arg(m_selectedMarker.name));
}

QList<Marker::Marker> MarkersWidget::getMarkers() const
{
    QList<Marker::Marker> markers;
    for (int row = 0; row < m_markers->rowCount(); row++) {
        Marker::Marker marker;
        QTableWidgetItem *item = m_markers->item(row, 1);
        marker.id = item->data(MarkerData::Id).toInt();
        marker.name = item->text();
        marker.color = item->foreground().color();
        marker.sorting = static_cast<Marker::Sorting>(item->data(MarkerData::Sorting).toInt());
        markers.append(marker);
    }
    return markers;
}

void MarkersWidget::checkShowContextMenu(int, const int column)
{
    if (column == 1 || column == 2) {
        showContextMenu(m_markers->viewport()->mapFromGlobal(QCursor::pos()));
    }
}

void MarkersWidget::updateVisibilityIcon(const int id, const int row)
{
    if (! m_isVisible[id]) {
        m_markers->item(row, 3)->setIcon(m_hidden);
    } else {
        if (m_markers->item(row, 0)->text().toInt() == 0) {
            m_markers->item(row, 3)->setIcon(m_unused);
        } else {
            m_markers->item(row, 3)->setIcon(m_visible);
        }
    }
}

void MarkersWidget::changeVisibility(const int row, const int column)
{
    if (column != 3) {
        // Not clicked on eye symbol
        return;
    }

    const int id = m_markers->item(row, 1)->data(MarkerData::Id).toInt();
    m_isVisible[id] = ! m_isVisible[id];
    updateVisibilityIcon(id, row);

    emit visibilityChanged();
}

void MarkersWidget::updateVisibilityIcons()
{
    for (int row = 0; row < m_markers->rowCount(); row++) {
        updateVisibilityIcon(m_markers->item(row, 1)->data(MarkerData::Id).toInt(), row);
    }
    emit visibilityChanged();
}

void MarkersWidget::reset()
{
    m_isVisible.clear();
    m_defaultMarker->setCurrentIndex(m_defaultMarker->findData(0));
    m_markerCondition->setCurrentIndex(0);
    m_markerConditionString->setText(QStringLiteral(DEFAULT_NAMES_SEPARATOR).simplified());
    m_conditionMarker->setCurrentIndex(m_conditionMarker->findData(0));
    m_conditionalMarkerBox->setChecked(false);
}

bool MarkersWidget::isMarkerVisible(const int id) const
{
    return m_isVisible.value(id);
}

void MarkersWidget::updateCounters(const QMap<int, int> &counter)
{
    m_allMarkersCount = 0;
    int markerCount;
    int id;

    for (int row = 0; row < m_markers->rowCount(); row++) {
        id = m_markers->item(row, 1)->data(MarkerData::Id).toInt();
        markerCount = counter[id];
        m_markers->item(row, 0)->setText(QString::number(markerCount));
        m_allMarkersCount += markerCount;
    }

    updateVisibilityIcons();
    m_markers->resizeColumnsToContents();
}

void MarkersWidget::showAll()
{
    for (int id : m_isVisible.keys()) {
        m_isVisible[id] = true;
    }
    updateVisibilityIcons();
}

void MarkersWidget::showOnly(const int idToShow)
{
    for (int id : m_isVisible.keys()) {
        m_isVisible[id] = (id == idToShow);
    }
    updateVisibilityIcons();
}

int MarkersWidget::defaultMarker(const QString &name) const
{
    int marker = m_defaultMarker->currentData().toInt();
    if (m_conditionalMarkerBox->isChecked()) {
        bool contains = name.contains(m_markerConditionString->text().simplified());
        // The first m_markerCondition entry is "doesn't contain" (index 0), the second is
        // "contains" (index 1), so we can simply use the selected index here, because it will be
        // (per official C++ behavior) silently casted to a bool value in the following comparison
        if (contains == m_markerCondition->currentIndex()) {
            marker = m_conditionMarker->currentData().toInt();
        }
    }
    return marker;
}

void MarkersWidget::addChecksumData(QCryptographicHash &hash) const
{
    for (int i = 0; i < m_markers->rowCount(); i++) {
        QTableWidgetItem *item = m_markers->item(i, 1);

        const int id = item->data(MarkerData::Id).toInt();
        const QString name = id != 0 ? item->text() : QString();

        hash.addData(QByteArray::number(id));
        hash.addData(name.toUtf8());
        hash.addData(QByteArray::number(item->foreground().color().rgb()));
        hash.addData(QByteArray::number(item->data(MarkerData::Sorting).toInt()));
    }
}

void MarkersWidget::setServer(Server *server)
{
    m_server = server;
    m_isServer = true;
}

void MarkersWidget::setClient(Client *client)
{
    m_client = client;
    m_isClient = true;
}

void MarkersWidget::serverStopped()
{
    m_isServer = false;
}

void MarkersWidget::clientDisconnected()
{
    m_isClient = false;
}

void MarkersWidget::processNetworkChange()
{
    if (m_markerMenu->isVisible()) {
        m_markerMenu->hide();
    }

    reload();

    MarkerDialog *dialog = findChild<MarkerDialog *>();
    if (dialog != nullptr && dialog->isEditing()) {
        QTableWidgetItem *item = m_markers->item(m_selectedMarkerSequence, 1);
        if (item == nullptr ||
            (item->text() != m_selectedMarker.name
             || item->data(MarkerData::Id).toInt() != m_selectedMarker.id
             || item->data(MarkerData::Sorting) != m_selectedMarker.sorting
             || item->foreground().color() != m_selectedMarker.color)) {

            dialog->reject();
            dialog->deleteLater();

            // This is invoked via a QTimer::singleShot call so that the function returns
            // immediately and the network processing is not blocked until the dialog is closed
            QTimer::singleShot(0, this, [this] {
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("Die Markierung „%1“ wurde zwischenzeitlich von einem anderen "
                       "Netzwerkteilnehmer entweder gelöscht, umbenannt oder geändert. Das Ändern "
                       "ist nicht mehr möglich und wird abgebrochen.").arg(
                       m_selectedMarker.name)); } );
            return;
        }
    }
}

void MarkersWidget::setProcessingRequest(const bool processing)
{
    setEnabled(! processing);
    if (processing) {
        QApplication::setOverrideCursor(Qt::WaitCursor);
    }
}

void MarkersWidget::setAutomaticMarkingBoxVisible(const bool state)
{
    m_automaticMarkingBox->setVisible(state);
}
