/* Copyright (c) 2012-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "PlayersPage.h"
#include "PlayersPageDelegate.h"
#include "MarkersWidget.h"
#include "MarkersPopup.h"

#include "ServerPage/Server.h"
#include "ClientPage/Client.h"

#include "shared/SharedObjects.h"
#include "shared/StringTools.h"
#include "shared/Database.h"
#include "shared/SearchWidget.h"
#include "shared/SearchEngine.h"
#include "shared/TitleMenu.h"
#include "shared/BackupEngine.h"
#include "shared/ResourceFinder.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QTableWidget>
#include <QHeaderView>
#include <QScrollBar>
#include <QAction>
#include <QMessageBox>
#include <QApplication>
#include <QSplitter>
#include <QCryptographicHash>
#include <QCheckBox>
#include <QDir>
#include <QTimer>

// C++ includes
#include <functional>

PlayersPage::PlayersPage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent), m_stringTools(sharedObjects->stringTools()),
                       m_db(sharedObjects->database()),
                       m_searchEngine(sharedObjects->searchEngine())
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    // Register new players/pairs

    QWidget *registerWidget = new QWidget;
    QVBoxLayout *registerLayout = new QVBoxLayout(registerWidget);

    m_newPlayersBox = new QGroupBox;
    m_newPlayersBox->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    QHBoxLayout *newPlayersBoxLayout = new QHBoxLayout(m_newPlayersBox);
    registerLayout->addWidget(m_newPlayersBox);

    m_nameLabel = new QLabel;
    newPlayersBoxLayout->addWidget(m_nameLabel);

    m_playersName = new QLineEdit;
    newPlayersBoxLayout->addWidget(m_playersName);
    connect(m_playersName, &QLineEdit::textChanged, [this] (const QString &text) {
            m_registerButton->setEnabled(! text.simplified().isEmpty()); } );
    connect(m_playersName, &QLineEdit::returnPressed, this, &PlayersPage::registerPlayers);

    m_registerButton = new QPushButton(tr("Anmelden"));
    m_registerButton->setEnabled(false);
    newPlayersBoxLayout->addWidget(m_registerButton);
    connect(m_registerButton, &QPushButton::clicked, this, &PlayersPage::registerPlayers);

    // Registered players/pairs

    m_playersBox = new QGroupBox;
    QVBoxLayout *playersBoxLayout = new QVBoxLayout(m_playersBox);
    registerLayout->addWidget(m_playersBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    playersBoxLayout->addWidget(m_isSearchLabel);

    m_playersList = new QTableWidget(this);
    m_playersList->setColumnCount(3);
    m_playersList->setColumnHidden(1, true);
    m_playersList->setColumnHidden(2, true);
    m_delegate = new PlayersPageDelegate(m_playersList);
    m_playersList->setItemDelegate(m_delegate);
    connect(m_delegate, &PlayersPageDelegate::startedEdit,
            this, [this] (int row) { m_editedRow = row; } );
    connect(m_delegate, &PlayersPageDelegate::editEntered, this, &PlayersPage::editEntered);
    m_playersList->setHorizontalHeaderLabels( { tr("Name"), QString(), tr("Markierung") } );
    m_playersList->horizontalHeader()->hide();
    m_playersList->verticalHeader()->hide();
    const int iconSize = QFontMetrics(QFont(font())).boundingRect(QStringLiteral("M")).height()
#ifdef Q_OS_WIN32
                         * 1.9; // On Windows, the icon looks better if it's a bit bigger
#else
                         * 1.5; // On Linux, a bit smaller looks better
#endif
    m_playersList->setIconSize(QSize(iconSize, iconSize));
    m_playersList->setShowGrid(false);
    m_playersList->setWordWrap(false);
    m_playersList->setSelectionMode(QTableWidget::NoSelection);
    m_playersList->setFocusPolicy(Qt::NoFocus);
    m_playersList->setHorizontalScrollMode(QAbstractItemView::ScrollPerPixel);
    connect(m_playersList->verticalScrollBar(), &QScrollBar::valueChanged,
            m_playersList, &QTableWidget::resizeColumnsToContents);
    m_playersList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_playersList, &QTableWidget::customContextMenuRequested,
            this, &PlayersPage::showContextMenu);
    connect(m_playersList, &QTableWidget::cellClicked, this, &PlayersPage::showMarkerPopup);

    m_markerIcon = QIcon(sharedObjects->resourceFinder()->find(QStringLiteral("marker.svg")));

    playersBoxLayout->addWidget(m_playersList);

    // Search widget
    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &PlayersPage::updateSearch);
    playersBoxLayout->addWidget(m_searchWidget);

    // Markers

    m_markersWidget = new MarkersWidget(this, sharedObjects);
    connect(m_markersWidget, &MarkersWidget::markersChanged, this, &PlayersPage::markersChanged);
    connect(m_markersWidget, &MarkersWidget::statusUpdate, this, &PlayersPage::statusUpdate);
    connect(m_markersWidget, &MarkersWidget::visibilityChanged,
            this, &PlayersPage::markersVisibilityChanged);

    QWidget *markersPartWidget = new QWidget;
    QVBoxLayout *markersPartWidgetLayout = new QVBoxLayout(markersPartWidget);
    markersPartWidgetLayout->addWidget(m_markersWidget);

    // Column options

    m_showSetMarkerColumn = new QCheckBox(tr("Spalte „Markieren“ anzeigen"));
    connect(m_showSetMarkerColumn, &QCheckBox::stateChanged,
            this, &PlayersPage::showSetMarkerColumn);
    markersPartWidgetLayout->addWidget(m_showSetMarkerColumn);

    m_showMarkerColumn = new QCheckBox(tr("Spalte „Markierung“ anzeigen"));
    connect(m_showMarkerColumn, &QCheckBox::stateChanged, this, &PlayersPage::showMarkerColumn);
    markersPartWidgetLayout->addWidget(m_showMarkerColumn);

    // Arrange the registering part and the markers part in a splitter
    QSplitter *splitter = new QSplitter(this);
    layout->addWidget(splitter);
    splitter->addWidget(registerWidget);
    splitter->setCollapsible(0, false);
    splitter->setStretchFactor(0, 5);
    splitter->addWidget(markersPartWidget);
    splitter->setCollapsible(1, true);
    splitter->setStretchFactor(1, 3);

    // Status label
    m_statusLabel = new QLabel;
    m_statusLabel->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    layout->addWidget(m_statusLabel);

    // Markers popup
    m_markersPopup = new MarkersPopup(this);
    connect(m_markersPopup, &QDialog::rejected, this, &PlayersPage::resetHighlight);
    connect(m_markersPopup, &QDialog::accepted, this, &PlayersPage::setMarkerFromPopup);

    // Flash scrollto target reset timer
    m_resetHighlightTimer = new QTimer(this);
    m_resetHighlightTimer->setInterval(320);
    m_resetHighlightTimer->setSingleShot(true);
    connect(m_resetHighlightTimer, &QTimer::timeout, this, &PlayersPage::resetHighlight);

    // Context menu

    m_contextMenu = new TitleMenu(this);

    m_editAction = m_contextMenu->addAction(QString());
    connect(m_editAction, &QAction::triggered,
            [this] { m_playersList->editItem(m_playersList->currentItem()); });

    m_deleteAction = m_contextMenu->addAction(QString());
    connect(m_deleteAction, &QAction::triggered, this, &PlayersPage::deletePlayers);

    m_contextMenu->addSeparator();

    m_setOnlyMarkerAction = m_contextMenu->addAction(QString());
    m_setOnlyMarkerAction->setVisible(false);

    m_setMarkerMenu = m_contextMenu->addMenu(tr("Markieren als"));

    m_removeMarkerAction = m_contextMenu->addAction(tr("Markierung entfernen"));
    connect(m_removeMarkerAction, &QAction::triggered, this, &PlayersPage::removeMarker);

    // Whole list menu

    m_listMenu = new TitleMenu(tr("Ganze Liste"), this);
    m_listMenu->setTitle(tr("Ganze Liste"));

    m_contextMenu->addSeparator();
    m_contextMenu->addMenu(m_listMenu);

    m_listSetOnlyMarkerAction = m_listMenu->addAction(QString());

    m_listSetMarkerMenu = m_listMenu->addMenu(tr("Markieren als"));

    m_listRemoveOnlyMarkerAction = m_listMenu->addAction(QString());

    m_listRemoveMarkerMenu = m_listMenu->addMenu(tr("Markierungen entfernen"));

    m_listRemoveAllMarkersAction = m_listMenu->addAction(tr("Alle Markierungen entfernen"));
    connect(m_listRemoveAllMarkersAction, &QAction::triggered,
            this, &PlayersPage::listRemoveAllMarkers);

    m_listMenu->addSeparator();

    m_deleteOnlyMarkedAction = m_listMenu->addAction(QString());

    m_deleteMarkedMenu = m_listMenu->addMenu(QString());

    m_deleteUnmarkedAction = m_listMenu->addAction(QString());
    connect(m_deleteUnmarkedAction, &QAction::triggered, this, &PlayersPage::deleteUnmarked);

    m_deleteAllAction = m_listMenu->addAction(QString());
    connect(m_deleteAllAction, &QAction::triggered, this, &PlayersPage::deleteAll);
}

void PlayersPage::reload()
{
    m_newPlayersBox->show();

    switch (m_db->tournamentMode()) {
    case Database::TournamentMode::FixedPairs:
        m_newPlayersBox->setTitle(tr("Neue Paare anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Paare"));
        m_nameLabel->setText(tr("Name(n):"));
        m_editAction->setText(tr("Name(n) ändern"));
        m_deleteAction->setText(tr("Paar löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Paare mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Paare löschen"));
        m_deleteAllAction->setText(tr("Alle Paare löschen"));
        break;
    case Database::TournamentMode::SinglePlayers:
        m_newPlayersBox->setTitle(tr("Neue Spieler anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Spieler"));
        m_nameLabel->setText(tr("Name:"));
        m_editAction->setText(tr("Name ändern"));
        m_deleteAction->setText(tr("Spieler löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Spieler mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Spieler löschen"));
        m_deleteAllAction->setText(tr("Alle Spieler löschen"));
        break;
    }

    m_newPlayersBox->setEnabled(m_db->isWritable());
    if (m_db->isWritable()) {
        m_playersList->setEditTriggers(QAbstractItemView::DoubleClicked);
    } else {
        m_playersList->setEditTriggers(QAbstractItemView::NoEditTriggers);
    }

    m_markersWidget->reload(); // This will also trigger updatePlayersList()
}

void PlayersPage::markersChanged()
{
    const QList<Marker::Marker> markers = m_markersWidget->getMarkers();
    const int count = markers.count();

    // No need to process count < 2, because the actions will be hidden in this case

    if (count == 2) {
        // We have exactly one marker

        m_setOnlyMarkerAction->disconnect();
        m_listSetOnlyMarkerAction->disconnect();
        m_listRemoveOnlyMarkerAction->disconnect();
        m_deleteOnlyMarkedAction->disconnect();

        for (const Marker::Marker &marker : markers) {
            if (marker.id != 0) {
                m_setOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_setOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&PlayersPage::setMarker, this, marker.name, marker.id));

                m_listSetOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_listSetOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&PlayersPage::listSetMarker, this, marker.name, marker.id));

                m_listRemoveOnlyMarkerAction->setText(tr("Markierung „%1“ entfernen").arg(
                                                          marker.name));
                connect(m_listRemoveOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&PlayersPage::listRemoveMarker,
                                        this, marker.name, marker.id));

                m_deleteOnlyMarkedAction->setText((m_db->isPairMode()
                    ? tr("Alle als „%1“ markierten Paare löschen")
                    : tr("Alle als „%1“ markierten Spieler löschen")).arg(marker.name));
                m_deleteOnlyMarkedAction->setData(marker.id);
                connect(m_deleteOnlyMarkedAction, &QAction::triggered,
                        this, std::bind(&PlayersPage::deleteMarked, this, marker.name, marker.id));

                break;
            }
        }
    }

    if (count > 2) {
        // We have multiple markers
        m_setMarkerMenu->clear();
        m_listSetMarkerMenu->clear();
        m_listRemoveMarkerMenu->clear();
        m_deleteMarkedMenu->clear();

        for (const Marker::Marker &marker : markers) {
            if (marker.id == 0) {
                continue;
            }

            QAction *setMarkerAction = m_setMarkerMenu->addAction(marker.name);
            setMarkerAction->setData(marker.id);
            connect(setMarkerAction, &QAction::triggered,
                    this, std::bind(&PlayersPage::setMarker, this, marker.name, marker.id));

            QAction *listSetMarkerAction = m_listSetMarkerMenu->addAction(marker.name);
            listSetMarkerAction->setData(marker.id);
            connect(listSetMarkerAction, &QAction::triggered,
                    this, std::bind(&PlayersPage::listSetMarker, this, marker.name, marker.id));

            QAction *listRemoveMarkerAction = m_listRemoveMarkerMenu->addAction(marker.name);
            listRemoveMarkerAction->setData(marker.id);
            connect(listRemoveMarkerAction, &QAction::triggered,
                    this, std::bind(&PlayersPage::listRemoveMarker, this, marker.name, marker.id));

            QAction *deleteMarkedAction = m_deleteMarkedMenu->addAction(marker.name);
            deleteMarkedAction->setData(marker.id);
            connect(deleteMarkedAction, &QAction::triggered,
                    this, std::bind(&PlayersPage::deleteMarked, this, marker.name, marker.id));
        }
    }

    m_setOnlyMarkerAction->setVisible(count == 2);
    m_setMarkerMenu->menuAction()->setVisible(count > 2);
    m_removeMarkerAction->setVisible(count > 1);

    m_listSetOnlyMarkerAction->setVisible(count == 2);
    m_listSetMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveOnlyMarkerAction->setVisible(count == 2);
    m_listRemoveMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveAllMarkersAction->setVisible(count > 2);

    m_deleteOnlyMarkedAction->setVisible(count == 2);
    m_deleteMarkedMenu->menuAction()->setVisible(count > 2);
    m_deleteUnmarkedAction->setVisible(count > 2);

    listModified();

    // Update the markers popup
    m_markersPopup->updateMarkers(markers);
}

void PlayersPage::updatePlayersList()
{
    // Be sure not to produce a segfault via the resetHighlight timer call
    m_highlightedItem = nullptr;
    if (m_resetHighlightTimer->isActive()) {
        m_resetHighlightTimer->stop();
    }

    // Create a list of all blocks we need
    QMap<int, int> targetBlock;
    QList<QStringList> block;
    QMap <int, QColor> markerColor;
    QMap <int, QString> markerName;
    for (const Marker::Marker &marker : m_markersWidget->getMarkers()) {
        switch (marker.sorting) {
        case Marker::Sorting::NewBlock:
            block.append(QStringList());
            targetBlock[marker.id] = block.count() - 1;
            break;
        case Marker::Sorting::Merged:
            targetBlock[marker.id] = block.count() - 1;
            break;
        }
        if (marker.id != 0) {
            markerColor[marker.id] = marker.color;
            markerName[marker.id] = marker.name;
        }
    }

    // Fetch the players list and a list of all marked players
    QStringList allPlayers = m_db->listPlayers();
    m_stringTools->sortList(allPlayers);
    const QMap<QString, int> playersMarks = m_db->playersMarks();

    const int oldPlayersCount = m_playersList->rowCount();
    const int playersCount = allPlayers.size();
    m_playersList->blockSignals(true);
    m_playersList->setRowCount(playersCount);
    m_searchEngine->clearSearchCache();

    // Iterate through all players, add them to their respective blocks and count the markers
    QMap<int, int> markersCount;

    for (const QString &name : allPlayers) {
        m_searchEngine->updateSearchCache(name);
        block[targetBlock[playersMarks.value(name)]].append(name);
        markersCount[playersMarks.value(name)]++;
    }

    // Display all blocks

    int blockRow = 0;
    for (const QStringList &list : block) {
        for (const QString &name : list) {
            QTableWidgetItem *nameCell = new QTableWidgetItem(name);
            QTableWidgetItem *setMarkerCell = new QTableWidgetItem;
            setMarkerCell->setIcon(m_markerIcon);
            setMarkerCell->setFlags(setMarkerCell->flags() & (~Qt::ItemIsEditable));
            QTableWidgetItem *markerCell = new QTableWidgetItem;
            markerCell->setFlags(markerCell->flags() & (~Qt::ItemIsEditable));
            nameCell->setData(PlayersData::Marker, playersMarks.value(name));
            if (playersMarks.value(name) != 0) {
                nameCell->setForeground(markerColor.value(playersMarks.value(name)));
                markerCell->setText(markerName.value(playersMarks.value(name)));
            }
            m_playersList->setItem(blockRow, 0, nameCell);
            m_playersList->setItem(blockRow, 1, setMarkerCell);
            m_playersList->setItem(blockRow, 2, markerCell);
            blockRow++;
        }
    }
    updateVisibility();

    // Update the context menus

    if (m_listSetMarkerMenu->menuAction()->isVisible()) {
        for (QAction *action : m_listSetMarkerMenu->actions()) {
            action->setEnabled(markersCount[action->data().toInt()] < playersCount);
        }
        for (QAction *action : m_listRemoveMarkerMenu->actions()) {
            action->setEnabled(markersCount[action->data().toInt()] > 0);
        }
        for (QAction *action : m_deleteMarkedMenu->actions()) {
            action->setEnabled(markersCount[action->data().toInt()] > 0);
        }
    }

    if (markersCount.value(0) == playersCount) {
        m_listRemoveOnlyMarkerAction->setEnabled(false);
        m_listRemoveMarkerMenu->setEnabled(false);
        m_listRemoveAllMarkersAction->setEnabled(false);
    } else {
        m_listRemoveOnlyMarkerAction->setEnabled(true);
        m_listRemoveMarkerMenu->setEnabled(true);
        m_listRemoveAllMarkersAction->setEnabled(true);
    }

    if (m_db->tournamentStarted() || markersCount.value(0) == playersCount) {
        m_deleteOnlyMarkedAction->setEnabled(false);
        m_deleteMarkedMenu->setEnabled(false);
        m_deleteUnmarkedAction->setEnabled(false);
    } else {
        m_deleteOnlyMarkedAction->setEnabled(
            markersCount[m_deleteOnlyMarkedAction->data().toInt()] < playersCount);
        m_deleteMarkedMenu->setEnabled(true);
        m_deleteUnmarkedAction->setEnabled(markersCount[0] > 0);
    }

    // Update the GUI

    m_markersWidget->updateCounters(markersCount);

    m_playersList->resizeColumnsToContents();
    m_playersList->blockSignals(false);

    m_searchWidget->setVisible(playersCount > 0);
    emit registrationsAvailable(playersCount > 0);

    QString registrationStatus = m_db->isPairMode() ? tr("%n Paar(e) angemeldet", "", playersCount)
                                                    : tr("%n Spieler angemeldet", "", playersCount);
    if (! m_db->playersDividable(playersCount) && playersCount > 0) {
        registrationStatus.append(QStringLiteral(" – <span style=\"color:red;\">%1</span>").arg(
                                                 tr("nicht durch %1 teilbar!").arg(
                                                    QString::number(m_db->necessaryDivisor()))));
    }
    m_statusLabel->setText(registrationStatus);

    if (playersCount != oldPlayersCount) {
        emit playersCountChanged(playersCount);
    }

    if (m_db->tournamentStarted()) {
        m_newPlayersBox->hide();
    }

    scoreChanged();
}

void PlayersPage::scoreChanged()
{
    m_deleteAction->setEnabled(! m_db->tournamentStarted());
    m_deleteAllAction->setEnabled(! m_db->tournamentStarted());
    m_markersWidget->setAutomaticMarkingBoxVisible(! m_db->tournamentStarted());
}

void PlayersPage::reset()
{
    m_playersList->clearContents();
    m_newPlayersBox->show();
}

void PlayersPage::registerPlayers()
{
    if (! m_registerButton->isEnabled()) {
        return;
    }

    const QString name = m_playersName->text().simplified();

    // Check for an exact match. In this case, we can't register the pair/player
    const QString lowerName = m_stringTools->toLower(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (lowerName == m_stringTools->toLower(m_playersList->item(row, 0)->text())) {
            QMessageBox::critical(this,
                m_newPlayersBox->title(),
                tr("<p><b>%1 „%2“ kann nicht angemeldet werden</b></p>"
                   "<p>Der Name ist bereits vergeben.</p>").arg(
                   m_db->isPairMode() ? tr("Das Paar") : tr("Der Spieler"), name));
            m_playersName->selectAll();
            m_playersName->setFocus();
            return;
        }
    }

    // Search for a fuzzy match
    m_searchEngine->setSearchTerm(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (m_searchEngine->checkMatch(m_playersList->item(row, 0)->text())) {
            if (QMessageBox::warning(this,
                m_newPlayersBox->title(),
                tr("<p>Es ist bereits ein %1 mit einem ähnlichen Namen angemeldet:</p>"
                   "<p><table>"
                   "<tr><td>Eingegeben:&nbsp;</td><td><i>%2</i></td></tr>"
                   "<tr><td>Vorhanden:</td><td><i>%3</i></td></tr>"
                   "</table></p>"
                   "<p>Das könnte zu Verwechslungen führen. Soll %4 trotzdem angemeldet "
                   "werden?</p>").arg(m_db->isPairMode() ? tr("Paar") : tr("Spieler"),
                                      name,
                                      m_playersList->item(row, 0)->text(),
                                      m_db->isPairMode() ? tr("das Paar") : tr("der Spieler")),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No) {

                m_playersName->selectAll();
                m_playersName->setFocus();
                return;
            }
            break;
        }
    }

    if (m_isClient || m_isServer) {
        if (m_playersList->findItems(name, Qt::MatchExactly).count() != 0) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("Es wurde mittlerweile von einem anderen Netzwerkteilnehmer ein %1 mit dem "
                   "Namen „%2“ angemeldet. Die Anmeldung ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name));
            m_playersName->clear();
            return;
        }
    }

    if (m_isClient) {
        m_playersName->clear();
        m_playersList->setCurrentItem(nullptr);
        m_client->requestRegisterPlayers(name, m_markersWidget->defaultMarker(name));
        return;
    }

    emit statusUpdate(m_db->isPairMode() ? tr("Melde Paar an …") : tr("Melde Spieler an …"));
    QApplication::setOverrideCursor(Qt::WaitCursor);

    m_db->registerPlayers(name, m_markersWidget->defaultMarker(name));
    m_playersName->clear();
    m_playersName->setFocus();
    updatePlayersList();
    m_searchWidget->clear();

    emit statusUpdate((m_db->isPairMode() ? tr("Paar „%1“ angemeldet")
                                          : tr("Spieler „%1“ angemeldet")).arg(name));

    scrollTo(name);

    if (m_isServer) {
        m_server->broadcastRegisterPlayers(name, m_markersWidget->defaultMarker(name));
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::scrollTo(const QString &text)
{
    QTableWidgetItem *target = m_playersList->findItems(text, Qt::MatchExactly).at(0);
    if (target->row() == m_playersList->rowCount() - 1) {
        m_playersList->scrollToBottom();
    } else {
        m_playersList->scrollToItem(target, QAbstractItemView::PositionAtCenter);
    }

    highlightItem(target);
    m_resetHighlightTimer->start();
}

void PlayersPage::editEntered(const QString &originalName, const QString &name)
{
    // Check for an exact match. In this case, we can't rename the pair/player
    const QString lowerName = m_stringTools->toLower(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (row == m_editedRow) {
            continue;
        }
        if (lowerName == m_stringTools->toLower(m_playersList->item(row, 0)->text())) {
            QMessageBox::critical(this,
                m_newPlayersBox->title(),
                tr("<p><b>%1 „%2“ kann nicht umbenannt werden:</b></p>"
                   "<p>Der Name „%3“ ist bereits vergeben.</p>").arg(
                   m_db->isPairMode() ? tr("Das Paar") : tr("Der Spieler"), originalName, name));
            return;
        }
    }

    // Seach for a fuzzy match
    m_searchEngine->setSearchTerm(name);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (row == m_editedRow) {
            continue;
        }
        if (m_searchEngine->checkMatch(m_playersList->item(row, 0)->text())) {
            if (QMessageBox::warning(this,
                m_newPlayersBox->title(),
                tr("<p>Es ist bereits ein %1 mit einem ähnlichen Namen angemeldet:</p>"
                   "<p><table>"
                   "<tr><td>Eingegeben:&nbsp;</td><td><i>%2</i></td></tr>"
                   "<tr><td>Vorhanden:</td><td><i>%3</i></td></tr>"
                   "</table></p>"
                   "<p>Das könnte zu Verwechslungen führen. Soll %4 trotzdem umbenannt "
                   "werden?</p>").arg(m_db->isPairMode() ? tr("Paar") : tr("Spieler"),
                                      name,
                                      m_playersList->item(row, 0)->text(),
                                      m_db->isPairMode() ? tr("das Paar") : tr("Der Spieler")),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No) {

                return;
            }
            break;
        }
    }

    if (m_isClient || m_isServer) {
        if (m_playersList->findItems(originalName, Qt::MatchExactly).count() != 1) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("%1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer entweder "
                   "umbenannt oder gelöscht. Das Umbenennen ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_db->isPairMode() ? tr("Das Paar") : tr("Der Spieler"), originalName));
            return;
        }
        if (m_playersList->findItems(name, Qt::MatchExactly).count() != 0) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("Es wurde mittlerweile von einem anderen Netzwerkteilnehmer %1 mit dem "
                   "Namen „%2“ angemeldet oder %1 entsprechend umbenannt. Das Umbenennen ist "
                   "nicht mehr möglich und wird abgebrochen.").arg(
                   m_db->isPairMode() ? tr("ein Paar") : tr("ein Spieler"), name));
            return;
        }
    }

    if (m_isClient) {
        m_client->requestRenamePlayers(originalName, name);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(m_db->isPairMode() ? tr("Benenne Paar um …") : tr("Benenne Spieler um …"));
    m_db->renamePlayers(originalName, name);
    emit statusUpdate((m_db->isPairMode() ? tr("Paar „%1“ umbenannt zu „%2“")
                                          : tr("Spieler „%1“ umbenannt zu „%2“")).arg(
                                               originalName, name));
    emit nameEdited();
    listModified();
    if (m_playersList->findItems(name, Qt::MatchExactly).at(0)->row() != m_editedRow) {
        scrollTo(name);
    }

    if (m_isServer) {
        m_server->broadcastRenamePlayers(originalName, name);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::showContextMenu(QPoint point)
{
    if (m_playersList->rowCount() == 0 || ! m_db->isWritable()) {
        return;
    }

    QTableWidgetItem *selectedItem = m_playersList->itemAt(point);

    if (m_playersList->horizontalHeader()->isVisible()) {
        point.setY(point.y() + m_playersList->horizontalHeader()->height());
    }

    if (selectedItem == nullptr) {
        m_listMenu->exec(m_playersList->mapToGlobal(point));
        return;
    }

    const int itemColumn = selectedItem->column();
    if (itemColumn == 1) {
        showMarkerPopup(selectedItem->row(), itemColumn);
        return;
    }
    if (itemColumn != 0) {
        selectedItem = m_playersList->item(selectedItem->row(), 0);
    }

    m_selectedName = selectedItem->text();
    const int selectedMark = selectedItem->data(PlayersData::Marker).toInt();
    m_contextMenu->setTitle((m_db->isPairMode() ? tr("Paar „%1“") : tr("Spieler „%1“")).arg(
                             m_selectedName));
    m_removeMarkerAction->setEnabled(selectedMark != 0);
    for (QAction *action : m_setMarkerMenu->actions()) {
        action->setEnabled(action->data().toInt() != selectedMark);
    }

    m_contextMenu->exec(m_playersList->mapToGlobal(point));
}

void PlayersPage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_playersList->rowCount(); row++) {
            m_playersList->setRowHidden(row, ! isRowVisible(row));
        }
        m_playersList->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    int matches = 0;
    m_searchEngine->setSearchTerm(searchText);
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        if (m_searchEngine->checkMatch(m_playersList->item(row, 0)->text())) {
            if (isRowVisible(row)) {
                m_playersList->showRow(row);
            }
            matches++;
        } else {
            m_playersList->hideRow(row);
        }
    }
    m_searchWidget->updateMatches(matches);
    m_playersList->resizeColumnsToContents();

    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText));
    m_isSearchLabel->show();
}

void PlayersPage::listModified()
{
    m_searchWidget->saveSearch();
    updatePlayersList();
    m_searchWidget->restoreSearch();
}

void PlayersPage::resetSearch()
{
    m_searchWidget->clear();
}

void PlayersPage::displayMarkedEntry()
{
    const int lastRow = m_playersList->currentItem()->row();
    listModified();
    if (m_playersList->findItems(m_selectedName, Qt::MatchExactly).at(0)->row() != lastRow) {
        scrollTo(m_selectedName);
    }
}

void PlayersPage::setMarker(const QString &markerName, const int markerId)
{
    if (m_isClient) {
        m_client->requestSetMarker(m_selectedName, markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_db->isPairMode() ? tr("Markiere Paar „%1“ als „%2“ …")
                                          : tr("Markiere Spieler „%1“ als „%2“ …")).arg(
                                               m_selectedName, markerName));
    m_db->setMarker(m_selectedName, markerId);
    emit statusUpdate((m_db->isPairMode() ? tr("Paar „%1“ als „%2“ markiert")
                                          : tr("Spieler „%1“ als „%2“ markiert")).arg(
                                               m_selectedName, markerName));
    displayMarkedEntry();

    if (m_isServer) {
        m_server->broadcastSetMarker(m_selectedName, markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::removeMarker()
{
    if (m_isClient) {
        m_client->requestRemoveMarker(m_selectedName);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_db->isPairMode() ? tr("Entferne Markierung von Paar „%2“ …")
                                          : tr("Entferne Markierung von Spieler „%2“ …")).arg(
                                               m_selectedName));
    m_db->removeMarker(m_selectedName);
    emit statusUpdate((m_db->isPairMode() ? tr("Markierung von Paar „%2“ entfernt")
                                          : tr("Markierung von Spieler „%2“ entfernt")).arg(
                                               m_selectedName));
    displayMarkedEntry();

    if (m_isServer) {
        m_server->broadcastRemoveMarker(m_selectedName);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::listSetMarker(const QString &markerName, const int markerId)
{
    if (m_isClient) {
        m_client->requestListSetMarker(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_db->isPairMode() ? tr("Markiere die Paarliste als „%2“")
                                          : tr("Markiere die Spielerliste als „%2“")).arg(
                                               markerName));
    m_db->listSetMarker(markerId);
    emit statusUpdate((m_db->isPairMode() ? tr("Paarliste als „%2“ markiert")
                                          : tr("Spielerliste als „%2“ markiert")).arg(
                                               markerName));
    listModified();

    if (m_isServer) {
        m_server->broadcastListSetMarker(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::listRemoveMarker(const QString &markerName, const int markerId)
{
    if (m_isClient) {
        m_client->requestListRemoveMarker(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Entferne alle „%1“-Markierungen …").arg(markerName));
    m_db->listRemoveMarker(markerId);
    emit statusUpdate(tr("Alle „%1“-Markierungen entfernt").arg(markerName));
    listModified();

    if (m_isServer) {
        m_server->broadcastListRemoveMarker(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::listRemoveAllMarkers()
{
    if (m_isClient) {
        m_client->requestListRemoveAllMarkers();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Entferne alle Markierungen …"));
    m_db->listRemoveAllMarkers();
    emit statusUpdate(tr("Alle Markierungen entfernt"));
    listModified();

    if (m_isServer) {
        m_server->broadcastListRemoveAllMarkers();
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::deletePlayers()
{
    if (QMessageBox::question(this,
            m_db->isPairMode() ? tr("Paar löschen")
                               : tr("Spieler löschen"),
            (m_db->isPairMode() ? tr("Soll Paar „%1“ wirklich gelöscht werden?")
                                : tr("Soll Spieler „%1“ wirklich gelöscht werden?")).arg(
                                     m_selectedName),
            QMessageBox::Yes | QMessageBox::Cancel,
            QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    if (m_isClient || m_isServer) {
        if (m_playersList->findItems(m_selectedName, Qt::MatchExactly).count() != 1) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("%1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer entweder "
                   "gelöscht oder umbenannt. Das Löschen ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_db->isPairMode() ? tr("Das Paar") : tr("Der Spieler"), m_selectedName));
            return;
        }
    }

    if (m_isClient) {
        m_client->requestDeletePlayers(m_selectedName);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate((m_db->isPairMode() ? tr("Lösche Paar „%1“ …")
                                          : tr("Lösche Spieler „%1“ …")).arg(m_selectedName));
    m_db->deletePlayers(m_selectedName);
    emit statusUpdate((m_db->isPairMode() ? tr("Paar „%1“ gelöscht")
                                          : tr("Spieler „%1“ gelöscht")).arg(m_selectedName));

    listModified();

    if (m_isServer) {
        m_server->broadcastDeletePlayers(m_selectedName);
    }

    QApplication::restoreOverrideCursor();
}

void PlayersPage::cacheChecksum()
{
    if (m_isServer || m_isClient) {
        m_cachedChecksum = checksum();
    }
}

bool PlayersPage::checkCachedChecksum(const QString &title)
{
    if (! m_isClient && ! m_isServer) {
        return true;
    }

    if (checksum() != m_cachedChecksum) {
        QMessageBox::warning(this, title,
            tr("<p><b>Kollidierende Netzwerkänderung</b></p>"
               "<p>Die Anmeldungsliste wurde zwischenzeitlich von einem anderen Netzwerkteilnehmer "
               "verändert. Die Aktion kann nicht mehr durchgeführt werden und wird abgebrochen."
               "</p>"));
        return false;
    }

    return true;
}

bool PlayersPage::confirmListDeleteAction(const QString &title, const QString &text)
{
    m_backupFile.clear();

    QMessageBox messageBox(QMessageBox::Warning, title, text,
                           QMessageBox::Yes | QMessageBox::Cancel, this);
    QCheckBox *createBackup = new QCheckBox(tr("Vorher ein Backup der Datenbank anlegen"));
    createBackup->setChecked(true);
    messageBox.setCheckBox(createBackup);

    if (messageBox.exec() == QMessageBox::Cancel) {
        return false;
    }

    if (createBackup->isChecked()) {
        const QString openedDb = m_db->dbFile();
        BackupEngine backupEngine(openedDb);
        m_db->closeDatabaseConnection();
        bool success = backupEngine.createBackup();
        m_db->openDatabase(openedDb);

        if (! success) {
            QMessageBox::critical(this, title,
                tr("<p><b>Das Anlegen eines Backups ist fehlgeschlagen!</b></p>"
                   "<p>Die Datei konnte nicht angelegt werden. Bitte die Zugriffsrechte von „%1“ "
                   "bzw. dem Verzeichnis „%2“) überprüfen!</p>").arg(
                   backupEngine.backupFile(), QDir::toNativeSeparators(backupEngine.path())));
            return false;
        }

        m_backupFile = backupEngine.backupFile();
    }

    return true;
}

void PlayersPage::checkDisplayBackup(const QString &title)
{
    if (! m_backupFile.isEmpty()) {
        QMessageBox::information(this, title,
            tr("Es wurde erfolgreich ein Backup der Datenbank mit dem Dateinamen „%1“ "
               "erstellt.").arg(m_backupFile));
    }
}

void PlayersPage::deleteMarked(const QString &markerName, int markerId)
{
    cacheChecksum();
    const QString title = tr("Markierte %1 löschen").arg(
                             m_db->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
            tr("Sollen wirklich alle als „%1“ markierten %2 gelöscht werden?").arg(markerName,
               m_db->isPairMode() ? tr("Paare") : tr("Spieler")))
        || ! checkCachedChecksum(title)) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteMarkedPlayers(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle als „%1“ markierten %2 …").arg(
                         markerName, m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_db->deleteMarkedPlayers(markerId);
    emit statusUpdate(tr("Alle als „%1“ markierten %2 gelöscht").arg(
                         markerName, m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    listModified();

    if (m_isServer) {
        m_server->broadcastDeleteMarkedPlayers(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void PlayersPage::deleteUnmarked()
{
    cacheChecksum();
    const QString title = tr("Unmarkierte %1 löschen").arg(
                             m_db->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
            tr("Sollen wirklich alle unmarkierten %1 gelöscht werden?").arg(
               m_db->isPairMode() ? tr("Paare") : tr("Spieler")))
        || ! checkCachedChecksum(title)) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteUnmarkedPlayers();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle unmarkierten %1 …").arg(
                         m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_db->deleteUnmarkedPlayers();
    emit statusUpdate(tr("Alle unmarkierten %1 gelöscht").arg(
                         m_db->isPairMode() ? tr("Paare") : tr("Spieler")));

    listModified();

    if (m_isServer) {
        m_server->broadcastDeleteUnmarkedPlayers();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void PlayersPage::deleteAll()
{
    const QString title = tr("Alle %1 löschen").arg(
                             m_db->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
              tr("Sollen wirklich alle %1 gelöscht werden?").arg(
                 m_db->isPairMode() ? tr("Paare") : tr("Spieler")))) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteAllPlayers();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    emit statusUpdate(tr("Lösche alle %1 …").arg(
                         m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    m_db->deleteAllPlayers();
    emit statusUpdate(tr("Alle %1 gelöscht").arg(
                         m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    listModified();

    if (m_isServer) {
        m_server->broadcastDeleteAllPlayers();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void PlayersPage::resetMarkersWidget()
{
    m_markersWidget->reset();
}

void PlayersPage::markersVisibilityChanged()
{
    m_searchWidget->saveSearch();
    updateVisibility();
    m_searchWidget->restoreSearch();
}

void PlayersPage::updateVisibility()
{
    int hidden = 0;
    for (int row = 0; row < m_playersList->rowCount(); row++) {
        bool visible = isRowVisible(row);
        m_playersList->setRowHidden(row, ! visible);
        if (! visible) {
            hidden++;
        }
    }
    if (hidden > 0) {
        m_playersBox->setTitle(m_db->isPairMode()
                                   ? tr("Angemeldete Paare (%n ausgeblendet)", "", hidden)
                                   : tr("Angemeldete Spieler (%n ausgeblendet)", "", hidden));
    } else {
        m_playersBox->setTitle(m_db->isPairMode() ? tr("Angemeldete Paare")
                                                  : tr("Angemeldete Spieler"));
    }
    m_playersList->resizeColumnsToContents();
}

bool PlayersPage::isRowVisible(int row) const
{
    return m_markersWidget->isMarkerVisible(
               m_playersList->item(row, 0)->data(PlayersData::Marker).toInt());
}

void PlayersPage::showSetMarkerColumn(const int state)
{
    m_playersList->setColumnHidden(1, state == Qt::Unchecked);
    m_playersList->resizeColumnsToContents();
}

void PlayersPage::showMarkerColumn(const int state)
{
    m_playersList->horizontalHeader()->setHidden(state == Qt::Unchecked);
    m_playersList->setColumnHidden(2, state == Qt::Unchecked);
    m_playersList->resizeColumnsToContents();
}

QString PlayersPage::checksum() const
{
    QCryptographicHash hash(QCryptographicHash::Md5);

    for (int row = 0; row < m_playersList->rowCount(); row++) {
        hash.addData(m_playersList->item(row, 0)->text().toUtf8());
        hash.addData(QByteArray::number(
            m_playersList->item(row, 0)->data(PlayersData::Marker).toInt()));
    }

    m_markersWidget->addChecksumData(hash);

    return QLatin1String(hash.result().toHex());
}

void PlayersPage::setServer(Server *server)
{
    m_server = server;
    m_isServer = true;
    m_markersWidget->setServer(server);
    connect(server, &Server::nameEdited, this, &PlayersPage::nameEdited);
}

void PlayersPage::serverStopped()
{
    m_isServer = false;
    m_markersWidget->serverStopped();
    markersChanged();
}

void PlayersPage::hideContextMenus()
{
    for (QMenu *menu : findChildren<QMenu *>()) {
        menu->hide();
    }
}

void PlayersPage::processNetworkChange(const int clientId)
{
    const bool contextMenuVisible = m_contextMenu->isVisible();
    const bool listMenuVisible = m_listMenu->isVisible();
    const bool editorOpen = m_delegate->isEditing();
    const bool markersPopupVisible = m_markersPopup->isVisible();

    int processedRow = -1;
    int verticalOffset = 0;
    int originalY = 0;

    QTableWidgetItem *currentItem = m_playersList->currentItem();
    if (currentItem != nullptr) {
        processedRow = currentItem->row();
        originalY = m_playersList->mapToGlobal(
                        m_playersList->visualItemRect(currentItem).topLeft()).y();
    }

    // Check if we're doing anything. If not, we can simply display the change (possibly our own)

    if (! contextMenuVisible && ! listMenuVisible && ! editorOpen && ! markersPopupVisible) {
        m_searchWidget->saveSearch();
        updatePlayersList();

        if (m_isClient && clientId == m_client->id() && ! m_scrollToTarget.isEmpty()) {
            if (m_playersList->findItems(m_scrollToTarget, Qt::MatchExactly).at(0)->row()
                != processedRow) {

                // The last selected row doesn't match the row of the scrollTarget, so we want to
                // be sure that the item is shown and scroll to it
                scrollTo(m_scrollToTarget);
            }

            // Otherwise, we can assume that the item is inside the viewport anyway.
        }

        m_searchWidget->restoreSearch();
        return;
    }

    // Some action is currently executed, so we want to restore it

    // Cache what we have

    QString processedText;
    if (contextMenuVisible || markersPopupVisible) {
        processedText = m_selectedName;
    } else if (editorOpen) {
        processedText = m_delegate->editedText();
    }

    QPoint listMenuPos;
    if (listMenuVisible) {
        // Process this first so that the list menu will be restored if it's visible,
        // even if it was shown by using the context menu's list menu action
        listMenuPos = m_listMenu->pos();
    }

    QPoint contextMenuPos;
    if (contextMenuVisible) {
        contextMenuPos = m_playersList->mapFromGlobal(m_contextMenu->pos());
    }

    QPoint markersPopupPos;
    if (markersPopupVisible) {
        markersPopupPos = m_markersPopup->pos();
    }

    // Abort all actions
    if (contextMenuVisible || listMenuVisible) {
        hideContextMenus();
    }
    if (editorOpen) {
        m_delegate->saveEditorState();
    }
    if (m_markersPopup->isVisible()) {
        m_markersPopup->reject();
    }

    // Display the change
    listModified();

    // Restore the "whole list" menu (we don't need a specific item to still exist for this one)
    if (listMenuVisible) {
        if (m_playersList->rowCount() > 0) {
            QTimer::singleShot(0, this, [this, listMenuPos] { m_listMenu->exec(listMenuPos); } );
        } else {
            QTimer::singleShot(0, this, [this] {
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("Es wurden von einem anderen Netzwerkteilnehmer alle Anmeldungen gelöscht. "
                       "Ein Bearbeiten der ganzen Liste ist nicht mehr möglich.")); } );
        }
        return;
    }

    // From here on, we need the processed item, so check if we still have it

    QTableWidgetItem *processedItem;
    const QList<QTableWidgetItem *> matches = m_playersList->findItems(processedText,
                                                                       Qt::MatchExactly);
    if (matches.isEmpty()) {
        QTimer::singleShot(0, this, [this, processedText] {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("<p>%1 „%2“ wurde durch eine Änderung eines anderen Netzwerkteilnehmers "
                   "entweder umbenannt oder gelöscht.</p>"
                   "<p>Das Bearbeiten wird abgebrochen.</p>").arg(
                   m_db->isPairMode() ? tr("Der gerade bearbeitete Spieler")
                                      : tr("Das gerade bearbeitete Paar"),
                   processedText)); } );
        return;
    } else {
        processedItem = matches.at(0);
        m_playersList->setCurrentItem(processedItem);
    }

    // Try to scroll so that it's at the same place on screen
    if (processedItem->row() != processedRow) {
        m_playersList->verticalScrollBar()->setValue(
            m_playersList->verticalScrollBar()->value() - processedRow + processedItem->row());
        verticalOffset = m_playersList->mapToGlobal(
                             m_playersList->visualItemRect(processedItem).topLeft()).y()
                         - originalY;
    }

    // Restore the context menu

    if (contextMenuVisible) {
        contextMenuPos.setY(contextMenuPos.y() + verticalOffset);

        QTableWidgetItem *iteamAtContextMenuPos = m_playersList->itemAt(contextMenuPos);
        if (iteamAtContextMenuPos == nullptr
            || m_playersList->item(iteamAtContextMenuPos->row(), 0) != processedItem) {

            // This can happen if the context menu has been displayed from bottom to top and now,
            // it can be displayed from top to bottom or vice versa. In this case, display the
            // context menu at the vertical middle of the processed item
            contextMenuPos.setY(m_playersList->visualItemRect(processedItem).topLeft().y()
                                + m_playersList->rowHeight(processedItem->row()) / 2);
        }

        QTimer::singleShot(0, this, [this, contextMenuPos] { showContextMenu(contextMenuPos); } );
        return;
    }

    // Restore the editor
    if (editorOpen) {
        QTimer::singleShot(0, this, [this, processedItem] {
            m_playersList->editItem(processedItem);
            m_delegate->restoreEditorState();
        } );
        return;
    }

    // Restore the markers popup
    if (markersPopupVisible) {
        markersPopupPos.setY(markersPopupPos.y() + verticalOffset);
        QTimer::singleShot(0, this, [this, processedItem, markersPopupPos] {
            highlightItem(processedItem);
            m_markersPopup->show(processedItem->data(PlayersData::Marker).toInt(), markersPopupPos);
        } );
        return;
    }
}

void PlayersPage::setClient(Client *client)
{
    m_isClient = true;
    m_client = client;
    m_markersWidget->setClient(client);
    connect(client, &Client::nameEdited, this, &PlayersPage::nameEdited);
}

void PlayersPage::clientDisconnected()
{
    m_isClient = false;
    m_markersWidget->clientDisconnected();
    markersChanged();
}

void PlayersPage::setProcessingRequest(bool processing, const QString &scrollToTarget)
{
    m_newPlayersBox->setEnabled(! processing);
    m_playersBox->setEnabled(! processing);
    if (processing) {
        m_scrollToTarget = scrollToTarget;
        QApplication::setOverrideCursor(Qt::WaitCursor);
    }
}

MarkersWidget * PlayersPage::markersWidget() const
{
    return m_markersWidget;
}

void PlayersPage::finishRegistration()
{
    if (QMessageBox::warning(this,
        tr("Anmeldung beenden"),
        tr("<p>Wenn die Anmeldung beendet ist, kann (mindestens während das Netzwerk läuft) kein "
           "Netzwerkteilnehmer mehr neue %1 anmelden oder löschen!</p>"
           "<p>Soll die Anmeldung wirklich beendet werden?</p>").arg(
           m_db->isPairMode() ? tr("Paare") : tr("Spieler")),
        QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Cancel) {

        return;
    }

    if (m_isClient) {
        m_client->requestFinishRegistration();
        return;
    }

    // Otherwise, we're the server
    networkFinishRegistration();
    m_server->broadcastFinishRegistration();
}

void PlayersPage::networkFinishRegistration(const int clientId)
{
    QString networkChange;
    if ((m_isClient && clientId != m_client->id()) || (m_isServer && clientId != 0)) {
        networkChange = tr(" von einem anderen Netzwerkteilnehmer");
        if (m_contextMenu->isVisible()) {
            hideContextMenus();
        }
        if (m_delegate->isEditing()) {
            m_delegate->abortEdit();
        }
    }

    m_db->setNetworkTournamentStarted(true);
    reload();
    emit networkTournamentStarted(true);

    QTimer::singleShot(0, this, [this, networkChange] {
        QMessageBox::information(this,
            tr("Anmeldung beendet"),
            tr("<p>Die Anmeldung wurde%1 beendet. Es können ab jetzt keine neuen %2 mehr "
               "angemeldet oder bestehende Anmeldungen gelöscht werden.</p>"
               "<p>Sollte das trotzdem noch nötig sein, dann bitte den Server stoppen, neu "
               "starten und alle Clients neu verbinden. Sofern noch keine Ergebnisse eingegeben "
               "wurden, ist danach das komplette Bearbeiten der Anmeldungsliste wieder möglich."
               "</p>").arg(networkChange, m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
    } );
}

void PlayersPage::showMarkerPopup(const int row, const int column)
{
    if (column != 1) {
        return;
    }

    QTableWidgetItem *item = m_playersList->item(row, 0);
    m_selectedName = item->text();
    highlightItem(item);
    m_markersPopup->show(item->data(PlayersData::Marker).toInt());
}

void PlayersPage::highlightItem(QTableWidgetItem *item)
{
    m_lastForeground = item->foreground();
    m_lastBackground = item->background();
    item->setForeground(m_playersList->palette().highlightedText());
    item->setBackground(m_playersList->palette().highlight());
    m_highlightedItem = item;
}

void PlayersPage::resetHighlight()
{
    if (m_highlightedItem == nullptr) {
        return;
    }

    m_highlightedItem->setForeground(m_lastForeground);
    m_highlightedItem->setBackground(m_lastBackground);
}

void PlayersPage::setMarkerFromPopup()
{
    if (m_markersPopup->markerId() == 0) {
        removeMarker();
    } else {
        setMarker(m_markersPopup->markerName(), m_markersPopup->markerId());
    }
}

QList<bool> PlayersPage::columns() const
{
    return QList<bool> { ! m_playersList->isColumnHidden(1), ! m_playersList->isColumnHidden(2) };
}

void PlayersPage::restoreColumns(const QList<bool> &visibility)
{
    // We need to call this via a QTimer::singleShot so that
    // it's processed after the constructor is finished
    QTimer::singleShot(0, this, [this, visibility] {
        m_showSetMarkerColumn->setChecked(visibility.at(0));
        m_showMarkerColumn->setChecked(visibility.at(1));
    });
}
