/* Copyright (c) 2012-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef PLAYERSPAGE_H
#define PLAYERSPAGE_H

// Qt includes
#include <QWidget>
#include <QIcon>

// Local classes

class SharedObjects;
class StringTools;
class Database;
class SearchEngine;
class PlayersPageDelegate;
class TitleMenu;
class SearchWidget;
class MarkersWidget;
class MarkersPopup;

class Server;
class Client;

// Qt classes
class QGroupBox;
class QLabel;
class QLineEdit;
class QPushButton;
class QTableWidget;
class QAction;
class QMenu;
class QTableWidgetItem;
class QCheckBox;

class PlayersPage : public QWidget
{
    Q_OBJECT

public:
    enum PlayersData {
        Marker = Qt::UserRole
    };

    explicit PlayersPage(QWidget *parent, SharedObjects *sharedObjects);
    void reload();
    void updatePlayersList();
    void reset();
    void resetSearch();
    void resetMarkersWidget();
    void disableConditionalMarker();
    QList<bool> columns() const;
    void restoreColumns(const QList<bool> &visibility);

    void setServer(Server *server);
    void serverStopped();
    void setClient(Client *client);
    void clientDisconnected();
    QString checksum() const;
    MarkersWidget * markersWidget() const;
    void processNetworkChange(const int clientId = 0);
    void setProcessingRequest(bool processing, const QString &scrollToTarget = QString());
    void networkFinishRegistration(const int clientId = 0);

public slots:
    void scoreChanged();
    void finishRegistration();

signals:
    void statusUpdate(const QString &message);
    void playersCountChanged(int count);
    void registrationsAvailable(bool state);
    void nameEdited();
    void networkTournamentStarted(bool state);

private slots:
    void registerPlayers();
    void editEntered(const QString &originalName, const QString &name);
    void showContextMenu(QPoint point);
    void scrollTo(const QString &text);
    void deletePlayers();
    void updateSearch(const QString &searchText);
    void markersChanged();
    void setMarker(const QString &markerName, const int markerId);
    void removeMarker();
    void listSetMarker(const QString &markerName, const int markerId);
    void listRemoveMarker(const QString &markerName, const int markerId);
    void listRemoveAllMarkers();
    void deleteMarked(const QString &markerName, const int markerId);
    void deleteUnmarked();
    void deleteAll();
    void updateVisibility();
    void markersVisibilityChanged();
    void showSetMarkerColumn(const int state);
    void showMarkerColumn(const int state);
    void showMarkerPopup(const int row, const int column);
    void highlightItem(QTableWidgetItem *item);
    void resetHighlight();
    void setMarkerFromPopup();

private: // Functions
    void listModified();
    bool isRowVisible(int row) const;
    void updateStatusLabel();
    bool confirmListDeleteAction(const QString &title, const QString &text);
    void checkDisplayBackup(const QString &title);
    void cacheChecksum();
    bool checkCachedChecksum(const QString &title);
    void hideContextMenus();
    void displayMarkedEntry();

private: // Variables
    StringTools *m_stringTools;
    Database *m_db;
    SearchEngine *m_searchEngine;

    QGroupBox *m_newPlayersBox;
    QLabel *m_nameLabel;
    QLineEdit *m_playersName;
    QPushButton *m_registerButton;
    QTimer *m_resetHighlightTimer;

    QGroupBox *m_playersBox;
    QTableWidget *m_playersList;
    QIcon m_markerIcon;
    PlayersPageDelegate *m_delegate;
    QString m_selectedName;
    QLabel *m_statusLabel;
    SearchWidget *m_searchWidget;
    QLabel *m_isSearchLabel;
    QTableWidgetItem *m_highlightedItem;
    int m_editedRow;

    QCheckBox *m_showSetMarkerColumn;
    QCheckBox *m_showMarkerColumn;

    MarkersWidget *m_markersWidget;
    MarkersPopup *m_markersPopup;
    QBrush m_lastForeground;
    QBrush m_lastBackground;

    TitleMenu *m_contextMenu;

    QAction *m_editAction;
    QAction *m_deleteAction;

    QAction *m_setOnlyMarkerAction;
    QMenu *m_setMarkerMenu;
    QAction *m_removeMarkerAction;

    TitleMenu *m_listMenu;

    QAction *m_listSetOnlyMarkerAction;
    QMenu *m_listSetMarkerMenu;
    QAction *m_listRemoveOnlyMarkerAction;
    QMenu *m_listRemoveMarkerMenu;
    QAction *m_listRemoveAllMarkersAction;

    QAction *m_deleteOnlyMarkedAction;
    QMenu *m_deleteMarkedMenu;
    QAction *m_deleteUnmarkedAction;
    QAction *m_deleteAllAction;

    Server *m_server;
    Client *m_client;
    bool m_isServer = false;
    bool m_isClient = false;
    QString m_scrollToTarget;
    QString m_cachedChecksum;
    QString m_backupFile;

};

#endif // PLAYERSPAGE_H
