/* Copyright (c) 2012-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "RankingPage.h"

#include "shared/SharedObjects.h"
#include "shared/SearchEngine.h"
#include "shared/Database.h"
#include "shared/SearchWidget.h"
#include "shared/TableDelegate.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QTableWidget>
#include <QHeaderView>
#include <QCheckBox>
#include <QComboBox>

RankingPage::RankingPage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent), m_db(sharedObjects->database()),
                       m_searchEngine(sharedObjects->searchEngine())
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    QGroupBox *settingsBox = new QGroupBox;
    QHBoxLayout *settingsBoxLayout = new QHBoxLayout(settingsBox);
    layout->addWidget(settingsBox);

    m_upToRoundLabel = new QLabel(tr("Rangliste für Runde"));
    m_upToRoundLabel->setAlignment(Qt::AlignRight);
    m_upToRoundLabel->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    settingsBoxLayout->addWidget(m_upToRoundLabel);

    m_upToRound = new QComboBox;
    m_upToRound->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    m_upToRound->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    connect(m_upToRound, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &RankingPage::showRankingUpToRound);
    settingsBoxLayout->addWidget(m_upToRound);

    settingsBoxLayout->addStretch();

    m_includeOpponentGoals = new QCheckBox(tr("Gegnerische Tore berücksichtigen"));
    m_includeOpponentGoals->setWhatsThis(tr(
        "Wenn diese Option ausgewählt ist, dann werden bei der Erstellung der Rangliste die "
        "„geschossenen Tore“ der Gegner in den gewonnenen Bobbln des jeweiligen %1 berücksichtigt "
        "– je weniger, desto besser.").arg(m_db->isPairMode() ? tr("Paars") : tr("Spielers"))
    );
    connect(m_includeOpponentGoals, &QCheckBox::stateChanged,
            this, &RankingPage::includeOpponentGoalsChanged);
    settingsBoxLayout->addWidget(m_includeOpponentGoals);

    m_rankingBox = new QGroupBox;
    QVBoxLayout *rankingBoxLayout = new QVBoxLayout(m_rankingBox);
    layout->addWidget(m_rankingBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->hide();
    rankingBoxLayout->addWidget(m_isSearchLabel);

    m_interimLabel = new QLabel;
    m_interimLabel->setWordWrap(true);
    rankingBoxLayout->addWidget(m_interimLabel);
    m_interimLabel->hide();

    m_ranking = new QTableWidget;
    m_ranking->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_ranking->setItemDelegate(new TableDelegate(this, TableDelegate::HighlightMode::Custom));
    m_ranking->verticalHeader()->hide();
    m_ranking->setShowGrid(false);
    m_ranking->setWordWrap(false);
    m_ranking->setSelectionMode(QTableWidget::NoSelection);
    m_ranking->setFocusPolicy(Qt::NoFocus);
    m_ranking->setColumnCount(5);
    rankingBoxLayout->addWidget(m_ranking);

    m_multipleRanks = new QLabel;
    m_multipleRanks->setWordWrap(true);
    m_multipleRanks->hide();
    rankingBoxLayout->addWidget(m_multipleRanks);

    m_searchWidget = new SearchWidget;
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &RankingPage::updateSearch);
    rankingBoxLayout->addWidget(m_searchWidget);
}

void RankingPage::dataChanged()
{
    m_roundsState = m_db->roundsState();
    m_allRounds = m_roundsState.size();

    m_upToRound->blockSignals(true);
    m_upToRound->clear();
    for (int i = 1; i <= m_allRounds; i++) {
        m_upToRound->addItem(QString::number(i));
    }
    m_upToRound->setCurrentIndex(m_allRounds - 1);
    m_upToRound->blockSignals(false);

    if (m_upToRound->count() == 1) {
        m_upToRound->hide();
        m_upToRoundLabel->hide();
    } else {
        m_upToRound->show();
        m_upToRoundLabel->show();
    }

    reload(m_allRounds);
}

void RankingPage::includeOpponentGoalsChanged(bool state)
{
    m_db->setIncludeOpponentGoals(state);
    reload(m_upToRound->currentIndex() + 1);
}

void RankingPage::showRankingUpToRound(int round)
{
    reload(round + 1);
}

void RankingPage::reload(int upToRound)
{
    m_searchWidget->saveSearch();

    QList<int> incompleteRounds;
    for (int round : m_roundsState.keys()) {
        if (! m_roundsState[round]) {
            incompleteRounds.append(round);
        }
    }

    const int missingRounds = incompleteRounds.count();
    QString interimText;
    if (missingRounds > 0) {
        emit enableExport(false);
        interimText = tr("<table><tr><td><b>Achtung!</b></td><td>Das ist nur ein Zwischenergebnis. "
                         "Es fehlen noch Ergebnisse aus ");
        if (missingRounds == 1) {
            interimText += tr("Runde %1.").arg(QString::number(incompleteRounds[0]));
        } else if (missingRounds == 2) {
            interimText += tr("den Runden %1 und %2.").arg(QString::number(incompleteRounds[0]),
                                                           QString::number(incompleteRounds[1]));
        } else {
            interimText += tr("den Runden ");
            for (int i = 0; i < missingRounds - 1; i++) {
                interimText += tr("%1, ").arg(QString::number(incompleteRounds[i]));
            }
            interimText = interimText.left(interimText.size() - 2);
            interimText += tr(" und %1.").arg(QString::number(incompleteRounds.last()));
        }
        interimText += QStringLiteral(
            "<br/>%1 <span style=\"color:red\">%2</span></td></tr></table>").arg(
            m_db->isPairMode() ? tr("Paare, von denen noch Spielstände fehlen, sind")
                               : tr("Spieler, von denen noch Spielstände fehlen, sind"),
            tr("rot markiert"));
    } else {
        emit enableExport(true);
    }

    if (upToRound < m_allRounds) {
        if (m_roundsState[upToRound]) {
            m_rankingBox->setTitle(tr("Zwischenstand: Rangliste nach Runde %1").arg(upToRound));
        } else {
            m_rankingBox->setTitle(tr("Zwischenstand: Rangliste im Verlauf von Runde %1").arg(
                                      upToRound));
        }
        interimText = tr("<table><tr><td><b>Achtung!</b></td><td>Das ist nicht die aktuelle "
                         "Rangliste! Um diese anzuzeigen <b>bitte Runde %1 auswählen!</b></td>"
                         "</tr></table>").arg(m_allRounds)
                      + interimText;
    } else {
        if (! m_roundsState[m_allRounds]) {
            m_rankingBox->setTitle(tr("Aktuelle Rangliste im Verlauf der %1. Runde").arg(
                                      QString::number(m_allRounds)));
        } else {
            if (m_allRounds == 1) {
                m_rankingBox->setTitle(tr("Aktuelle Rangliste nach einer Runde"));
            } else {
                m_rankingBox->setTitle(tr("Aktuelle Rangliste nach %1 Runden").arg(
                                          QString::number(m_allRounds)));
            }
        }
    }

    if (interimText.isEmpty()) {
        m_interimLabel->hide();
    } else {
        m_interimLabel->setText(interimText);
        m_interimLabel->show();
    }

    m_ranking->setColumnHidden(4, ! m_includeOpponentGoals->isChecked());

    QList<Database::RankData> ranking = m_db->ranking(upToRound);
    int allBoogers = upToRound * m_db->boogersPerRound();

    m_ranking->setHorizontalHeaderLabels( { tr("Platz"),
                                            m_db->isPairMode() ? tr("Paar") : tr("Spieler"),
                                            tr("Bobbl"),
                                            tr("Tore"),
                                            tr("Gegn. Tore") } );
    m_ranking->horizontalHeaderItem(2)->setWhatsThis(tr(
        "In dieser Spalte werden die jeweils gewonnenen Bobbl angezeigt – je mehr, desto besser"));
    m_ranking->horizontalHeaderItem(3)->setWhatsThis(tr(
        "„Geschossene Tore“ sind die durch %1 in den verlorenen Bobbln erreichten Punkte – je "
        "mehr, desto besser.").arg(
        m_db->isPairMode() ? tr("das jeweilige Paar") : tr("den jeweiligen Spieler")));
    m_ranking->horizontalHeaderItem(4)->setWhatsThis(tr(
        "„Gegenerische geschossene Tore“ sind die Punkte, die die Gegener in den gewonnenen "
        "Bobbln des jeweiligen %1 erreicht haben – je weniger, desto besser.").arg(
        m_db->isPairMode() ? tr("Paars") : tr("Spielers")));

    m_ranking->setRowCount(ranking.size());

    int row = 0;
    int lastRank = 0;
    int color = 1;
    bool drawTopLine = false;
    QMap<int, int> multipleRanks;

    for (const Database::RankData &currentRank : ranking) {
        if (currentRank.rank != lastRank) {
            color++;
            color = color % 2;
            if (lastRank != 0) {
                drawTopLine = true;
            }
        } else {
            // This will insert rank keys if they don't exist yet and increment them if they exist
            multipleRanks[currentRank.rank]++;
        }
        lastRank = currentRank.rank;

        QTableWidgetItem *rankCell = new QTableWidgetItem(QString::number(currentRank.rank));
        QTableWidgetItem *nameCell = new QTableWidgetItem(currentRank.name);
        QTableWidgetItem *boogersCell = new QTableWidgetItem(QString::number(currentRank.boogers));
        QTableWidgetItem *goalsCell = new QTableWidgetItem(QString::number(currentRank.goals));
        QTableWidgetItem *opponentGoalsCell = new QTableWidgetItem(
                                                      QString::number(currentRank.opponentGoals));

        if (currentRank.playedBoogers != allBoogers) {
            rankCell->setForeground(Qt::red);
            nameCell->setForeground(Qt::red);
            boogersCell->setForeground(Qt::red);
            goalsCell->setForeground(Qt::red);
            opponentGoalsCell->setForeground(Qt::red);
        }

        if (color) {
            rankCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
            nameCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
            boogersCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
            goalsCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
            opponentGoalsCell->setData(TableDelegate::Display::DrawBackground, QVariant(true));
        }

        if (drawTopLine && row > 0) {
            rankCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            nameCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            boogersCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            goalsCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            opponentGoalsCell->setData(TableDelegate::Display::DrawTopLine, QVariant(true));
            drawTopLine = false;
        }

        m_ranking->setItem(row, 0, rankCell);
        m_ranking->setItem(row, 1, nameCell);
        m_ranking->setItem(row, 2, boogersCell);
        m_ranking->setItem(row, 3, goalsCell);
        m_ranking->setItem(row, 4, opponentGoalsCell);

        row++;
    }

    if (multipleRanks.count() > 0) {
        QString multipleRanksText = tr("Mehrfach besetzte Ranglistenplätze: ");
        for (int rank : multipleRanks.keys()) {
            multipleRanksText += tr("<b>%1</b> (%2x), ").arg(
                                    QString::number(rank),
                                    QString::number(multipleRanks[rank] + 1));
        }
        multipleRanksText.chop(2);
        m_multipleRanks->setText(multipleRanksText);
        m_multipleRanks->show();
    } else {
        m_multipleRanks->hide();
    }

    m_ranking->resizeColumnsToContents();
    m_searchWidget->restoreSearch();
}

void RankingPage::updateSearch(const QString &searchText)
{
    if (searchText.isEmpty()) {
        for (int row = 0; row < m_ranking->rowCount(); row++) {
            if (m_ranking->isRowHidden(row)) {
                m_ranking->showRow(row);
            }
        }
        m_ranking->resizeColumnsToContents();
        m_isSearchLabel->hide();
        return;
    }

    int matches = 0;
    m_searchEngine->setSearchTerm(searchText);
    for (int row = 0; row < m_ranking->rowCount(); row++) {
        if (m_searchEngine->checkMatch(m_ranking->item(row, 1)->text())) {
            if (m_ranking->isRowHidden(row)) {
                m_ranking->showRow(row);
            }
            matches++;
        } else {
            if (! m_ranking->isRowHidden(row)) {
                m_ranking->hideRow(row);
            }
        }
    }
    m_searchWidget->updateMatches(matches);
    m_ranking->resizeColumnsToContents();

    if (! m_isSearchLabel->isVisible()) {
        m_isSearchLabel->show();
    }
    m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(searchText));
}

void RankingPage::setIncludeOpponentGoals(bool state)
{
    m_includeOpponentGoals->setChecked(state);
}

void RankingPage::resetSearch()
{
    m_searchWidget->clear();
}
