/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "PlayersComboBox.h"
#include "PlayersLineEdit.h"
#include "PlayersListView.h"

#include "shared/SharedStyles.h"
#include "shared/SearchEngine.h"

// Qt includes
#include <QDebug>
#include <QStringListModel>
#include <QCompleter>

PlayersComboBox::PlayersComboBox(QWidget *parent, SearchEngine *searchEngine, const int pair)
    : NumbersComboBox(parent), m_searchEngine(searchEngine), m_pair(pair)
{
    setEditable(true);
    setInsertPolicy(QComboBox::NoInsert);
    setCompleter(0);

    connect(this, QOverload<int>::of(&QComboBox::activated), this, &PlayersComboBox::finishSearch);
    connect(this, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, [this] { emit selectionChanged(m_pair, currentNumber()); });

    m_matches = new QStringListModel(this);
    QCompleter *completer = new QCompleter(m_matches, this);
    completer->setCompletionMode(QCompleter::UnfilteredPopupCompletion);
    connect(completer, QOverload<const QString &>::of(&QCompleter::activated),
            this, &PlayersComboBox::completerActivated);

    PlayersListView *listView = new PlayersListView;
    completer->setPopup(listView);
    connect(listView, &PlayersListView::popupHidden, this, &PlayersComboBox::popupHidden);

    PlayersLineEdit *lineEdit = new PlayersLineEdit;
    lineEdit->setCompleter(completer);
    setLineEdit(lineEdit);

    disconnect(lineEdit, &QLineEdit::editingFinished, 0, 0);
    connect(lineEdit, &QLineEdit::textEdited, this, &PlayersComboBox::lineEditTextChanged);
    connect(lineEdit, &PlayersLineEdit::escapePressed,
            this, &PlayersComboBox::resetLineEditText);
    connect(lineEdit, &QLineEdit::editingFinished, this, &PlayersComboBox::processGivenText);

    m_isSearching = false;
}

void PlayersComboBox::lineEditTextChanged(const QString &text)
{
    if (! m_isSearching) {
        m_isSearching = true;
        m_completerActivated = false;
        lineEdit()->setStyleSheet(SharedStyles::STYLE_RED_TEXT);
        emit disableSaveScore();
    }

    QStringList matches;
    m_searchEngine->setSearchTerm(text);
    for(int i = 0; i < model()->rowCount(); i++) {
        const QString name = model()->index(i, 0).data().toString();
        if (m_searchEngine->checkMatch(name)) {
            matches.append(name);
        }
    }
    m_matches->setStringList(matches);
}

void PlayersComboBox::resetLineEditText()
{
    lineEdit()->setText(itemText(currentIndex()));
    finishSearch();
    lineEdit()->setFocus();
}

void PlayersComboBox::processGivenText()
{
    if (! m_isSearching || m_completerActivated) {
        return;
    }

    if (m_matches->rowCount() == 1) {
        completerActivated(m_matches->index(0, 0).data().toString());
        return;
    }

    resetLineEditText();
}

void PlayersComboBox::completerActivated(const QString &text)
{
    m_completerActivated = true;
    setCurrentIndex(findText(text));
    finishSearch();
}

void PlayersComboBox::finishSearch()
{
    if (m_isSearching) {
        m_isSearching = false;
        lineEdit()->setStyleSheet(QString());
    }
    emit checkSavePossible();
}

void PlayersComboBox::popupHidden()
{
    if (m_completerActivated || m_matches->rowCount() != 1) {
        return;
    }

    const QString onlyMatch = m_matches->index(0, 0).data().toString();
    if (lineEdit()->text() != onlyMatch) {
        completerActivated(onlyMatch);
    }
}
