/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "Server.h"
#include "Connection.h"

#include "PlayersPage/PlayersPage.h"
#include "PlayersPage/MarkersWidget.h"

#include "shared/ServerProtocol.h"
#include "shared/defaultValues.h"
#include "shared/SharedObjects.h"
#include "shared/StringTools.h"

// Qt includes
#include <QDebug>
#include <QTimer>
#include <QJsonArray>
#include <QApplication>

// C++ includes
#include <functional>

using namespace ServerProtocol;

Server::Server(QObject *parent, SharedObjects *sharedObjects, PlayersPage *playersPage)
    : QTcpServer(parent),
      m_db(sharedObjects->database()),
      m_playersPage(playersPage),
      m_markersWidget(playersPage->markersWidget()),
      m_locale(sharedObjects->stringTools()->locale())
{
}

void Server::close()
{
    for (Connection *connection : m_connections) {
        connection->close();
        removeConnection(connection);
    }
    QTcpServer::close();
}

void Server::incomingConnection(qintptr descriptor)
{
    // Create a new connection for the connection request
    Connection *connection = new Connection(this);
    if (! connection->setSocketDescriptor(descriptor)) {
        connection->deleteLater();
        return;
    }
    connect(connection, &Connection::disconnected,
            this, std::bind(&Server::removeConnection, this, connection));
    connect(connection, &Connection::handshakeResult,
            this, std::bind(&Server::processHandshake, this, connection, std::placeholders::_1));
    connect(connection, &Connection::clientRequest,
            this, std::bind(&Server::processClientRequest, this, connection,
                            std::placeholders::_1, std::placeholders::_2));
    connect(connection, &AbstractNetworkEngine::protocolError, this, &Server::protocolError);

    // Initiate a handshake

    m_pendingConnections.append(connection);
    emit message(tr("Verbindungsanfrage von %1, sende Handshake").arg(connection->ip()));

    // Send greeting message
    connection->sendData(HANDSHAKE_MESSAGE_INITIALIZE);

    // Start a concurrent timeout call
    QTimer::singleShot(NETWORK_CONNECT_TIMEOUT, this,
                       std::bind(&Server::checkHandshakeTimeout, this, connection));
}

void Server::removeConnection(Connection *connection)
{
    if (m_pendingConnections.removeOne(connection)) {
        emit message(tr("Verbindung mit %1 beendet").arg(connection->ip()));
    } else if (m_connections.removeOne(connection)) {
        emit message(tr("Verbindung mit %1 beendet").arg(clientString(connection)));
        emit statusUpdate(tr("Verbindung mit %1 beendet").arg(clientString(connection)));
        emit clientCountChanged(m_connections.count());
    }
    connection->deleteLater();
}

void Server::checkHandshakeTimeout(Connection *connection)
{
    if (m_pendingConnections.contains(connection)) {
        emit message(tr("Keine Antwort von %1").arg(connection->ip()));
        removeConnection(connection);
    }
}

void Server::broadcastMessage(QJsonObject &jsonObject, const int clientId,
                              const QString &messageText)
{
    emit message(messageText);
    jsonObject.insert(KEY_CHECKSUM, m_playersPage->checksum());
    jsonObject.insert(KEY_CLIENT_ID, clientId);
    for (Connection *connection : m_connections) {
        connection->sendData(jsonObject);
    }
}

int Server::clientCount() const
{
    return m_connections.count();
}

const QString Server::clientString(Connection *connection) const
{
    return QStringLiteral("%1 (Client-ID %2)").arg(connection->ip(),
                                                   QString::number(connection->id()));
}

// =========
// Handshake
// =========

void Server::processHandshake(Connection *connection, const QString &phase)
{
    if (! m_pendingConnections.contains(connection)) {
        return;
    }

    if (phase == HANDSHAKE_PHASE_SYNCHRONIZE) {
        emit message(tr("Synchronisation von %1 angefragt, sende Daten").arg(connection->ip()));

        QJsonObject settings {
            { HANDSHAKE_KEY_TOURNAMENT_MODE,    m_db->isPairMode() ? HANDSHAKE_FIXED_PAIRS
                                                                   : HANDSHAKE_SINGLE_PLAYERS },
            { HANDSHAKE_KEY_BOOGER_SCORE,       m_db->boogerScore() },
            { HANDSHAKE_KEY_BOOGERS_PER_ROUND,  m_db->boogersPerRound() },
            { HANDSHAKE_KEY_TOURNAMENT_STARTED, m_db->networkTournamentStarted() } };

        QJsonArray markers;
        for (const Marker::Marker &marker : m_db->markers()) {
            markers.append(QJsonObject {
                { HANDSHAKE_KEY_MARKER_ID,      marker.id },
                { HANDSHAKE_KEY_MARKER_NAME,    marker.name },
                { HANDSHAKE_KEY_MARKER_COLOR,   Connection::colorToString(marker.color) },
                { HANDSHAKE_KEY_MARKER_SORTING, static_cast<int>(marker.sorting) }
            });
        }

        QMap<QString, int> markedPlayers = m_db->playersMarks();
        QJsonArray players;
        for (const QString &name : m_db->listPlayers()) {
            int marker = 0;
            if (markedPlayers.contains(name)) {
                marker = markedPlayers.value(name);
            }
            players.append(QJsonObject { { name, marker } });
        }

        connection->sendData(QJsonObject { { KEY_ACTION,             ACTION_HANDSHAKE },
                                           { HANDSHAKE_KEY_PHASE,    HANDSHAKE_PHASE_SYNCHRONIZE },
                                           { HANDSHAKE_KEY_LOCALE,   m_locale },
                                           { HANDSHAKE_KEY_SETTINGS, settings },
                                           { HANDSHAKE_KEY_MARKERS,  markers },
                                           { HANDSHAKE_KEY_PLAYERS,  players },
                                           { KEY_CHECKSUM,           m_playersPage->checksum() } });

    } else if (phase == HANDSHAKE_PHASE_ALLOCATE_ID) {
        m_clientIdCounter++;
        connection->setId(m_clientIdCounter);
        connection->sendData(QJsonObject { { KEY_ACTION,          ACTION_HANDSHAKE },
                                           { HANDSHAKE_KEY_PHASE, HANDSHAKE_PHASE_ALLOCATE_ID },
                                           { KEY_CLIENT_ID,       m_clientIdCounter } });
        m_pendingConnections.removeOne(connection);
        m_connections.append(connection);
        emit message(tr("Synchronisation mit %1 war erfolgreich, Verbindung hergestellt "
                        "(Client-ID %2)").arg(connection->ip(), QString::number(connection->id())));
        emit statusUpdate(tr("Verbindung mit %1 hergestellt (Client-ID %2)").arg(
                             connection->ip(), QString::number(connection->id())));
        emit clientCountChanged(m_connections.count());
    }
}

// =================
// Broadcast changes
// =================

// Players
// =======

void Server::broadcastRegisterPlayers(const QString &name, const int marker, const int clientId)
{
    QJsonObject message { { KEY_ACTION,            ACTION_PLAYERS_REGISTER },
                          { PLAYERS_KEY_NAME,      name },
                          { PLAYERS_KEY_MARKER_ID, marker } };
    broadcastMessage(message, clientId, tr("Sende Anmelden von „%1“").arg(name));
}

void Server::broadcastRenamePlayers(const QString &name, const QString &newName, const int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_PLAYERS_RENAME },
                          { PLAYERS_KEY_NAME,     name },
                          { PLAYERS_KEY_NEW_NAME, newName } };
    broadcastMessage(message, clientId, tr("Sende Umbenennen von „%1“ zu „%2“").arg(name, newName));
}

void Server::broadcastDeletePlayers(const QString &name, const int clientId)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_DELETE },
                          { PLAYERS_KEY_NAME, name } };
    broadcastMessage(message, clientId, tr("Sende Löschen von „%1“").arg(name));
}

void Server::broadcastDeleteMarkedPlayers(const QString &markerName, const int markerId,
                                          const int clientId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_DELETE_MARKED },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    broadcastMessage(message, clientId,
                     tr("Sende Löschen aller als „%1“ markierten %2").arg(
                        markerName, m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Server::broadcastDeleteUnmarkedPlayers(const int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_UNMARKED } };
    broadcastMessage(message, clientId, tr("Sende Löschen aller unmarkierten %2").arg(
                                           m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Server::broadcastDeleteAllPlayers(const int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_DELETE_ALL } };
    broadcastMessage(message, clientId, tr("Sende Löschen aller %2").arg(
                                           m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
}

void Server::broadcastSetMarker(const QString &name, const QString &markerName,
                                const int markerId, const int clientId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_SET_MARKER },
                          { PLAYERS_KEY_NAME,        name },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    broadcastMessage(message, clientId,
                     tr("Sende Markieren von „%1“ als „%2“").arg(name, markerName));
}

void Server::broadcastRemoveMarker(const QString &name, const int clientId)
{
    QJsonObject message { { KEY_ACTION,       ACTION_PLAYERS_REMOVE_MARKER },
                          { PLAYERS_KEY_NAME, name } };
    broadcastMessage(message, clientId, tr("Sende Entfernen der Markierung von „%1“").arg(name));
}

void Server::broadcastListSetMarker(const QString &markerName, const int markerId,
                                    const int clientId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_LIST_SET_MARKER },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    broadcastMessage(message, clientId,
                     tr("Sende Markieren aller %1 als „%2“").arg(
                        m_db->isPairMode() ? tr("Paare") : tr("Spieler"), markerName));
}

void Server::broadcastListRemoveMarker(const QString &markerName, const int markerId,
                                       const int clientId)
{
    QJsonObject message { { KEY_ACTION,              ACTION_PLAYERS_LIST_REMOVE_MARKER },
                          { PLAYERS_KEY_MARKER_NAME, markerName },
                          { PLAYERS_KEY_MARKER_ID,   markerId } };
    broadcastMessage(message, clientId,
                     tr("Sende Entfernen aller „%1“-Markierungen").arg(markerName));
}

void Server::broadcastListRemoveAllMarkers(const int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_PLAYERS_LIST_REMOVE_ALL_MARKERS } };
    broadcastMessage(message, clientId, tr("Sende Entfernen aller Markierungen"));
}

// Markers
// =======

void Server::broadcastAddMarker(const QString &name, const QColor &color,
                                const Marker::Sorting sorting, const int clientId)
{
    QJsonObject message { { KEY_ACTION,          ACTION_MARKERS_ADD },
                          { MARKERS_KEY_NAME,    name },
                          { MARKERS_KEY_COLOR,   Connection::colorToString(color) },
                          { MARKERS_KEY_SORTING, sorting } };
    broadcastMessage(message, clientId, tr("Sende Hinzufügen der Markierung „%1“").arg(name));
}

void Server::broadcastDeleteMarker(const QString &name, const int id, const int sequence,
                                   const int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_DELETE },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_SEQUENCE, sequence } };
    broadcastMessage(message, clientId, tr("Sende Hinzufügen der Markierung „%1“").arg(name));
}

void Server::broadcastMoveMarkerUp(const QString &name, const int id, const int sequence,
                                   const int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_MOVE_UP },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_SEQUENCE, sequence } };
    broadcastMessage(message, clientId, tr("Sende Verschieben der Markierung „%1“").arg(name));
}

void Server::broadcastMoveMarkerDown(const QString &name, const int id, const int sequence,
                                     const int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_MOVE_DOWN },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_SEQUENCE, sequence } };
    broadcastMessage(message, clientId, tr("Sende Verschieben der Markierung „%1“").arg(name));
}

void Server::broadcastEditMarker(const QString &name, const int id, const QString &newName,
                                 const QColor &color, const Marker::Sorting sorting,
                                 const int clientId)
{
    QJsonObject message { { KEY_ACTION,           ACTION_MARKERS_EDIT },
                          { MARKERS_KEY_NAME,     name },
                          { MARKERS_KEY_ID,       id },
                          { MARKERS_KEY_NEW_NAME, newName },
                          { MARKERS_KEY_COLOR,    Connection::colorToString(color) },
                          { MARKERS_KEY_SORTING,  sorting } };
    broadcastMessage(message, clientId, tr("Sende Ändern der Markierung „%1“").arg(name));
}

// Finish registration
// ===================

void Server::broadcastFinishRegistration(const int clientId)
{
    QJsonObject message { { KEY_ACTION, ACTION_FINISH_REGISTRATION } };
    broadcastMessage(message, clientId, tr("Sende Beenden der Anmeldung"));
}

// ===============
// Client requests
// ===============

void Server::processClientRequest(Connection *connection,
                                              const QString &action, const QJsonObject &jsonObject)
{
    QApplication::setOverrideCursor(Qt::BusyCursor);

    const int clientId = jsonObject.value(KEY_CLIENT_ID).toInt();

    bool rejected = false;
    QString messageText;
    RejectReason reason;

    if (m_playersPage->checksum() != jsonObject.value(KEY_CHECKSUM).toString()) {
        rejected = true;
        messageText = tr("Prüfsummen stimmen nicht überein");
        reason = RejectReason::OutdatedRequest;
    }

    if (m_db->networkTournamentStarted() && (   action == ACTION_PLAYERS_REGISTER
                                             || action == ACTION_PLAYERS_DELETE
                                             || action == ACTION_PLAYERS_DELETE_MARKED
                                             || action == ACTION_PLAYERS_DELETE_UNMARKED
                                             || action == ACTION_PLAYERS_DELETE_ALL)) {

        rejected = true;
        messageText = tr("Die Anmeldung wurde bereits beendet");
        reason = RejectReason::RegistrationFinished;
    }

    if (action == ACTION_FINISH_REGISTRATION && (m_db->networkTournamentStarted()
                                                 || ! m_db->playersDividable())) {

        rejected = true;
        messageText = tr("Das Beenden der Anmeldung ist nicht möglich");
        reason = RejectReason::FinishRegistrationNotPossible;
    }

    if (rejected) {
        emit message(tr("Ungültige Anfrage von %1: %2").arg(clientString(connection), messageText));
        connection->sendData(QJsonObject {
            { KEY_ACTION,        ACTION_REJECTED },
            { KEY_REJECT_REASON, static_cast<int>(reason) },
            { KEY_CLIENT_ID,     clientId }
        });
        QApplication::restoreOverrideCursor();
        return;
    }

// Players
// =======

    if (action == ACTION_PLAYERS_REGISTER) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        const int marker = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Anfrage zum Anmelden von „%1“ von %2").arg(
                        name, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ angemeldet").arg(
                             m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name));
        m_db->registerPlayers(name, marker);
        m_playersPage->processNetworkChange();
        broadcastRegisterPlayers(name, marker, clientId);

    } else if (action == ACTION_PLAYERS_RENAME) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        const QString newName = jsonObject.value(PLAYERS_KEY_NEW_NAME).toString();
        emit message(tr("Anfrage zum Umbenennen von „%1“ zu „%2“ von %3").arg(
                        name, newName, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ zu „%3“ umbenannt").arg(
                             m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name, newName));
        m_db->renamePlayers(name, newName);
        m_playersPage->processNetworkChange();
        emit nameEdited();
        broadcastRenamePlayers(name, newName, clientId);

    } else if (action == ACTION_PLAYERS_DELETE) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        emit message(tr("Anfrage zum Löschen von „%1“ von %2").arg(name, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ gelöscht").arg(
                             m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name));
        m_db->deletePlayers(name);
        m_playersPage->processNetworkChange();
        broadcastDeletePlayers(name, clientId);

    } else if (action == ACTION_PLAYERS_DELETE_MARKED) {
        const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
        const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Anfrage zum Löschen aller als „%1“ markierten %2 von %3").arg(
                        markerName, m_db->isPairMode() ? tr("Paare") : tr("Spieler"),
                        clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Alle als %1 markierten „%2“ gelöscht").arg(
                             markerName, m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
        m_db->deleteMarkedPlayers(markerId);
        m_playersPage->processNetworkChange();
        broadcastDeleteMarkedPlayers(markerName, markerId, clientId);

    } else if (action == ACTION_PLAYERS_DELETE_UNMARKED) {
        emit message(tr("Anfrage zum Löschen aller unmarkierten %1 von %2").arg(
                        m_db->isPairMode() ? tr("Paare") : tr("Spieler"),
                        clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Alle unmarkierten %1 gelöscht").arg(
                             m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
        m_db->deleteUnmarkedPlayers();
        m_playersPage->processNetworkChange();
        broadcastDeleteUnmarkedPlayers(clientId);

    } else if (action == ACTION_PLAYERS_DELETE_ALL) {
        emit message(tr("Anfrage zum Löschen aller %1 von %2").arg(
                        m_db->isPairMode() ? tr("Paare") : tr("Spieler"),
                        clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Alle %1 gelöscht").arg(
                             m_db->isPairMode() ? tr("Paare") : tr("Spieler")));
        m_db->deleteAllPlayers();
        m_playersPage->processNetworkChange();
        broadcastDeleteAllPlayers(clientId);

    } else if (action == ACTION_PLAYERS_SET_MARKER) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
        const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Anfrage zum Markieren von „%1“ als „%2“ von %3").arg(
                        name, markerName, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: %1 „%2“ als „%3“ markiert").arg(
                             m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name, markerName));
        m_db->setMarker(name, markerId);
        m_playersPage->processNetworkChange();
        broadcastSetMarker(name, markerName, markerId, clientId);

    } else if (action == ACTION_PLAYERS_REMOVE_MARKER) {
        const QString name = jsonObject.value(PLAYERS_KEY_NAME).toString();
        emit message(tr("Anfrage zum Entfernen der Markierung von „%1“ von %3").arg(
                        name, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Markierung von %1 „%2“ entfernt").arg(
                             m_db->isPairMode() ? tr("Paar") : tr("Spieler"), name));
        m_db->removeMarker(name);
        m_playersPage->processNetworkChange();
        broadcastRemoveMarker(name, clientId);

    } else if (action == ACTION_PLAYERS_LIST_SET_MARKER) {
        const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
        const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Anfrage zum Markieren aller %1 als „%2“ von %3").arg(
                        m_db->isPairMode() ? tr("Paare") : tr("Spieler"), markerName,
                        clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Alle %1 als „%2“ markiert").arg(
                             m_db->isPairMode() ? tr("Paare") : tr("Spieler"), markerName));
        m_db->listSetMarker(markerId);
        m_playersPage->processNetworkChange();
        broadcastListSetMarker(markerName, markerId, clientId);

    } else if (action == ACTION_PLAYERS_LIST_REMOVE_MARKER) {
        const QString markerName = jsonObject.value(PLAYERS_KEY_MARKER_NAME).toString();
        const int markerId = jsonObject.value(PLAYERS_KEY_MARKER_ID).toInt();
        emit message(tr("Anfrage zum Entfernen aller „%1“-Markierungen von %2").arg(
                        markerName, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Alle „%1“-Markierungen entfernt").arg(markerName));
        m_db->listRemoveMarker(markerId);
        m_playersPage->processNetworkChange();
        broadcastListRemoveMarker(markerName, markerId, clientId);

    } else if (action == ACTION_PLAYERS_LIST_REMOVE_ALL_MARKERS) {
        emit message(tr("Anfrage zum Entfernen aller Markierungen von %1").arg(
                        clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Alle Markierungen entfernt"));
        m_db->listRemoveAllMarkers();
        m_playersPage->processNetworkChange();
        broadcastListRemoveAllMarkers(clientId);

// Markers
// =======

    } else if (action == ACTION_MARKERS_ADD) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        const QColor color = Connection::colorFromValue(jsonObject.value(MARKERS_KEY_COLOR));
        const Marker::Sorting sorting = static_cast<Marker::Sorting>(
            jsonObject.value(MARKERS_KEY_SORTING).toInt());
        emit message(tr("Anfrage zum Hinzufügen der Markierung „%1“ von %2").arg(
                        name, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ hinzugefügt").arg(name));
        m_db->addMarker(name, color, sorting);
        m_markersWidget->processNetworkChange();
        broadcastAddMarker(name, color, sorting, clientId);

    } else if (action == ACTION_MARKERS_DELETE) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        const int id = jsonObject.value(MARKERS_KEY_ID).toInt();
        const int sequence = jsonObject.value(MARKERS_KEY_SEQUENCE).toInt();
        emit message(tr("Anfrage zum Löschen der Markierung „%1“ von %2").arg(
                        name, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ gelöscht").arg(name));
        m_db->deleteMarker(id, sequence);
        m_markersWidget->processNetworkChange();
        broadcastDeleteMarker(name, id, sequence, clientId);

    } else if (action == ACTION_MARKERS_MOVE_UP) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        const int id = jsonObject.value(MARKERS_KEY_ID).toInt();
        const int sequence = jsonObject.value(MARKERS_KEY_SEQUENCE).toInt();
        emit message(tr("Anfrage zum Verschieben der Markierung „%1“ von %2").arg(
                        name, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ verschoben").arg(name));
        m_db->moveMarkerUp(id, sequence);
        m_markersWidget->processNetworkChange();
        broadcastMoveMarkerUp(name, id, sequence, clientId);

    } else if (action == ACTION_MARKERS_MOVE_DOWN) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        const int id = jsonObject.value(MARKERS_KEY_ID).toInt();
        const int sequence = jsonObject.value(MARKERS_KEY_SEQUENCE).toInt();
        emit message(tr("Anfrage zum Verschieben der Markierung „%1“ von %2").arg(
                        name, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ verschoben").arg(name));
        m_db->moveMarkerDown(id, sequence);
        m_markersWidget->processNetworkChange();
        broadcastMoveMarkerDown(name, id, sequence, clientId);

    } else if (action == ACTION_MARKERS_EDIT) {
        const QString name = jsonObject.value(MARKERS_KEY_NAME).toString();
        const int id = jsonObject.value(MARKERS_KEY_ID).toInt();
        const QString newName = jsonObject.value(MARKERS_KEY_NEW_NAME).toString();
        const QColor color = Connection::colorFromValue(jsonObject.value(MARKERS_KEY_COLOR));
        const Marker::Sorting sorting = static_cast<Marker::Sorting>(
            jsonObject.value(MARKERS_KEY_SORTING).toInt());
        emit message(tr("Anfrage zum Ändern der Markierung „%1“ von %2").arg(
                         name, clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Markierung „%1“ geändert").arg(name));
        m_db->editMarker(id, newName, color, sorting);
        m_markersWidget->processNetworkChange();
        broadcastEditMarker(name, id, newName, color, sorting, clientId);

// Finish registration
// ===================

    } else if (action == ACTION_FINISH_REGISTRATION) {
        emit message(tr("Anfrage zum Beenden der Anmeldung von %1").arg(clientString(connection)));
        emit statusUpdate(tr("Netzwerkänderung: Anmeldung beendet"));
        m_playersPage->networkFinishRegistration(clientId);
        broadcastFinishRegistration(clientId);

// Ranking
// =======

    } else if (action == ACTION_RANKING_REQUEST) {
        emit message(tr("Rangliste von %1 angefragt").arg(clientString(connection)));

        if (! m_db->localTournamentStarted()) {
            connection->sendData(QJsonObject { { KEY_ACTION, ACTION_RANKING_REQUEST },
                                               { KEY_CLIENT_ID, clientId },
                                               { RANKING_KEY_AVAILABLE, false } });
            emit message(tr("Es ist noch keine Rangliste verfügbar"));
            QApplication::restoreOverrideCursor();
            return;
        }

        QJsonArray names;
        QJsonArray boogers;
        QJsonArray goals;
        QJsonArray opponentGoals;
        QJsonArray playedBoogers;
        for (const Database::RankData &rank : m_db->ranking()) {
            names.append(QJsonValue(rank.name));
            boogers.append(QJsonValue(rank.boogers));
            goals.append(QJsonValue(rank.goals));
            opponentGoals.append(QJsonValue(rank.opponentGoals));
            playedBoogers.append(QJsonValue(rank.playedBoogers));
        }

        connection->sendData(QJsonObject { { KEY_ACTION,                 ACTION_RANKING_REQUEST },
                                           { KEY_CLIENT_ID,              clientId },
                                           { RANKING_KEY_AVAILABLE,      true },
                                           { RANKING_KEY_NAMES,          names },
                                           { RANKING_KEY_BOOGERS,        boogers },
                                           { RANKING_KEY_GOALS,          goals },
                                           { RANKING_KEY_OPPONENT_GOALS, opponentGoals },
                                           { RANKING_KEY_PLAYED_BOOGERS, playedBoogers } });

        emit message(tr("Rangliste an %1 gesendet").arg(clientString(connection)));

    } else {
        // This should not happen
        emit message(tr("Ungültige Anfrage von %1: Unbekannte Aktion").arg(
                        clientString(connection)));
        connection->sendData(QJsonObject { { KEY_ACTION,    ACTION_UNKNOWN },
                                           { KEY_CLIENT_ID, clientId } });
        QApplication::restoreOverrideCursor();
        return;
    }

    QApplication::restoreOverrideCursor();
}
