/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "ServerPage.h"
#include "Server.h"
#include "DiscoverServer.h"

#include "shared/LogDisplay.h"
#include "shared/SocketAddressWidget.h"
#include "shared/Database.h"
#include "shared/SharedStyles.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QLabel>
#include <QPushButton>
#include <QMessageBox>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QApplication>
#include <QStyle>
#include <QTimer>

ServerPage::ServerPage(QWidget *parent, SharedObjects *sharedObjects, PlayersPage *playersPage)
    : QWidget(parent)
{
    QVBoxLayout *layout = new QVBoxLayout(this);

    m_discoverServer = new DiscoverServer(this);

    // Server settings

    m_server = new Server(this, sharedObjects, playersPage);
    connect(m_server, &Server::statusUpdate, this, &ServerPage::statusUpdate);
    connect(m_server, &Server::clientCountChanged, this, &ServerPage::updateClientCount);

    m_settingsBox = new QGroupBox(tr("Socket-Adresse für den Server"));
    QHBoxLayout *settingsLayout = new QHBoxLayout(m_settingsBox);
    m_settingsBox->show();
    layout->addWidget(m_settingsBox);

    m_serverStoppedMessage = tr("<span style=\"%1\">Server<br/>gestoppt</span>").arg(
                                SharedStyles::STYLE_RED_TEXT);
    m_address = new QLabel(m_serverStoppedMessage);
    m_address->setTextInteractionFlags(Qt::TextSelectableByMouse);
    settingsLayout->addWidget(m_address);

    m_socketAddress = new SocketAddressWidget(SocketAddressWidget::Mode::Server);
    settingsLayout->addWidget(m_socketAddress);

    // Buttons

    m_startServer = new QPushButton(tr("Server starten"));
    m_startServer->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogYesButton));
    connect(m_socketAddress, &SocketAddressWidget::settingsOkay,
            m_startServer, &QPushButton::setEnabled);
    connect(m_startServer, &QPushButton::clicked, this, &ServerPage::startServer);
    settingsLayout->addWidget(m_startServer);

    m_stopServer = new QPushButton(tr("Server stoppen"));
    m_stopServer->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogCancelButton));
    connect(m_stopServer, &QPushButton::clicked, this, &ServerPage::stopServer);
    settingsLayout->addWidget(m_stopServer);

    settingsLayout->addStretch();

    m_closeButton = new QPushButton(tr("Schließen"));
    m_closeButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogCloseButton));
    connect(m_closeButton, &QPushButton::clicked, this, &ServerPage::closeServer);
    settingsLayout->addWidget(m_closeButton);

    // Server log

    m_displayBox = new QGroupBox(tr("Netzwerkinformationen"));
    QVBoxLayout *displayLayout = new QVBoxLayout(m_displayBox);
    layout->addWidget(m_displayBox);

    m_display = new LogDisplay;
    connect(m_server, &Server::protocolError, m_display, &LogDisplay::displayProtocolError);
    connect(m_server, &Server::message, m_display, &LogDisplay::displayMessage);
    connect(m_discoverServer, &DiscoverServer::message, m_display, &LogDisplay::displayMessage);

    displayLayout->addWidget(m_display);

    m_stopServer->hide();
    m_startServer->setFocus();
}

void ServerPage::startServer()
{
    m_display->clear();

    if (! m_server->listen(QHostAddress(m_socketAddress->ip()), m_socketAddress->port())) {
        m_display->displayMessage(tr("Der Muckturnier-Server konnte nicht gestartet werden: "
                                     "%1").arg(m_server->errorString()));
        QMessageBox::critical(this,
            tr("Server starten"),
            tr("<p>Der Muckturnier-Server konnte nicht gestartet werden:</p><p>%1</p>").arg(
               m_server->errorString()));
        return;
    }
    m_display->displayMessage(tr("Muckturnier-Server gestartet"));

    if (! m_discoverServer->serve(m_socketAddress->ip(), m_socketAddress->port())) {
        QTimer::singleShot(0, this, [this] {
            QMessageBox::warning(this,
                tr("Server starten"),
                tr("<p><b>Der Discover-Server konnte nicht gestartet werden</b></p>"
                   "<p>Die Fehlermeldung war: %1</p>"
                   "<p>Das automatische Auffinden des Muckturnier-Servers wird nicht "
                   "funktionieren, die Clients müssen die Socket-Adresse (IP-Adresse und Port) "
                   "manuell eingeben.</p>"
                   "<p><b>Die Funktion des Muckturnier-Servers ist dadurch aber nicht "
                   "eingeschränkt!</b></p>").arg(m_discoverServer->errorString()));
        });
        m_display->displayMessage(tr("Starten des Discover-Servers fehlgeschlagen"));
    } else {
        m_display->displayMessage(tr("Discover-Server gestartet"));
    }

    m_startServer->hide();
    m_socketAddress->hide();
    m_stopServer->show();
    m_closeButton->clearFocus();
    updateClientCount(0);

    emit serverStarted(m_server);
    emit statusUpdate(tr("Server unter IP %1 Port %2 gestartet").arg(
                         m_socketAddress->ip(), QString::number(m_socketAddress->port())));
}

void ServerPage::stopServer()
{
    if (! checkConnections()) {
        return;
    }

    m_server->close();
    m_display->displayMessage(tr("Muckturnier-Server gestoppt"));

    m_discoverServer->close();
    m_display->displayMessage(tr("Discover-Server gestoppt"));

    m_address->setText(m_serverStoppedMessage);
    m_startServer->show();
    m_stopServer->hide();
    m_socketAddress->show();
    m_startServer->setFocus();

    emit serverStopped();
    emit statusUpdate(tr("Server gestoppt"));
}

bool ServerPage::checkServerRunning()
{
    return checkConnections(tr("Muckturnier beenden"));
}

bool ServerPage::checkConnections(const QString &messageTitle)
{
    if (! m_server->isListening()) {
        return true;
    }

    if (m_server->clientCount() > 0) {
        if (QMessageBox::warning(this,
                messageTitle.isEmpty() ? tr("Server schließen") : messageTitle,
                tr("<p>Es ist mindestens ein Client angemeldet.</p>"
                   "<p>Wenn der Server gestoppt wird, dann werden alle Verbindungen getrennt "
                   "und es werden keine Änderungen mehr an die anderen Netzwerkteilnehmer gesendet "
                   "oder von ihnen empfangen!</p>"),
                QMessageBox::Ok | QMessageBox::Cancel,
                QMessageBox::Cancel) == QMessageBox::Cancel) {

                return false;
        }
    } else {
        if (QMessageBox::warning(this,
                messageTitle.isEmpty() ? tr("Server schließen") : messageTitle,
                tr("<p>Wenn der Server gestoppt wird, dann sind keine Verbindungen zum "
                   "Muckturnier-Netzwerk mehr möglich!</p>"),
                QMessageBox::Ok | QMessageBox::Cancel,
                QMessageBox::Cancel) == QMessageBox::Cancel) {

                return false;
        }
    }
    return true;
}

void ServerPage::closeServer()
{
    if (m_server->isListening() && ! checkConnections()) {
        return;
    }
    emit pageClosed();
    parent()->deleteLater();
}

void ServerPage::updateClientCount(int count)
{
    m_address->setText(tr("<span style=\"%1\">Server läuft</span><br/>"
                          "IP: <b>%2</b> Port: <b>%3</b><br/>"
                          "%4").arg(SharedStyles::STYLE_GREEN_TEXT,
                                    m_socketAddress->ip(),
                                    QString::number(m_socketAddress->port()),
                                    tr("%n Client(s) angemeldet", "", count)));
}
