/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "StringsPage.h"
#include "SpacesLineEdit.h"

#include "shared/Settings.h"
#include "shared/defaultValues.h"

// Qt includes
#include <QDebug>
#include <QGroupBox>
#include <QVBoxLayout>
#include <QGridLayout>
#include <QLabel>
#include <QComboBox>
#include <QMessageBox>
#include <QPushButton>

namespace
{
const QLatin1Char TABULATOR('\t');
}

StringsPage::StringsPage(QWidget *parent, Settings *settings)
    : AbstractSettingsPage(parent, tr("Zeichenketten")),
      m_settings(settings)
{
    QLabel *stringsLabel = new QLabel(tr("Folgende Zeichenketten für die Anzeige können "
                                         "eingestellt werden:"));
    stringsLabel->setWordWrap(true);
    layout()->addWidget(stringsLabel);

    QGridLayout *settingsLayout = new QGridLayout;
    layout()->addLayout(settingsLayout);

    int row = 0;

    settingsLayout->addWidget(new QLabel(tr("Bobbl-Symbol:")), row, 0);

    m_boogerSymbol = new QComboBox;
    m_boogerSymbol->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_boogerSymbol->addItem(QStringLiteral("\u2022")); // "BULLET"
    m_boogerSymbol->addItem(QStringLiteral("\u26AB")); // "MEDIUM BLACK CIRCLE"
    m_boogerSymbol->addItem(QStringLiteral(DEFAULT_BOOGER_SYMBOL)); // "BLACK CIRCLE"
    m_boogerSymbol->addItem(QStringLiteral("\u2B24")); // "BLACK LARGE CIRCLE"
    m_boogerSymbol->addItem(QStringLiteral("\u204E")); // "LOW ASTERISK"
    m_boogerSymbol->addItem(QStringLiteral("\u2055")); // "FLOWER PUNCTUATION MARK"
    m_boogerSymbol->addItem(QStringLiteral("\u274B")); // "HEAVY EIGHT TEARDROP-SPOKED PROPELLER"
    m_boogerSymbol->addItem(QStringLiteral("\u273D")); // "HEAVY TEARDROP-SPOKED ASTERISK"
    m_boogerSymbol->setWhatsThis(tr("Es kann jedes sichtbare Unicode-Zeichen als Bobbl-Symbol "
                                    "benutzt werden. In der Liste finden sich ein paar Vorschläge "
                                    "für geeignete Zeichen."));
    m_boogerSymbol->setEditable(true);
    m_boogerSymbol->setEditText(m_settings->boogerSymbol());
    settingsLayout->addWidget(m_boogerSymbol, row, 1);

    settingsLayout->addWidget(new QLabel(tr("Trenner für Paarnamen:\n(Für Einzelspieler-Turniere)")),
                         ++row, 0);

    m_namesSeparator = new SpacesLineEdit;
    m_namesSeparator->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_namesSeparator->setText(m_settings->namesSeparator());
    settingsLayout->addWidget(m_namesSeparator, row, 1);
    connect(m_namesSeparator, &SpacesLineEdit::textChanged,
            this, &StringsPage::updateSeparatorExample);

    QLabel *exampleLabel = new QLabel(tr("Beispiel:"));
    exampleLabel->setAlignment(Qt::AlignRight);
    settingsLayout->addWidget(exampleLabel, ++row, 0);

    m_separatorExample = new QLabel;
    settingsLayout->addWidget(m_separatorExample, row, 1);

    layout()->addStretch();
    QHBoxLayout *controlLayout = new QHBoxLayout;
    layout()->addLayout(controlLayout);
    controlLayout->addStretch();

    QPushButton *resetStringsButton = new QPushButton(tr("Zurücksetzen"));
    connect(resetStringsButton, &QPushButton::clicked, this, &StringsPage::resetStrings);
    controlLayout->addWidget(resetStringsButton);

    updateSeparatorExample();
}

void StringsPage::updateSeparatorExample()
{
    QString text = m_namesSeparator->text();
    if (text.contains(TABULATOR)) {
        QMessageBox::warning(this,
            tr("Trenner für Paarnamen"),
            tr("<p>Das Verwenden von Tabulatoren im Trenner für Paarnamen verursacht Probleme "
               "beim Export der Turnierdaten als CSV, da hier Tabulatoren als Feldtrenner benutzt "
               "werden.</p>"
               "<p> Es können daher keine Tabulatoren im Trenner benutzt werden.</p>"));
        m_namesSeparator->setText(text.remove(TABULATOR));
        return;
    }
    m_separatorExample->setText(tr("„Müller Andreas%1Meier Bernd“").arg(text));
}

void StringsPage::resetStrings()
{
    m_boogerSymbol->setCurrentText(QStringLiteral(DEFAULT_BOOGER_SYMBOL));
    m_namesSeparator->setText(QStringLiteral(DEFAULT_NAMES_SEPARATOR));
}

void StringsPage::saveSettings()
{
    const QString boogerSymbol = m_boogerSymbol->currentText().simplified();
    m_settings->saveBoogerSymbol(boogerSymbol.isEmpty() ? QStringLiteral(DEFAULT_BOOGER_SYMBOL)
                                                        : boogerSymbol);
    m_settings->saveNamesSeparator(m_namesSeparator->text().isEmpty()
                                       ? QStringLiteral(DEFAULT_NAMES_SEPARATOR)
                                       : m_namesSeparator->text());
}
