/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "TournamentTemplatePage.h"

#include "shared/Database.h"
#include "shared/TableDelegate.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QLabel>
#include <QTableWidget>
#include <QPushButton>
#include <QMessageBox>
#include <QHeaderView>

TournamentTemplatePage::TournamentTemplatePage(QWidget *parent, Database *db, Settings *settings)
    : AbstractSettingsPage(parent, tr("Vorlage Turniereinstellungen")),
      m_db(db), m_settings(settings)
{
    QLabel *tournamentTemplateLabel = new QLabel(tr("Folgende Turniereinstellungen werden für "
                                                    "neue Turnierdatenbanken voreingestellt:"));
    tournamentTemplateLabel->setWordWrap(true);
    layout()->addWidget(tournamentTemplateLabel);

    m_templateDisplay = new QTableWidget;
    m_templateDisplay->horizontalHeader()->hide();
    m_templateDisplay->verticalHeader()->hide();
    m_templateDisplay->setShowGrid(false);
    m_templateDisplay->setSelectionMode(QTableWidget::NoSelection);
    m_templateDisplay->setFocusPolicy(Qt::NoFocus);
    m_templateDisplay->setEditTriggers(QAbstractItemView::NoEditTriggers);
    m_templateDisplay->setItemDelegate(new TableDelegate(m_templateDisplay));
    m_templateDisplay->setColumnCount(2);
    m_templateDisplay->setRowCount(5);
    const int height = m_templateDisplay->rowHeight(0) * 5 + 4;
    m_templateDisplay->setMinimumHeight(height);
    layout()->addWidget(m_templateDisplay);

    m_templateDisplay->setItem(0, 0, new QTableWidgetItem(tr("Turniermodus:")));
    m_templateDisplay->setItem(1, 0, new QTableWidgetItem(tr("Punkte pro Bobbl:")));
    m_templateDisplay->setItem(2, 0, new QTableWidgetItem(tr("Bobbl pro Runde:")));
    m_templateDisplay->setItem(3, 0, new QTableWidgetItem(tr("Automatische Tisch- bzw. "
                                                             "Paarauswahl:")));
    m_templateDisplay->setItem(4, 0, new QTableWidgetItem(tr("Eingabe ohne Tischnummern:")));

    layout()->addStretch();

    QHBoxLayout *controlLayout = new QHBoxLayout;
    layout()->addLayout(controlLayout);
    controlLayout->addStretch();

    QPushButton *restoreTemplateButton = new QPushButton(tr("Zurücksetzen"));
    connect(restoreTemplateButton, &QPushButton::clicked,
            this, &TournamentTemplatePage::restoreTemplate);
    controlLayout->addWidget(restoreTemplateButton);

    QPushButton *adoptCurrentSettingsButton = new QPushButton(tr("Aktuelle Einstellungen "
                                                                 "übernehmen"));
    connect(adoptCurrentSettingsButton, &QPushButton::clicked,
            this, &TournamentTemplatePage::adoptCurrentSettings);
    controlLayout->addWidget(adoptCurrentSettingsButton);
    adoptCurrentSettingsButton->setEnabled(! m_db->dbFile().isEmpty());

    m_template = m_settings->tournamentTemplate();
    updateTemplateDisplay();
}

void TournamentTemplatePage::updateTemplateDisplay()
{
    m_templateDisplay->setItem(0, 1, new QTableWidgetItem(
        m_template.pairMode ? tr("Feste Paare") : tr("Einzelne Spieler")));
    m_templateDisplay->setItem(1, 1, new QTableWidgetItem(
        QString::number(m_template.boogerScore)));
    m_templateDisplay->setItem(2, 1, new QTableWidgetItem(
        QString::number(m_template.boogersPerRound)));
    m_templateDisplay->setItem(3, 1, new QTableWidgetItem(
        m_template.autoSelectPairs ? tr("Ja") : tr("Nein")));
    m_templateDisplay->setItem(4, 1, new QTableWidgetItem(
        m_template.noTableNumbers ? tr("Ja") : tr("Nein")));

    m_templateDisplay->resizeColumnsToContents();
    const int width = m_templateDisplay->columnWidth(0) + m_templateDisplay->columnWidth(1) + 4;
    m_templateDisplay->setMinimumWidth(width);
}

void TournamentTemplatePage::restoreTemplate()
{
    m_template = m_settings->defaultTournamentTemplate();
    updateTemplateDisplay();
}

void TournamentTemplatePage::adoptCurrentSettings()
{
    m_template.pairMode = m_db->isPairMode();
    m_template.boogerScore = m_db->boogerScore();
    m_template.boogersPerRound = m_db->boogersPerRound();
    m_template.autoSelectPairs = m_db->autoSelectPairs();
    m_template.noTableNumbers = m_db->noTableNumbers();
    updateTemplateDisplay();
}

void TournamentTemplatePage::saveSettings()
{
    m_settings->saveTournamentTemplate(m_template);
}
