/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "AbstractDiscoverEngine.h"

#include "shared/defaultValues.h"

// Qt includes
#include <QDebug>
#include <QUdpSocket>
#include <QJsonDocument>
#include <QJsonObject>

AbstractDiscoverEngine::AbstractDiscoverEngine(QObject *parent) : QObject(parent)
{
    m_socket = new QUdpSocket(this);
    connect(m_socket, &QAbstractSocket::readyRead, this, &AbstractDiscoverEngine::readDatagram);
}

bool AbstractDiscoverEngine::listen()
{
    return m_socket->bind(NETWORK_DISCOVER_PORT, QUdpSocket::ShareAddress);
}

QString AbstractDiscoverEngine::errorString()
{
    return m_socket->errorString();
}

void AbstractDiscoverEngine::close()
{
    m_socket->abort();
}

void AbstractDiscoverEngine::readDatagram()
{
    while (m_socket->state() == QAbstractSocket::BoundState && m_socket->hasPendingDatagrams()) {
        // Once Qt 5.7 support is dropped, we can use QUdpSocket::receiveDatagram() here
        QByteArray datagram;
        QHostAddress senderAddress;
        datagram.resize(int(m_socket->pendingDatagramSize()));
        m_socket->readDatagram(datagram.data(), datagram.size(), &senderAddress);

        QJsonParseError error;
        const QJsonDocument json = QJsonDocument::fromJson(datagram, &error);
        if (error.error != QJsonParseError::NoError || ! json.isObject()) {
            emit invalidData(senderAddress);
            continue;
        }

        processData(json.object(), senderAddress);
    }
}

void AbstractDiscoverEngine::broadcastData(const QJsonObject &jsonObject)
{
    m_socket->writeDatagram(QJsonDocument(jsonObject).toJson(QJsonDocument::Compact),
                            QHostAddress::Broadcast, NETWORK_DISCOVER_PORT);
}

void AbstractDiscoverEngine::sendData(const QJsonObject &jsonObject, const QHostAddress &address)
{
    m_socket->writeDatagram(QJsonDocument(jsonObject).toJson(QJsonDocument::Compact),
                            address, NETWORK_DISCOVER_PORT);
}
