/* Copyright (c) 2012-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "Database.h"
#include "StringTools.h"
#include "version.h"
#include "defaultValues.h"
#include "Characters.h"

// Qt includes
#include <QDebug>
#include <QSqlQuery>
#include <QDir>
#include <QSqlError>
#include <QLockFile>
#include <QCryptographicHash>

namespace
{

// The revisions we're using
constexpr int DB_VERSION = 7;

// Key literals for the database's "config" table
const QString KEY_MUCKTURNIER_VERSION    = QStringLiteral("MUCKTURNIER_VERSION");
const QString KEY_DB_VERSION             = QStringLiteral("DB_VERSION");
const QString KEY_TOURNAMENT_MODE        = QStringLiteral("TOURNAMENT_MODE");
const QString KEY_BOOGER_SCORE           = QStringLiteral("BOOGER_SCORE");
const QString KEY_BOOGERS_PER_ROUND      = QStringLiteral("BOOGERS_PER_ROUND");
const QString KEY_NO_TABLE_NUMBERS       = QStringLiteral("NO_TABLE_NUMBERS");
const QString KEY_AUTO_SELECT_PAIRS      = QStringLiteral("AUTO_SELECT_PAIRS");
const QString KEY_INCLUDE_OPPONENT_GOALS = QStringLiteral("INCLUDE_OPPONENT_GOALS");

// This one is only used to check if we can write to the database when opening it
// and is removed again afterwards, so it does never show up in a database in use.
const QString KEY_WRITE_TEST = QStringLiteral("WRITE_TEST");

// Values for KEY_TOURNAMENT_MODE
const QString FIXED_PAIRS    = QStringLiteral("FIXED_PAIRS");
const QString SINGLE_PLAYERS = QStringLiteral("SINGLE_PLAYERS");

// Literal bool values
const QString BOOL_FALSE = QStringLiteral("0");
const QString BOOL_TRUE  = QStringLiteral("1");

// Internally used connection name
const QString CONNECTION_NAME = QStringLiteral("SQLiteDB");

}

Database::Database(QObject *parent, StringTools *stringTools)
    : QObject(parent), m_stringTools(stringTools)
{
}

Database::~Database()
{
    closeDatabaseConnection();
}

bool Database::startTransaction()
{
    bool success = m_db.transaction();
    if (! success) {
        emitSqlError();
    }
    return success;
}

bool Database::commitTransaction()
{
    bool success = m_db.commit();
    if (! success) {
        emitSqlError();
    }
    return success;
}

bool Database::queryExec(QSqlQuery &query)
{
    bool success = query.exec();
    if (! success) {
        emitSqlError(query);
    }
    return success;
}

bool Database::queryExec(QSqlQuery &query, const QString &text)
{
    bool success = query.exec(text);
    if (! success) {
        emitSqlError(query);
    }
    return success;
}

bool Database::queryExecBatch(QSqlQuery &query)
{
    bool success = query.execBatch();
    if (! success) {
        emitSqlError(query);
    }
    return success;
}

void Database::emitSqlError()
{
    QSqlError error = m_db.lastError();
    emit sqlError(tr("<p>%1<br/>%2</p>").arg(error.databaseText(), error.driverText()));
}

void Database::emitSqlError(const QSqlQuery &query)
{
    QSqlError error = query.lastError();
    emit sqlError(tr("<p>%1<br/>%2</p>").arg(error.databaseText(), error.driverText()),
                    query.executedQuery());
}

bool Database::updateValue(QSqlQuery &query, const QString &key, bool state)
{
    query.prepare(QStringLiteral("UPDATE config SET value = ? WHERE key = ?"));
    query.bindValue(0, state ? BOOL_TRUE : BOOL_FALSE);
    query.bindValue(1, key);
    return queryExec(query);
}

void Database::closeDatabaseConnection()
{
    if (! m_db.isOpen()) {
        return;
    }

    // Save the display options if the database is writable
    if (m_isWritable && ! m_dbCheckRunning) {
        startTransaction();
        QSqlQuery query(m_db);
        updateValue(query, KEY_NO_TABLE_NUMBERS, m_noTableNumbers);
        updateValue(query, KEY_AUTO_SELECT_PAIRS, m_autoSelectPairs);
        updateValue(query, KEY_INCLUDE_OPPONENT_GOALS, m_includeOpponentGoals);
        commitTransaction();
    }

    // Close the existing connection
    m_db.close();
    m_db = QSqlDatabase::database();
    QSqlDatabase::removeDatabase(CONNECTION_NAME);

    // Remove the lock file
    delete m_lockFile;

    // Clear cached variables
    m_dbFileName.clear();
    m_dbFilePath.clear();
    m_tournamentStarted = false;
}

bool Database::connectDatabase(const QString &dbFile)
{
    closeDatabaseConnection();

    // Create a Lockfile

    QString lockFileName = QDir::temp().absoluteFilePath(
        QStringLiteral("muckturnier.%1.lock").arg(
                       QString::fromUtf8(QCryptographicHash::hash(dbFile.toUtf8(),
                       QCryptographicHash::Md5).toHex()))
    );

    m_lockFile = new QLockFile(lockFileName);
    m_lockFile->setStaleLockTime(0);

    if (! m_lockFile->tryLock(0)) {
        m_errorMessage = tr("<p>Die Datenbank ist bereits von einer anderen Instanz von "
                            "Muckturnier geöffnet.</p>"
                            "<p>Sollte dies nicht der Fall sein (evtl. nach einen Crash, "
                            "Stromausfall etc.), dann bitte die Lock-Datei „%1“ manuell löschen "
                            "und die Datenbank erneut öffnen.</p>").arg(
                            QDir::toNativeSeparators(lockFileName));
        return false;
    }

    // Connect the database
    m_db = QSqlDatabase::addDatabase(QStringLiteral("QSQLITE"), CONNECTION_NAME);
    m_db.setDatabaseName(dbFile);

    updateDbFileInfo();

    if (! m_db.open()) {
        // This should not happen
        QSqlError error = m_db.lastError();
        m_errorMessage = tr("Unerwarteter Fehler: Konnte keine Datenbankverbindung herstellen.</p>"
                            "<p>Die Fehlermeldung war: %2</p>").arg(error.text());
        return false;
    }

    return true;
}

void Database::updateDbFileInfo()
{
    m_dbFileName = m_db.databaseName();
    QFileInfo info(m_dbFileName);
    m_dbFilePath = info.canonicalPath();
}

bool Database::create(const QString &dbFile)
{
    if (! connectDatabase(dbFile)) {
        return false;
    }

    startTransaction();
    QSqlQuery query(m_db);

    // Setup the tables
    queryExec(query, QStringLiteral("CREATE TABLE config(\n"
                                    "    key TEXT,\n"
                                    "    value TEXT\n"
                                    ")"));
    queryExec(query, QStringLiteral("CREATE TABLE players(\n"
                                    "    id INTEGER PRIMARY KEY,\n"
                                    "    name TEXT,\n"
                                    "    marker INTEGER\n"
                                    ")"));
    queryExec(query, QStringLiteral("CREATE TABLE markers(\n"
                                    "    id INTEGER PRIMARY KEY,\n"
                                    "    name TEXT,\n"
                                    "    color INTEGER,\n"
                                    "    sorting INTEGER,\n"
                                    "    sequence INTEGER\n"
                                    ")"));
    queryExec(query, QStringLiteral("CREATE TABLE assignment(\n"
                                    "    round INTEGER,\n"
                                    "    pair_or_player INTEGER,\n"
                                    "    pair_number INTEGER,\n"
                                    "    player_number INTEGER,\n"
                                    "    table_number INTEGER\n"
                                    ")"));
    queryExec(query, QStringLiteral("CREATE TABLE score(\n"
                                    "    round INTEGER,\n"
                                    "    table_number INTEGER,\n"
                                    "    booger INTEGER,\n"
                                    "    pair_or_player INTEGER,\n"
                                    "    score INTEGER\n"
                                    ")"));

    // Insert the default config values
    query.prepare(QStringLiteral("INSERT INTO config(key, value) VALUES(?, ?)"));
    query.addBindValue(QVariantList { KEY_MUCKTURNIER_VERSION,
                                      KEY_DB_VERSION,
                                      KEY_TOURNAMENT_MODE,
                                      KEY_BOOGER_SCORE,
                                      KEY_BOOGERS_PER_ROUND,
                                      KEY_NO_TABLE_NUMBERS,
                                      KEY_AUTO_SELECT_PAIRS,
                                      KEY_INCLUDE_OPPONENT_GOALS });
    query.addBindValue(QVariantList { QStringLiteral(MT_VERSION),
                                      DB_VERSION,
                                      FIXED_PAIRS,
                                      DEFAULT_BOOGER_SCORE,
                                      DEFAULT_BOOGERS_PER_ROUND,
                                      BOOL_FALSE,
                                      BOOL_TRUE,
                                      BOOL_FALSE });
    queryExecBatch(query);

    commitTransaction();

    readSettingsFromDatabase();
    m_isWritable = true;
    m_tournamentStarted = false;

    updateDbFileInfo();

    return true;
}

void Database::readSettingsFromDatabase()
{
    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("SELECT key, value FROM config"));

    while (query.next()) {
        QString key = query.value(0).toString();
        QString value = query.value(1).toString();

        if (key == KEY_MUCKTURNIER_VERSION) {
            m_dbValueMuckturnierVersion = value;
        } else if (key == KEY_TOURNAMENT_MODE) {
            if (value == FIXED_PAIRS) {
                m_settingsTournamentMode = TournamentMode::FixedPairs;
            } else if (value == SINGLE_PLAYERS) {
                m_settingsTournamentMode = TournamentMode::SinglePlayers;
            }
        } else if (key == KEY_BOOGER_SCORE) {
            m_settingsBoogerScore = value.toInt();
        } else if (key == KEY_BOOGERS_PER_ROUND) {
            m_settingsBoogersPerRound = value.toInt();
        } else if (key == KEY_NO_TABLE_NUMBERS) {
            m_noTableNumbers = value == BOOL_FALSE ? false : true;
        } else if (key == KEY_AUTO_SELECT_PAIRS) {
            m_autoSelectPairs = value == BOOL_FALSE ? false : true;
        } else if (key == KEY_INCLUDE_OPPONENT_GOALS) {
            m_includeOpponentGoals = value == BOOL_FALSE ? false : true;
        }
    }
}

bool Database::tournamentStarted() const
{
    return m_tournamentStarted || m_networkTournamentStarted;
}

bool Database::localTournamentStarted() const
{
    return m_tournamentStarted;
}

Database::Error Database::checkDatabase(const QString &dbFile)
{
    // Step 1: Check if it's a file and if we can read it

    QFileInfo fileInfo = QFileInfo(dbFile);
    if (! fileInfo.exists()) {
        m_errorMessage = tr("<p>Die Datei „%1“ konnte nicht gefunden werden</p>").arg(
                            QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }
    if (! (fileInfo.isFile() || fileInfo.isSymLink())) {
        m_errorMessage = tr("<p>„%1“ ist weder eine Datei, noch ein symbolischer Link auf eine "
                            "Datei</p>").arg(QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }
    if (! fileInfo.isReadable()) {
        m_errorMessage = tr("<p>Die Datei „%1“ ist nicht lesbar. Bitte die Zugriffsrechte "
                            "überprüfen!</p>").arg(QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    }

    // Step 2: Check if we can connect the database and if it's actually an SQLite database

    if (! connectDatabase(dbFile)) {
        return Error::CouldNotOpen;
    }

    startTransaction();
    QSqlQuery query(m_db);

    query.exec(QStringLiteral("PRAGMA quick_check"));
    QString testResult;
    if (query.next()) {
        testResult = query.value(0).toString();
    }

    if (testResult != QLatin1String("ok")) {
        QSqlError error = query.lastError();
        m_errorMessage = tr("<p>Der Datenbankcheck für die Datei „%1“ ist fehlgeschlagen.</p>"
                            "<p>Die Fehlermeldung war: %2</p>").arg(
                            QDir::toNativeSeparators(dbFile), error.text());
        commitTransaction();
        m_isWritable = false;
        return Error::CouldNotOpen;
    }

    // Step 3: Check if it's a Muckturnier database

    // Let's see if we have all tables
    queryExec(query, QStringLiteral("SELECT name FROM sqlite_master "
                                    "WHERE type = 'table' "
                                    "AND name = 'config' "
                                    "OR name = 'players' "
                                    "OR name = 'assignment' "
                                    "OR name = 'score'"));
    int tableCount = 0;
    while (query.next()) {
        tableCount++;
    }
    if (tableCount != 4) {
        m_errorMessage = tr("<p>In der Datenbank „%1“ sind nicht alle nötigen Tabellen vorhanden. "
                            "Entweder ist die Datei keine Muckturnier-Datenbank, oder sie wurde "
                            "mit einer zu alten Version von Muckturnier (vor 0.5.1) erstellt."
                            "</p>").arg(QDir::toNativeSeparators(dbFile));
        commitTransaction();
        m_isWritable = false;
        return Error::CouldNotOpen;
    }

    /*
     * Let's see if the database is writable.
     *
     * We do have to check if a database is writable by actually trying to insert some data (is
     * this a bug or a feature?!) because:
     *
     * - If the file itself has write access, QFileInfo::isWritable() will return true, but if the
     *   parent directory is not writable, we won't be able to write to the database.
     *
     * - Checking for the parent directory to be writable in addition does also not help, as we will
     *   get false for "My Documents" on Windows, although we can write to it.
     *
     * - Opening the file via QFile::open() does actually work if the file is writable and the
     *   parent directory is not (with the QIODevice::WriteOnly flag, the file will even be
     *   trucated!), but we still can't insert data in this case.
     *
     * So here we go:
     */

    query.prepare(QStringLiteral("INSERT INTO config(key, value) VALUES(?, ?)"));
    query.bindValue(0, KEY_WRITE_TEST);
    query.bindValue(1, QStringLiteral("test"));

    // We use query.exec() here directly, so that no error is displayed if the database is not
    // writable
    if (query.exec()) {
        query.prepare(QStringLiteral("DELETE FROM config WHERE key = ?"));
        query.bindValue(0, KEY_WRITE_TEST);
        queryExec(query);
        m_isWritable = true;
    } else {
        m_isWritable = false;
    }

    // Let's see if we have the correct database version

    query.prepare(QStringLiteral("SELECT value FROM config WHERE key = ?"));
    query.bindValue(0, KEY_DB_VERSION);
    queryExec(query);

    int dbVersion = 0;
    if (query.next()) {
        dbVersion = query.value(0).toInt();
    }

    commitTransaction();

    if (dbVersion < 3) {
        m_errorMessage = tr("<p>Es wurde in der Datenbank „%1“ kein gültiger Konfigurationseintrag "
                            "zur verwendeten Datenbankrevision gefunden. Die Datei ist entweder "
                            "keine Muckturnier-Datenbank oder sie wurde mit einer zu alten Version "
                            "von Muckturnier (vor 0.5.1) erstellt.</p>").arg(
                            QDir::toNativeSeparators(dbFile));
        return Error::CouldNotOpen;
    } else if (dbVersion > DB_VERSION) {
        m_errorMessage = tr("<p>Die Datenbank „%1“ hat die falsche Revision (ist %2, die benutzte "
                            "Version von Muckturnier nutzt %3). Sie wurde vermutlich mit einer "
                            "neueren Version erstellt.</p>"
                            "<p>Bitte auf die neueste Version aktualisieren (zu finden unter "
                            "<a href=\"http://muckturnier.org/\">http://muckturnier.org/</a>) und "
                            "die Datenbank damit erneut öffnen.</p>").arg(
                            QDir::toNativeSeparators(dbFile), QString::number(dbVersion),
                                                              QString::number(DB_VERSION));
        return Error::CouldNotOpen;
    } else if (dbVersion == 3) {
        return Error::DbVersion3;
    } else if (dbVersion == 4) {
        return Error::DbVersion4;
    } else if (dbVersion == 5) {
        return Error::DbVersion5;
    } else if (dbVersion == 6) {
        return Error::DbVersion6;
    } else if (dbVersion != DB_VERSION) {
        // This should not happen
        m_errorMessage = tr("<p>Die Datenbank „%1“ hat die falsche Revision (ist %2, die benutzte "
                            "Version von Muckturnier nutzt %3). Die Datenbank kann nicht geöffnet "
                            "werden").arg(
                            QDir::toNativeSeparators(dbFile), QString::number(dbVersion),
                                                              QString::number(DB_VERSION));
        return Error::CouldNotOpen;
    }

    // If we reach here, we actually can assume that this is a valid Muckturnier database ;-)
    m_dbCheckRunning = true;
    closeDatabaseConnection();
    m_dbCheckRunning = false;
    return Error::NoError;
}

Database::Error Database::openDatabase(const QString &dbFile)
{
    Error testResult = checkDatabase(dbFile);
    if (testResult != Error::NoError) {
        return testResult;
    }

    connectDatabase(dbFile);
    readSettingsFromDatabase();
    checkIfTournamentStarted();

    return m_isWritable ? Error::NoError
                        : Error::NoErrorReadOnly;
}

const QString & Database::errorMessage() const
{
    return m_errorMessage;
}

void Database::reset()
{
    startTransaction();
    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("DELETE FROM players"));
    queryExec(query, QStringLiteral("DELETE FROM markers"));
    queryExec(query, QStringLiteral("DELETE FROM assignment"));
    queryExec(query, QStringLiteral("DELETE FROM score"));
    commitTransaction();
    m_tournamentStarted = false;
}

void Database::saveSettings()
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("UPDATE config SET value = ? WHERE key = ?"));
    query.addBindValue(QVariantList { isPairMode() ? FIXED_PAIRS : SINGLE_PLAYERS,
                                      m_settingsBoogerScore,
                                      m_settingsBoogersPerRound });
    query.addBindValue(QVariantList { KEY_TOURNAMENT_MODE,
                                      KEY_BOOGER_SCORE,
                                      KEY_BOOGERS_PER_ROUND });
    queryExecBatch(query);
    commitTransaction();
}

void Database::checkIfTournamentStarted()
{
    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("SELECT COUNT(round) FROM SCORE"));
    query.next();
    m_tournamentStarted = query.value(0).toInt() != 0;
}

int Database::necessaryDivisor() const
{
    return isPairMode() ? 2 : 4;
}

bool Database::playersDividable(int count) const
{
    return count > 0 && count % necessaryDivisor() == 0;
}

bool Database::playersDividable()
{
    return playersDividable(playersCount());
}

QStringList Database::listPlayers()
{
    QStringList playersList;
    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("SELECT name FROM players"));
    while (query.next()) {
        playersList.append(query.value(0).toString());
    }
    return playersList;
}

void Database::registerPlayers(const QString &name, int marker)
{
    startTransaction();
    QSqlQuery query(m_db);

    query.prepare(QStringLiteral("INSERT INTO players(id, name, marker) VALUES(NULL, ?, ?)"));
    query.bindValue(0, name);
    if (marker == 0) {
        query.bindValue(1, QVariant(QVariant::Int)); // This will map to NULL
    } else {
        query.bindValue(1, marker);
    }

    queryExec(query);
    commitTransaction();
}

void Database::registerPlayers(const QStringList &names)
{
    startTransaction();
    QSqlQuery query(m_db);
    registerPlayers(query, names);
    commitTransaction();
}

void Database::registerPlayers(QSqlQuery &query, const QStringList &names)
{
    query.prepare(QStringLiteral("INSERT INTO players(id, name) VALUES(NULL, ?)"));
    query.addBindValue(stringListToVariantList(names));
    queryExecBatch(query);
}

void Database::deletePlayers(const QString &name)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("DELETE FROM players WHERE name = ?"));
    query.bindValue(0, name);
    queryExec(query);
    commitTransaction();
}

void Database::deletePlayers(const QStringList &names)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("DELETE FROM players WHERE name = ?"));
    query.addBindValue(stringListToVariantList(names));
    queryExecBatch(query);
    commitTransaction();
}

void Database::renamePlayers(const QString &oldName, const QString &newName)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("UPDATE players SET name = ? WHERE name = ?"));
    query.bindValue(0, newName);
    query.bindValue(1, oldName);
    queryExec(query);
    commitTransaction();
}

void Database::renamePlayers(const QStringList &oldNames, const QStringList &newNames)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("UPDATE players SET name = ? WHERE name = ?"));
    query.addBindValue(stringListToVariantList(newNames));
    query.addBindValue(stringListToVariantList(oldNames));
    queryExecBatch(query);
    commitTransaction();
}

int Database::playersCount()
{
    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("SELECT COUNT(id) AS count FROM players"));
    query.next();
    return query.value(0).toInt();
}


QMap<int, bool> Database::roundsState()
{
    QMap<int, bool> roundsStateMap;
    int allBoogers = playersCount() * m_settingsBoogersPerRound;
    int roundNumber = 1;

    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("SELECT round, COUNT(round) AS count FROM score "
                                    "GROUP BY round ORDER BY round"));

    // Check if each round already contains all boogers
    while (query.next()) {
        roundNumber = query.value(0).toInt();
        if (query.value(1).toInt() < allBoogers) {
            roundsStateMap[roundNumber] = false;
        } else {
            roundsStateMap[roundNumber] = true;
        }
    }

    // Be sure to have a state for each round. All data from a round could have been deleted.
    // If we have no data, round 1 will be added as open.
    for (int i = 1; i <= roundNumber; i++) {
        if (! roundsStateMap.contains(i)) {
            roundsStateMap[i] = false;
        }
    }

    return roundsStateMap;
}

QList<int> Database::missingTables(int round)
{
    QList<int> missingTableNumbers;

    // First, we assume we need all tables
    for (int i = 1; i <= playersCount() / necessaryDivisor(); i++) {
        missingTableNumbers.append(i);
    }

    // Then, we take away those that already have a score
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("SELECT table_number FROM score "
                                 "WHERE round = ? "
                                 "GROUP BY table_number"));
    query.bindValue(0, round);
    queryExec(query);
    while (query.next()) {
        missingTableNumbers.removeOne(query.value(0).toInt());
    }

    return missingTableNumbers;
}

QMap<QString, int> Database::missingPlayers(int round)
{
    QMap<QString, int> missingPlayersMap;
    QSqlQuery query(m_db);

    query.prepare(QStringLiteral("SELECT players.name, players.id FROM players "
                                 "WHERE NOT EXISTS ( "
                                     "SELECT score.pair_or_player FROM score "
                                     "WHERE players.id = score.pair_or_player "
                                     "AND round = ? "
                                 ")"));
    query.bindValue(0, round);
    queryExec(query);
    while (query.next()) {
        missingPlayersMap[query.value(0).toString()] = query.value(1).toInt();
    }

    return missingPlayersMap;
}

Database::Assignment Database::assignment(int roundNumber)
{
    int round = roundNumber - 1;
    int allTables = playersCount() / necessaryDivisor();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("SELECT pair_or_player, pair_number, table_number "
                                 "FROM assignment WHERE round = ?"));
    query.bindValue(0, round);
    queryExec(query);

    QMap<int, int> assignedTable;
    QMap<int, int> assignedPairNumber;
    QMap<int, QMap<int, int>> assignedPair;

    while(query.next()) {
        if (query.value(1).toInt() == 1) {
            assignedTable[query.value(0).toInt()] = query.value(2).toInt();
        } else {
            assignedTable[query.value(0).toInt()] = query.value(2).toInt() % allTables + 1;
        }
        assignedPairNumber[query.value(0).toInt()] = query.value(1).toInt();
    }

    for (int pairNumber : assignedTable.keys()) {
        assignedPair[assignedTable[pairNumber]][assignedPairNumber[pairNumber]] = pairNumber;
    }

    Assignment assignment;
    assignment.assignedTable = assignedTable;
    assignment.assignedPairNumber = assignedPairNumber;
    assignment.assignedPair = assignedPair;

    return assignment;
}

QList<Database::Score> Database::scores(int round)
{
    QSqlQuery query(m_db);
    switch (m_settingsTournamentMode) {
    case TournamentMode::FixedPairs:
        query.prepare(QStringLiteral(
            "SELECT "
                "score.table_number, "
                "players.name, "
                "score.score "
            "FROM score "
            "INNER JOIN players ON "
                "players.id = score.pair_or_player "
            "INNER JOIN assignment ON "
                "assignment.round = score.round "
                "AND assignment.pair_or_player = score.pair_or_player "
            "WHERE score.round = ? "
            "ORDER BY "
                "score.table_number, "
                "assignment.pair_number, "
                "score.booger"
        ));
        query.bindValue(0, round);
        break;
    case TournamentMode::SinglePlayers:
        // It's a bit more complicated for single players ;-)
        query.prepare(QStringLiteral(
            "SELECT "
                "score_one.table_number, "
                "player_one.name || ? || player_two.name, "
                "score_one.score "
            "FROM "
                "score AS score_one "
            "INNER JOIN players AS player_one ON "
                "player_one.id = score_one.pair_or_player "
            "INNER JOIN assignment AS assignment_one ON "
                "assignment_one.round = score_one.round "
                "AND assignment_one.table_number = score_one.table_number "
                "AND assignment_one.pair_or_player = score_one.pair_or_player "
            "INNER JOIN players AS player_two ON "
                "player_two.id = score_two.pair_or_player "
            "INNER JOIN score AS score_two ON "
                "score_two.round = score_one.round "
                "AND score_two.table_number = score_one.table_number "
                "AND score_two.booger = score_one.booger "
                "AND assignment_two.pair_number = assignment_one.pair_number "
                "AND assignment_two.player_number = 2 "
            "INNER JOIN assignment AS assignment_two ON "
                "assignment_two.round = score_one.round "
                "AND assignment_two.table_number = score_one.table_number "
                "AND assignment_two.pair_or_player = score_two.pair_or_player "
            "WHERE "
                "score_one.round = ? "
                "AND assignment_one.player_number = 1 "
            "ORDER BY "
                "score_one.table_number, "
                "assignment_one.pair_number, "
                "score_one.booger "
        ));
        query.bindValue(0, m_namesSeparator);
        query.bindValue(1, round);
        break;
    }
    queryExec(query);

    QList<Score> scores;
    while (query.next()) {
        Score score;
        score.table = query.value(0).toInt();
        score.pairName = query.value(1).toString();
        score.score = query.value(2).toInt();
        scores.append(score);
    }

    return scores;
}

QList<Database::BoogerResult> Database::singleScore(int round, int table)
{
    QList<BoogerResult> score;
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral(
        "SELECT "
            "score.pair_or_player, "
            "score.score "
        "FROM score "
        "INNER JOIN assignment ON "
            "assignment.round = score.round "
            "AND assignment.pair_or_player = score.pair_or_player "
        "WHERE "
            "score.round = ? "
            "AND score.table_number = ? "
        "ORDER BY "
            "assignment.pair_number, "
            "assignment.player_number, "
            "score.booger"
    ));
    query.bindValue(0, round);
    query.bindValue(1, table);
    queryExec(query);

    while (query.next()) {
        BoogerResult boogerResult;
        boogerResult.pairOrPlayer = query.value(0).toInt();
        boogerResult.score = query.value(1).toInt();
        score.append(boogerResult);
    }

    return score;
}

void Database::saveScore(QSqlQuery &query, int round, int table, int booger,
                         const BoogerResult &boogerResult)
{
    query.prepare(QStringLiteral(
        "INSERT INTO score(round, table_number, booger, pair_or_player, score) "
        "VALUES(?, ?, ?, ?, ?)"
    ));
    query.bindValue(0, round);
    query.bindValue(1, table);
    query.bindValue(2, booger);
    query.bindValue(3, boogerResult.pairOrPlayer);
    query.bindValue(4, boogerResult.score);
    queryExec(query);
}

void Database::updateAssignment(QSqlQuery &query, int round, int pairOrPlayer,
                                                  int pairNumber, int playerNumber, int table)
{
    query.prepare(QStringLiteral(
        "INSERT INTO assignment(round, pair_or_player, pair_number, player_number, table_number) "
        "VALUES(?, ?, ?, ?, ?)"
    ));
    query.bindValue(0, round);
    query.bindValue(1, pairOrPlayer);
    query.bindValue(2, pairNumber);
    query.bindValue(3, playerNumber);
    query.bindValue(4, table);
    queryExec(query);
}

void Database::saveScore(int round, int table, const QList<Database::BoogerResult> &score)
{
    startTransaction();
    QSqlQuery query(m_db);

    switch (m_settingsTournamentMode) {
    case FixedPairs:
        updateAssignment(query, round, score.at(0).pairOrPlayer,
                         1, 1, table);
        updateAssignment(query, round, score.at(m_settingsBoogersPerRound).pairOrPlayer,
                         2, 1, table);
        break;
    case SinglePlayers:
        updateAssignment(query, round, score.at(0).pairOrPlayer,
                         1, 1, table);
        updateAssignment(query, round, score.at(m_settingsBoogersPerRound).pairOrPlayer,
                         1, 2, table);
        updateAssignment(query, round, score.at(m_settingsBoogersPerRound * 2).pairOrPlayer,
                         2, 1, table);
        updateAssignment(query, round, score.at(m_settingsBoogersPerRound * 3).pairOrPlayer,
                         2, 2, table);
        break;
    }

    int booger = 1;
    for (const BoogerResult &boogerResult : score) {
        saveScore(query, round, table, booger, boogerResult);
        booger++;
        if (booger > m_settingsBoogersPerRound) {
            booger = 1;
        }
    }

    commitTransaction();
    m_tournamentStarted = true;
}

void Database::deleteScore(int round, int table)
{
    startTransaction();
    QSqlQuery query(m_db);

    query.prepare(QStringLiteral("DELETE FROM score WHERE round = ? AND table_number = ?"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    queryExec(query);

    query.prepare(QStringLiteral("DELETE FROM assignment WHERE round = ? AND table_number = ?"));
    query.bindValue(0, round);
    query.bindValue(1, table);
    queryExec(query);

    commitTransaction();
    checkIfTournamentStarted();
}

bool Database::ranksDiffer(const RankData &currentRank, const RankData &lastRank)
{
    if (m_includeOpponentGoals) {
        return currentRank.boogers != lastRank.boogers
               || currentRank.goals != lastRank.goals
               || currentRank.opponentGoals != lastRank.opponentGoals;
    } else {
        return currentRank.boogers != lastRank.boogers
               || currentRank.goals != lastRank.goals;
    }
}

QList<Database::RankData> Database::ranking(int upToRound)
{
    if (upToRound == -1) {
        upToRound = roundsState().count();
    }

    QSqlQuery query(m_db);
    query.prepare(QStringLiteral(
        "SELECT "
            "players.name, "
            "SUM(CASE WHEN one_score.score = ? THEN 1 ELSE 0 END) AS won_boogers, "
            "SUM(CASE WHEN one_score.score = ? THEN 0 ELSE one_score.score END) AS scored_goals, "
            "SUM(CASE WHEN other_score.score = ? THEN 0 ELSE other_score.score END) AS opponent_goals, "
            "COUNT(one_score.score) AS played_boogers "
        "FROM "
            "score AS one_score "
        "INNER JOIN players ON "
            "players.id = one_score.pair_or_player "
        "INNER JOIN assignment AS one_assignment ON "
            "one_assignment.round = one_score.round "
            "AND one_assignment.table_number = one_score.table_number "
            "AND one_assignment.pair_or_player = one_score.pair_or_player "
        "INNER JOIN assignment AS other_assignment ON "
            "other_assignment.round = other_score.round "
            "AND other_assignment.table_number = other_score.table_number "
            "AND other_assignment.pair_or_player = other_score.pair_or_player "
        "INNER JOIN score AS other_score ON "
            "other_score.round = one_score.round "
            "AND other_score.table_number = one_score.table_number "
            "AND other_score.booger = one_score.booger "
            "AND other_assignment.pair_number != one_assignment.pair_number "
        "WHERE one_score.round <= ? "
        "GROUP BY "
            "one_score.pair_or_player "
        "ORDER BY "
            "won_boogers DESC, "
            "scored_goals DESC, "
            "opponent_goals, "
            "players.name"
    ));
    for (int i = 0; i < 3; i++) {
        query.bindValue(i, m_settingsBoogerScore);
    }
    query.bindValue(3, upToRound);
    queryExec(query);

    int rankingLine = 0;
    RankData lastRank;
    lastRank.rank = 0;
    lastRank.boogers = 0;
    lastRank.goals = 0;
    lastRank.opponentGoals = 0;

    QList<RankData> ranking;

    while (query.next()) {
        RankData currentRank;
        currentRank.name = query.value(0).toString();
        currentRank.boogers = query.value(1).toInt();
        currentRank.goals = query.value(2).toInt();
        currentRank.opponentGoals = query.value(3).toInt();
        currentRank.playedBoogers = query.value(4).toInt();
        if (m_settingsTournamentMode == TournamentMode::SinglePlayers) {
            currentRank.boogers /= 2;
            currentRank.goals /= 2;
            currentRank.opponentGoals /= 2;
            currentRank.playedBoogers /= 2;
        }

        rankingLine++;
        if (ranksDiffer(currentRank, lastRank)) {
            currentRank.rank = rankingLine;
        } else {
            currentRank.rank = lastRank.rank;
        }

        ranking.append(currentRank);
        lastRank = currentRank;
    }

    if (ranking.count() < playersCount()) {
        // Some pair(s)/player(s) don't have a score yet,
        // so we insert empty scores to get a complete ranking list

        queryExec(query, QStringLiteral("SELECT players.name FROM players "
                                        "WHERE NOT EXISTS ( "
                                            "SELECT score.pair_or_player "
                                            "FROM score "
                                            "WHERE players.id = score.pair_or_player "
                                        ")"));

        QStringList missingNames;
        while (query.next()) {
            missingNames.append(query.value(0).toString());
        }
        m_stringTools->sortList(missingNames);

        RankData missingRank;
        missingRank.boogers = 0;
        missingRank.goals = 0;
        missingRank.opponentGoals = 0;
        missingRank.playedBoogers = 0;
        if (ranksDiffer(missingRank, lastRank)) {
            missingRank.rank = lastRank.rank + 1;
        } else {
            missingRank.rank = lastRank.rank;
        }

        for (const QString &name : missingNames) {
            missingRank.name = name;
            ranking.append(missingRank);
        }
    }

    return ranking;
}

const QString & Database::muckturnierVersion() const
{
    return m_dbValueMuckturnierVersion;
}

Database::TournamentMode Database::tournamentMode() const
{
    return m_settingsTournamentMode;
}

const QString & Database::dbFile() const
{
    return m_dbFileName;
}

const QString & Database::dbPath() const
{
    return m_dbFilePath;
}

void Database::setTournamentMode(Database::TournamentMode tournamentMode)
{
    m_settingsTournamentMode = tournamentMode;
}

int Database::dbVersion() const
{
    return DB_VERSION;
}

int Database::boogerScore() const
{
    return m_settingsBoogerScore;
}

void Database::setBoogerScore(int boogerScore)
{
    m_settingsBoogerScore = boogerScore;
}

int Database::boogersPerRound() const
{
    return m_settingsBoogersPerRound;
}

void Database::setBoogersPerRound(int boogersPerRound)
{
    m_settingsBoogersPerRound = boogersPerRound;
}

void Database::setNamesSeparator(const QString &text)
{
    m_namesSeparator = text;
}

void Database::updateDatabase(const QString &dbFileName, Error dbVersion)
{
    // Setting this prevents the current GUI-changeable settings (auto-select pairs, include
    // opponent goals etc.) to be written to the database when closing the connection. Those
    // settings aren't set to the GUI using this function and would be overwritten by the default
    // ones for a fixed pair tournament otherwise (which sucks notably for a single player
    // tournanment with "auto select pairs" being set to true and being unchangeable thus always
    // yielding incorrect warnings ;-)
    m_dbCheckRunning = true;

    connectDatabase(dbFileName);
    startTransaction();
    QSqlQuery query(m_db);

    switch (dbVersion) {
    case NoError:
    case NoErrorReadOnly:
    case CouldNotOpen:
        Q_UNREACHABLE();
        break;
    case Error::DbVersion3:
        updateDatabaseFromV3(query);
        break;
    case Error::DbVersion4:
        updateDatabaseFromV4(query);
        break;
    case Error::DbVersion5:
        updateDatabaseFromV5(query);
        break;
    case Error::DbVersion6:
        updateDatabaseFromV6(query);
        break;
    }

    commitTransaction();
    closeDatabaseConnection();

    m_dbCheckRunning = false;
}

void Database::updateDatabaseFromV3(QSqlQuery &query)
{
    // Get the tournament mode
    queryExec(query, QStringLiteral("SELECT value FROM config WHERE key = 'TOURNAMENT_MODE'"));
    query.next();
    const QString tournamentMode = query.value(0).toString();

    // Add the missing column to "assignment"
    queryExec(query, QStringLiteral("ALTER TABLE assignment ADD COLUMN player_number INTEGER"));

    // Fill in the data. For a single player mode database, the fetch result hopefully is still in
    // the correct order, so that the right players become player 1 and 2 inside a team.
    if (tournamentMode == QStringLiteral("FIXED_PAIRS")) {
        queryExec(query, QStringLiteral("UPDATE assignment SET player_number = 1"));
    } else {
        QVariantList round;
        QVariantList pair_or_player;
        QVariantList pair_number;
        QVariantList table_number;
        QVariantList player_number;
        queryExec(query, QStringLiteral("SELECT round, pair_or_player, pair_number, table_number "
                                        "FROM assignment "
                                        "ORDER BY round, table_number, pair_number"));
        int playerNumber = 1;
        while (query.next()) {
            round.append(query.value(0).toInt());
            pair_or_player.append(query.value(1).toInt());
            pair_number.append(query.value(2).toInt());
            table_number.append(query.value(3).toInt());
            playerNumber = (playerNumber + 1) % 2;
            player_number.append(playerNumber + 1);
        }

        query.prepare(QStringLiteral("UPDATE assignment SET player_number = ? "
                                     "WHERE round = ? "
                                     "AND pair_or_player = ? "
                                     "AND pair_number = ? "
                                     "AND table_number = ?"));
        query.addBindValue(player_number);
        query.addBindValue(round);
        query.addBindValue(pair_or_player);
        query.addBindValue(pair_number);
        query.addBindValue(table_number);
        queryExecBatch(query);
    }

    // Update the "DB_VERSION" key
    queryExec(query, QStringLiteral("UPDATE config SET value = 4 WHERE key = 'DB_VERSION'"));

    // Go on updating
    updateDatabaseFromV4(query);
}

void Database::updateDatabaseFromV4(QSqlQuery &query)
{
    // Get the boogers per round
    queryExec(query, QStringLiteral("SELECT value FROM config WHERE key = 'GAMES_PER_ROUND'"));
    query.next();
    const QString boogersPerRound = query.value(0).toString();

    // Update "GAMES_PER_ROUND" to "BOOGERS_PER_ROUND"
    queryExec(query, QStringLiteral("DELETE FROM config WHERE key = 'GAMES_PER_ROUND'"));
    query.prepare(QStringLiteral("INSERT INTO config(key, value) VALUES('BOOGERS_PER_ROUND', ?)"));
    query.addBindValue(boogersPerRound);
    queryExec(query);

    // Update the score table's "game" column to "booger"
    queryExec(query, QStringLiteral("ALTER TABLE score RENAME TO score_tmp"));
    queryExec(query, QStringLiteral("CREATE TABLE score(\n"
                                    "    round INTEGER,\n"
                                    "    table_number INTEGER,\n"
                                    "    booger INTEGER,\n"
                                    "    pair_or_player INTEGER,\n"
                                    "    score INTEGER\n"
                                    ")"));
    queryExec(query, QStringLiteral("INSERT INTO score(round, table_number, booger, "
                                                      "pair_or_player, score) "
                                    "SELECT round, table_number, game, pair_or_player, score "
                                    "FROM score_tmp"));
    queryExec(query, QStringLiteral("DROP TABLE score_tmp"));

    // Update the "DB_VERSION" key
    queryExec(query, QStringLiteral("UPDATE config SET value = 5 WHERE key = 'DB_VERSION'"));

    // Go on updating
    updateDatabaseFromV5(query);
}

void Database::updateDatabaseFromV5(QSqlQuery &query)
{
    // Insert the new config keys. We don't use the defined constants here in case they change with
    // some later version.
    query.prepare(QStringLiteral("INSERT INTO config(key, value) VALUES(?, ?)"));
    query.addBindValue(QVariantList { QStringLiteral("NO_TABLE_NUMBERS"),
                                      QStringLiteral("AUTO_SELECT_PAIRS"),
                                      QStringLiteral("INCLUDE_OPPONENT_GOALS") });
    query.addBindValue(QVariantList { QStringLiteral("0"),
                                      QStringLiteral("1"),
                                      QStringLiteral("0") });
    queryExecBatch(query);

    // Update the "DB_VERSION" key
    queryExec(query, QStringLiteral("UPDATE config SET value = 6 WHERE key = 'DB_VERSION'"));

    // Go on updating
    updateDatabaseFromV6(query);
}

void Database::updateDatabaseFromV6(QSqlQuery &query)
{
    // Add the new "markers" table
    queryExec(query, QStringLiteral("CREATE TABLE markers(\n"
                                    "    id INTEGER PRIMARY KEY,\n"
                                    "    name TEXT,\n"
                                    "    color INTEGER,\n"
                                    "    sorting INTEGER,\n"
                                    "    sequence INTEGER\n"
                                    ")"));

    // Add the new "marker" column to the "players" table
    queryExec(query, QStringLiteral("ALTER TABLE players ADD COLUMN marker INTEGER"));

    // Add the default markers using a direct query in case the format changes later:
    query.prepare(QStringLiteral("INSERT INTO markers(id, name, color, sorting, sequence) "
                                 "VALUES(?, ?, ?, ?, ?)"));
    query.addBindValue(QVariantList { 0, 1, 2, 3 });
    query.addBindValue(QVariantList { QString(),
                                      tr("Rot"),
                                      tr("Blau"),
                                      tr("Grün") });
    query.addBindValue(QVariantList { QColor(0, 0, 0, 0).rgb(),
                                      QColor(255, 0, 0, 0).rgb(),
                                      QColor(0, 0, 255, 0).rgb(),
                                      QColor(0, 128, 0, 0).rgb() });
    query.addBindValue(QVariantList { 0, 0, 0, 1 });
    query.addBindValue(QVariantList { 2, 0, 1, 3 });
    queryExecBatch(query);

    // Map the old markers to the new ones

    QVariantList oldNames;
    QVariantList newNames;
    QVariantList markerIds;

    // Search for marked names
    query.prepare(QStringLiteral("SELECT name FROM players"));
    queryExec(query);
    while (query.next()) {
        const QString name = query.value(0).toString();
        int marker = -1;
        if (name.startsWith(Characters::MARKER) && ! name.endsWith(Characters::MARKER)) {
            marker = 1;
        } else if (! name.startsWith(Characters::MARKER) && name.endsWith(Characters::MARKER)) {
            marker = 3;
        }
        if (marker == -1) {
            continue;
        }
        QString strippedName = name;
        strippedName.remove(Characters::MARKER);
        oldNames.append(name);
        newNames.append(strippedName);
        markerIds.append(marker);
    }

    // Update the names and the marker ids
    query.prepare(QStringLiteral("UPDATE players SET name = ?, marker = ? WHERE name = ?"));
    query.addBindValue(newNames);
    query.addBindValue(markerIds);
    query.addBindValue(oldNames);
    queryExecBatch(query);

    // Update the "DB_VERSION" key
    queryExec(query, QStringLiteral("UPDATE config SET value = 7 WHERE key = 'DB_VERSION'"));
}

bool Database::isWritable() const
{
    return m_isWritable;
}

bool Database::noTableNumbers() const
{
    return m_noTableNumbers;
}

bool Database::autoSelectPairs() const
{
    return m_autoSelectPairs;
}

bool Database::includeOpponentGoals() const
{
    return m_includeOpponentGoals;
}

void Database::setNoTableNumbers(bool state)
{
    m_noTableNumbers = state;
}

void Database::setAutoSelectPairs(bool state)
{
    m_autoSelectPairs = state;
}

void Database::setIncludeOpponentGoals(bool state)
{
    m_includeOpponentGoals = state;
}

const QVariantList Database::stringListToVariantList(const QStringList &stringList) const
{
    QVariantList variantList;
    for (const QString &text : stringList) {
        variantList.append(QVariant(text));
    }
    return variantList;
}

QList<Marker::Marker> Database::markers()
{
    QList<Marker::Marker> markers;
    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("SELECT id, name, color, sorting FROM markers "
                                    "ORDER BY sequence"));
    while (query.next()) {
        Marker::Marker marker;
        marker.id = query.value(0).toInt();
        marker.name = query.value(1).toString();
        marker.color = QColor(query.value(2).toInt());
        marker.sorting = static_cast<Marker::Sorting>(query.value(3).toInt());
        markers.append(marker);
    }

    return markers;
}

void Database::addMarkers(const QList<Marker::Marker> &markers)
{
    startTransaction();
    QSqlQuery query(m_db);
    for (const Marker::Marker &marker : markers) {
        addMarker(query, marker.name, marker.color, marker.sorting, marker.id);
    }
    commitTransaction();
}

void Database::addMarker(const QString &name, const QColor &color, Marker::Sorting sorting)
{
    startTransaction();
    QSqlQuery query(m_db);
    addMarker(query, name, color, sorting);
    commitTransaction();
}

void Database::addMarker(QSqlQuery &query, const QString &name, const QColor &color,
                         Marker::Sorting sorting, int id)
{
    query.prepare(QStringLiteral("INSERT INTO markers(id, name, color, sorting, sequence) "
                                 "VALUES(?, ?, ?, ?, (SELECT COUNT(sequence) FROM markers))"));
    if (id == -1) {
        query.bindValue(0, QVariant(QVariant::Int)); // This will map to NULL
    } else {
        query.bindValue(0, id);
    }
    query.bindValue(1, name);
    query.bindValue(2, color.rgb());
    query.bindValue(3, sorting);
    queryExec(query);
}

void Database::deleteMarker(int id, int sequence)
{
    startTransaction();
    QSqlQuery query(m_db);

    // Delete the marker
    query.prepare(QStringLiteral("DELETE FROM markers WHERE id = ?"));
    query.bindValue(0, id);
    queryExec(query);

    // Close gaps in the sequence counter
    query.prepare(QStringLiteral("UPDATE markers SET sequence = sequence - 1 WHERE sequence > ?"));
    query.bindValue(0, sequence);
    queryExec(query);

    // Remove all respective marker links from the pairs/players list
    query.prepare(QStringLiteral("UPDATE players SET marker = NULL WHERE marker = ?"));
    query.bindValue(0, id);
    queryExec(query);

    commitTransaction();
}

void Database::editMarker(int id, const QString &name, const QColor &color, Marker::Sorting sorting)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("UPDATE markers SET name = ?, color = ?, sorting = ? "
                                 "WHERE id = ?"));
    query.bindValue(0, name);
    query.bindValue(1, color.rgb());
    query.bindValue(2, sorting);
    query.bindValue(3, id);
    queryExec(query);
    commitTransaction();
}

void Database::moveMarkerDown(int id, int sequence)
{
    startTransaction();
    QSqlQuery query(m_db);

    if (sequence == 0) {
        // Be sure that the new first marker will create a new block
        query.prepare(QStringLiteral("UPDATE markers SET sorting = ? WHERE sequence = 1"));
        query.bindValue(0, Marker::Sorting::NewBlock);
        queryExec(query);
    }

    query.prepare(QStringLiteral("UPDATE markers SET sequence = sequence - 1 WHERE sequence = ?"));
    query.bindValue(0, sequence + 1);
    queryExec(query);

    query.prepare(QStringLiteral("UPDATE markers SET sequence = ? WHERE id = ?"));
    query.bindValue(0, sequence + 1);
    query.bindValue(1, id);
    queryExec(query);

    commitTransaction();
}

void Database::moveMarkerUp(int id, int sequence)
{
    startTransaction();
    QSqlQuery query(m_db);

    if (sequence == 1) {
        // Be sure that the new first marker will create a new block
        query.prepare(QStringLiteral("UPDATE markers SET sorting = ? WHERE sequence = 1"));
        query.bindValue(0, Marker::Sorting::NewBlock);
        queryExec(query);
    }

    query.prepare(QStringLiteral("UPDATE markers SET sequence = sequence + 1 WHERE sequence = ?"));
    query.bindValue(0, sequence - 1);
    queryExec(query);

    query.prepare(QStringLiteral("UPDATE markers SET sequence = ? WHERE id = ?"));
    query.bindValue(0, sequence - 1);
    query.bindValue(1, id);
    queryExec(query);

    commitTransaction();
}

void Database::setMarker(const QString &name, int markerId)
{
    startTransaction();
    QSqlQuery query(m_db);
    setMarker(query, name, markerId);
    commitTransaction();
}

void Database::setMarker(QSqlQuery &query, const QString &name, int markerId)
{
    query.prepare(QStringLiteral("UPDATE players SET marker = ? WHERE name = ?"));
    query.bindValue(0, markerId);
    query.bindValue(1, name);
    queryExec(query);
}

void Database::removeMarker(const QString &name)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("UPDATE players SET marker = NULL WHERE name = ?"));
    query.bindValue(0, name);
    queryExec(query);
    commitTransaction();
}

QMap<QString, int> Database::playersMarks()
{
    QMap<QString, int> markedPlayers;
    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("SELECT name, "
                                    "CASE WHEN marker IS NULL "
                                        "THEN 0 "
                                        "ELSE marker "
                                    "END "
                                    "FROM players"));
    while (query.next()) {
        markedPlayers[query.value(0).toString()] = query.value(1).toInt();
    }
    return markedPlayers;
}

void Database::listSetMarker(int markerId)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("UPDATE players SET marker = ?"));
    query.bindValue(0, markerId);
    queryExec(query);
    commitTransaction();
}

void Database::listRemoveMarker(int markerId)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("UPDATE players SET marker = NULL where marker = ?"));
    query.bindValue(0, markerId);
    queryExec(query);
    commitTransaction();
}

void Database::listRemoveAllMarkers()
{
    startTransaction();
    QSqlQuery query(m_db);
    listRemoveAllMarkers(query);
    commitTransaction();
}

void Database::listRemoveAllMarkers(QSqlQuery &query)
{
    queryExec(query, QStringLiteral("UPDATE players SET marker = NULL"));
}

void Database::deleteMarkedPlayers(int markerId)
{
    startTransaction();
    QSqlQuery query(m_db);
    query.prepare(QStringLiteral("DELETE FROM players where marker = ?"));
    query.bindValue(0, markerId);
    queryExec(query);
    commitTransaction();
}

void Database::deleteUnmarkedPlayers()
{
    startTransaction();
    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("DELETE FROM players where marker IS NULL"));
    commitTransaction();
}

void Database::deleteAllPlayers()
{
    startTransaction();
    QSqlQuery query(m_db);
    queryExec(query, QStringLiteral("DELETE FROM players"));
    commitTransaction();
}

bool Database::isPairMode() const
{
    return m_settingsTournamentMode == TournamentMode::FixedPairs;
}

bool Database::isSinglePlayerMode() const
{
    return m_settingsTournamentMode == TournamentMode::SinglePlayers;
}

void Database::adoptServerData(const QStringList &players, const QList<Marker::Marker> &markers,
                               const QMap<QString, int> &markedPlayers)
{
    startTransaction();
    QSqlQuery query(m_db);

    queryExec(query, QStringLiteral("DELETE FROM markers"));
    for (const Marker::Marker &marker : markers) {
        addMarker(query, marker.name,  marker.color, marker.sorting, marker.id);
    }

    queryExec(query, QStringLiteral("DELETE FROM players"));
    registerPlayers(query, players);

    for (const QString &name : markedPlayers.keys()) {
        setMarker(query, name, markedPlayers[name]);
    }

    commitTransaction();
}

void Database::adoptServerData(const QList<Marker::Marker> &markers,
                               const QMap<QString, int> &markedPlayers)
{
    startTransaction();
    QSqlQuery query(m_db);

    listRemoveAllMarkers(query);

    queryExec(query, QStringLiteral("DELETE FROM markers"));
    for (const Marker::Marker &marker : markers) {
        addMarker(query, marker.name,  marker.color, marker.sorting, marker.id);
    }

    for (const QString &name : markedPlayers.keys()) {
        setMarker(query, name, markedPlayers[name]);
    }

    commitTransaction();
}

void Database::setNetworkTournamentStarted(bool state)
{
    m_networkTournamentStarted = state;
}

bool Database::networkTournamentStarted() const
{
    return m_networkTournamentStarted;
}
