/* Copyright (c) 2012-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef DATABASE_H
#define DATABASE_H

// Local includes
#include "Marker.h"

// Qt includes
#include <QObject>
#include <QSqlDatabase>
#include <QMap>
#include <QColor>

// Local classes
class StringTools;

// Qt classes
class QSqlQuery;
class QLockFile;

class Database : public QObject
{
    Q_OBJECT

public:
    enum Error {
        NoError,
        NoErrorReadOnly,
        CouldNotOpen,
        DbVersion3,
        DbVersion4,
        DbVersion5,
        DbVersion6
    };

    enum TournamentMode {
        FixedPairs,
        SinglePlayers
    };

    struct Assignment {
        QMap<int, int> assignedTable;
        QMap<int, int> assignedPairNumber;
        QMap<int, QMap<int, int>> assignedPair;
    };

    struct Score {
        int table;
        QString pairName;
        int score;
    };

    struct BoogerResult {
        int pairOrPlayer;
        int score;
    };

    struct RankData {
        int rank;
        QString name;
        int boogers;
        int goals;
        int opponentGoals;
        int playedBoogers;
    };

    explicit Database(QObject *parent, StringTools *stringTools);
    ~Database();

    // Underlying database handling
    bool create(const QString &dbFile);
    Database::Error openDatabase(const QString &dbFile);
    Database::Error checkDatabase(const QString &dbFile);
    void updateDatabase(const QString &dbFileName, Error dbVersion);
    void updateDatabaseFromV3(QSqlQuery &query);
    void updateDatabaseFromV4(QSqlQuery &query);
    void updateDatabaseFromV5(QSqlQuery &query);
    void updateDatabaseFromV6(QSqlQuery &query);
    void saveSettings();
    void closeDatabaseConnection();
    void reset();

    // Settings
    void setTournamentMode(Database::TournamentMode tournamentMode);
    void setBoogerScore(int boogerScore);
    void setBoogersPerRound(int boogersPerRound);
    void setNamesSeparator(const QString &text);
    void setNoTableNumbers(bool state);
    bool noTableNumbers() const;
    void setAutoSelectPairs(bool state);
    bool autoSelectPairs() const;
    void setIncludeOpponentGoals(bool state);
    bool includeOpponentGoals() const;
    void setNetworkTournamentStarted(bool state);
    bool networkTournamentStarted() const;
    bool localTournamentStarted() const;

    // Functions only returning const values/references or boolean operations on them
    bool isWritable() const;
    bool tournamentStarted() const;
    bool isPairMode() const;
    bool isSinglePlayerMode() const;
    int dbVersion() const;
    Database::TournamentMode tournamentMode() const;
    int boogerScore() const;
    int boogersPerRound() const;
    int necessaryDivisor() const;
    bool playersDividable(int count) const;
    const QString & errorMessage() const;
    const QString & muckturnierVersion() const;
    const QString & dbFile() const;
    const QString & dbPath() const;

    // Data handling

    void addMarkers(const QList<Marker::Marker> &markers);
    void registerPlayers(const QString &name, int marker);
    void registerPlayers(const QStringList &names);
    void renamePlayers(const QString &oldName, const QString &newName);
    void renamePlayers(const QStringList &oldNames, const QStringList &newNames);
    void deletePlayers(const QString &name);
    void deletePlayers(const QStringList &names);
    void deleteMarkedPlayers(int markerId);
    void deleteUnmarkedPlayers();
    void deleteAllPlayers();

    void setMarker(const QString &name, int markerId);
    void listSetMarker(int markerId);
    void removeMarker(const QString &name);
    void listRemoveMarker(int markerId);
    void listRemoveAllMarkers();

    void addMarker(const QString &name, const QColor &color, Marker::Sorting sorting);
    void editMarker(int id, const QString &name, const QColor &color, Marker::Sorting sorting);
    void deleteMarker(int id, int sequence);
    void moveMarkerDown(int id, int sequence);
    void moveMarkerUp(int id, int sequence);

    void saveScore(int round, int table, const QList<Database::BoogerResult> &score);
    void deleteScore(int round, int table);

    void adoptServerData(const QStringList &players, const QList<Marker::Marker> &markers,
                         const QMap<QString, int> &markedPlayers);
    void adoptServerData(const QList<Marker::Marker> &markers,
                         const QMap<QString, int> &markedPlayers);

    // Functions returning data using queries
    QStringList listPlayers();
    QMap<int, bool> roundsState();
    QList<int> missingTables(int round);
    QList<Score> scores(int round);
    QList<BoogerResult> singleScore(int round, int table);
    QList<RankData> ranking(int upToRound = -1);
    QMap<QString, int> missingPlayers(int round);
    QList<Marker::Marker> markers();
    QMap<QString, int> playersMarks();
    int playersCount();
    Assignment assignment(int roundNumber);
    bool playersDividable();

signals:
    void sqlError(const QString &error, const QString &query = QString());

private: // Functions
    bool connectDatabase(const QString &dbFile);
    bool startTransaction();
    bool commitTransaction();
    bool queryExec(QSqlQuery &query);
    bool queryExec(QSqlQuery &query, const QString &text);
    bool queryExecBatch(QSqlQuery &query);
    void emitSqlError();
    void emitSqlError(const QSqlQuery &query);
    void readSettingsFromDatabase();
    void checkIfTournamentStarted();
    void saveScore(QSqlQuery &query, int round, int table, int booger,
                   const BoogerResult &boogerResult);
    void updateAssignment(QSqlQuery &query,
                          int round, int pairOrPlayer, int pairNumber, int playerNumber, int table);
    bool updateValue(QSqlQuery &query, const QString &key, bool state);
    bool lockDatabase(const QString &dbFileName);
    void updateDbFileInfo();
    const QVariantList stringListToVariantList(const QStringList &stringList) const;
    void addMarker(QSqlQuery &query, const QString &name, const QColor &color,
                   Marker::Sorting sorting, int id = -1);
    void setMarker(QSqlQuery &query, const QString &name, int markerId);
    void registerPlayers(QSqlQuery &query, const QStringList &names);
    void listRemoveAllMarkers(QSqlQuery &query);
    bool ranksDiffer(const RankData &currentRank, const RankData &lastRank);

private: // Variables
    StringTools *m_stringTools;
    QSqlDatabase m_db;
    bool m_tournamentStarted;
    QString m_errorMessage;
    QString m_dbValueMuckturnierVersion;
    TournamentMode m_settingsTournamentMode;
    int m_settingsBoogerScore;
    int m_settingsBoogersPerRound;
    bool m_isWritable;
    bool m_noTableNumbers;
    bool m_autoSelectPairs;
    bool m_includeOpponentGoals;
    QLockFile *m_lockFile;
    QString m_dbFileName;
    QString m_dbFilePath;
    bool m_dbCheckRunning;
    QString m_namesSeparator;
    bool m_networkTournamentStarted = false;

};

#endif // DATABASE_H
