/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "FileDialog.h"
#include "Database.h"

// Qt includes
#include <QDebug>
#include <QStandardPaths>
#include <QFileInfo>
#include <QFileDialog>
#include <QMessageBox>
#include <QDesktopServices>

FileDialog::FileDialog(QWidget *parent, Database *db, DeleteMode deleteMode)
    : TitleDialog(parent, deleteMode), m_db(db)
{
    addButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
}

Database * FileDialog::db() const
{
    return m_db;
}

void FileDialog::setTargetFileBaseName(const QString &targetFileBaseName)
{
    m_targetFileBaseName = targetFileBaseName;
}

QString FileDialog::targetFileBaseName() const
{
    return m_targetFileBaseName;
}

void FileDialog::setTargetFileExtension(const QString &targetFileExtension)
{
    m_targetFileExtension = targetFileExtension;
}

void FileDialog::setTargetFileDescription(const QString &targetFileDescription)
{
    m_targetFileDescription = targetFileDescription;
}

void FileDialog::setFileDialogTitle(const QString &fileDialogTitle)
{
    m_fileDialogTitle = fileDialogTitle;
}

QString FileDialog::getTargetFileName(const QString &directory)
{
    QString defaultFileName = directory + QStringLiteral("/%1.%2").arg(m_targetFileBaseName,
                                                                       m_targetFileExtension);
    int number = 0;
    while (QFileInfo::exists(defaultFileName)) {
        number++;
        defaultFileName = directory + QStringLiteral("/%1_%2.%3").arg(m_targetFileBaseName,
                                                                      QString::number(number),
                                                                      m_targetFileExtension);
    }

    QString fileName = QFileDialog::getSaveFileName(this,
        m_fileDialogTitle,
        QDir::toNativeSeparators(defaultFileName),
        QStringLiteral("%1 (*.%2);;").arg(m_targetFileDescription, m_targetFileExtension)
        + tr("Alle Dateien") + QStringLiteral(" (*)"));

    if (fileName.isEmpty()) {
        return QString();
    }

    // The user could have selected the currently open database

    if (fileName == db()->dbFile()) {
        QMessageBox messageBox(this);
        messageBox.setIcon(QMessageBox::Warning);
        messageBox.setWindowTitle(windowTitle());
        messageBox.setText(tr("Die ausgewählte Datei ist die momentan geöffnete Turnierdatenbank. "
                              "Soll sie wirklich überschrieben werden? Alle Daten gehen dabei "
                              "verloren!"));
        messageBox.setStandardButtons(QMessageBox::Yes | QMessageBox::Cancel);
        messageBox.setDefaultButton(QMessageBox::Cancel);

        if (messageBox.exec() == QMessageBox::Cancel) {
            return QString();
        }

        db()->closeDatabaseConnection();
    }

    // Check if we can write the file. If it exists, it will be deleted, but the user has been
    // asked by the QFileDialog if this is okay. And yes, we have to do it in this way, because
    // QFileInfo::isWritable does not work as expected on Windows (returns false for
    // "My Documents").
    QFile test(fileName);
    if (test.open(QIODevice::WriteOnly)) {
        test.close();
        test.remove();
    } else {
        QMessageBox::critical(this,
            windowTitle(),
            tr("Es bestehen keine Schreibrechte für die Datei „%1“. "
               "Bitte die Zugriffsrechte überprüfen!").arg(QDir::toNativeSeparators(fileName)),
            QMessageBox::Ok);
        return QString();
    }

    return fileName;
}

void FileDialog::successMessage(const QString &title, const QString &message)
{
    QMessageBox::information(this, title, message, QMessageBox::Ok);
}

void FileDialog::openFile(const QString &fileName)
{
    QDesktopServices::openUrl(QUrl::fromLocalFile(fileName));
}
