/* Copyright (c) 2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "Settings.h"
#include "defaultValues.h"

// Qt includes
#include <QDebug>
#include <QAction>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonArray>
#include <QStandardPaths>
#include <QDir>

namespace
{

// Options check state group
const QLatin1String OPTIONS_CHECK_STATE("options_check_state/");
const QString SAVE_MAINWINDOW_GEOMETRY = OPTIONS_CHECK_STATE
                                         + QLatin1String("save_mainwindow_geometry");
const QString SAVE_PLAYERS_PAGE_COLUMNS = OPTIONS_CHECK_STATE
                                          + QLatin1String("save_players_page_columns");
const QString SAVE_LAST_DB_PATH = OPTIONS_CHECK_STATE + QLatin1String("save_last_db_path");

// Session data
const QString SESSION_DATA = QStringLiteral("session_data/");
const QString LAST_DB_PATH = SESSION_DATA + QLatin1String("last_db_path");
const QString MAINWINDOW_GEOMETRY = SESSION_DATA + QLatin1String("mainwindow_geometry");
const QString PLAYERS_PAGE_COLUMNS = SESSION_DATA + QLatin1String("players_page_columns");

// Strings
const QLatin1String STRINGS("strings/");
const QString STRINGS_BOOGER_SYMBOL = STRINGS + QLatin1String("booger_symbol");
const QString STRINGS_NAMES_SEPARATOR = STRINGS + QLatin1String("names_separator");

// Tournament template

constexpr int VERSION_TEMPLATE_TOURNAMENT = 1;

const QString TOURNAMENT_TEMPLATE = QStringLiteral("tournament_template");

const QString TOURNAMENT_TEMPLATE_VERSION
    = TOURNAMENT_TEMPLATE + QLatin1String("/version");

const QString TOURNAMENT_TEMPLATE_DB_FILE_NAME_TEMPLATE
    = TOURNAMENT_TEMPLATE + QLatin1String("/db_file_name_template");

const QString TOURNAMENT_TEMPLATE_PAIR_MODE
    = TOURNAMENT_TEMPLATE + QLatin1String("/pair_mode");
const QString TOURNAMENT_TEMPLATE_BOOGER_SCORE
    = TOURNAMENT_TEMPLATE + QLatin1String("/booger_score");
const QString TOURNAMENT_TEMPLATE_BOOGERS_PER_ROUND
    = TOURNAMENT_TEMPLATE + QLatin1String("/boogers_per_round");
const QString TOURNAMENT_TEMPLATE_AUTO_SELECT_PAIRS
    = TOURNAMENT_TEMPLATE + QLatin1String("/auto_select_pairs");
const QString TOURNAMENT_TEMPLATE_NO_TABLE_NUMBERS
    = TOURNAMENT_TEMPLATE + QLatin1String("/no_table_numbers");

// Markers template

constexpr int VERSION_TEMPLATE_MARKERS = 1;

const QString MARKERS_TEMPLATE = QStringLiteral("markers_template");
const QString MARKERS_TEMPLATE_VERSION = MARKERS_TEMPLATE + QLatin1String("/version");
const QString MARKERS_TEMPLATE_DATA    = MARKERS_TEMPLATE + QLatin1String("/data");

const QLatin1String MARKERS_TEMPLATE_JSON_DATA   ("d");
const QLatin1String MARKERS_TEMPLATE_JSON_NAME   ("n");
const QLatin1String MARKERS_TEMPLATE_JSON_COLOR  ("c");
const QLatin1String MARKERS_TEMPLATE_JSON_SORTING("s");

}

Settings::Settings(QObject *parent) : QSettings(QStringLiteral("muckturnier.org"),
                                                QStringLiteral("muckturnier"), parent)
{
}

// Main window geometry

bool Settings::saveMainWindowGeometryEnabled() const
{
    const QVariant data = value(SAVE_MAINWINDOW_GEOMETRY);
    return data.isValid() ? data.toBool() : true;
}

void Settings::enableSaveMainWindowGeometry(bool state)
{
    setValue(SAVE_MAINWINDOW_GEOMETRY, state);
    if (! state) {
        remove(MAINWINDOW_GEOMETRY);
    }
}

void Settings::saveMainWindowGeometry(const QByteArray &data)
{
    setValue(MAINWINDOW_GEOMETRY, data);
}

QByteArray Settings::mainWindowGeometry() const
{
    return value(MAINWINDOW_GEOMETRY, QByteArray()).toByteArray();
}

// Players page columns visiblity

bool Settings::savePlayersPageColumnsEnabled() const
{
    const QVariant data = value(SAVE_PLAYERS_PAGE_COLUMNS);
    return data.isValid() ? data.toBool() : true;
}

void Settings::enableSavePlayersPageColumns(bool state)
{
    setValue(SAVE_PLAYERS_PAGE_COLUMNS, state);
    if (! state) {
        remove(PLAYERS_PAGE_COLUMNS);
    }
}

void Settings::savePlayersPageColumns(const QList<bool> &visibility)
{
    QVariantList values;
    for (bool visible : visibility) {
        values.append(visible);
    }
    setValue(PLAYERS_PAGE_COLUMNS, values);
}

QList<bool> Settings::playersPageColumns() const
{
    QVariantList values = value(PLAYERS_PAGE_COLUMNS).toList();
    return values.count() != 2 ? QList<bool> { false, false }
                               : QList<bool> { values.at(0).toBool(), values.at(1).toBool() };
}

void Settings::saveTournamentTemplate(const TournamentTemplate &tournamentTemplate)
{
    setValue(TOURNAMENT_TEMPLATE_VERSION, VERSION_TEMPLATE_TOURNAMENT);
    setValue(TOURNAMENT_TEMPLATE_PAIR_MODE, tournamentTemplate.pairMode);
    setValue(TOURNAMENT_TEMPLATE_BOOGER_SCORE, tournamentTemplate.boogerScore);
    setValue(TOURNAMENT_TEMPLATE_BOOGERS_PER_ROUND, tournamentTemplate.boogersPerRound);
    setValue(TOURNAMENT_TEMPLATE_AUTO_SELECT_PAIRS, tournamentTemplate.autoSelectPairs);
    setValue(TOURNAMENT_TEMPLATE_NO_TABLE_NUMBERS, tournamentTemplate.noTableNumbers);
}

// Last database path

bool Settings::saveLastDbPathEnabled() const
{
    const QVariant data = value(SAVE_LAST_DB_PATH);
    return data.isValid() ? data.toBool() : true;
}

void Settings::enableSaveLastDbPath(bool state)
{
    setValue(SAVE_LAST_DB_PATH, state);
    if (! state) {
        remove(LAST_DB_PATH);
    }
}

void Settings::saveLastDbPath(const QString &dbPath)
{
    setValue(LAST_DB_PATH, dbPath);
}

QString Settings::lastDbPath() const
{
    const QString path = value(LAST_DB_PATH, QStandardPaths::writableLocation(
                                                 QStandardPaths::DocumentsLocation)).toString();
    if (QDir(path).exists()) {
        return path;
    } else {
        return QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation);
    }
}

// Tournament template

Settings::TournamentTemplate Settings::defaultTournamentTemplate() const
{
    return TournamentTemplate { true,
                                DEFAULT_BOOGER_SCORE,
                                DEFAULT_BOOGERS_PER_ROUND,
                                true,
                                false };
}

Settings::TournamentTemplate Settings::tournamentTemplate() const
{
    QVariant version = value(TOURNAMENT_TEMPLATE_VERSION);
    if (version.isValid() && version.toInt() == VERSION_TEMPLATE_TOURNAMENT) {
        return TournamentTemplate {
            value(TOURNAMENT_TEMPLATE_PAIR_MODE,         true).toBool(),
            value(TOURNAMENT_TEMPLATE_BOOGER_SCORE,      DEFAULT_BOOGER_SCORE).toInt(),
            value(TOURNAMENT_TEMPLATE_BOOGERS_PER_ROUND, DEFAULT_BOOGERS_PER_ROUND).toInt(),
            value(TOURNAMENT_TEMPLATE_AUTO_SELECT_PAIRS, true).toBool(),
            value(TOURNAMENT_TEMPLATE_NO_TABLE_NUMBERS,  false).toBool() };
    } else {
        return defaultTournamentTemplate();
    }
}

void Settings::saveDbFileNameTemplate(const QString &text)
{
    setValue(TOURNAMENT_TEMPLATE_DB_FILE_NAME_TEMPLATE, text);
}

// Database file name template

QString Settings::dbFileNameTemplate() const
{
    return value(TOURNAMENT_TEMPLATE_DB_FILE_NAME_TEMPLATE,
                 QStringLiteral(DEFAULT_DB_FILE_NAME_TEMPLATE)).toString();
}

void Settings::saveMarkersTemplate(const QList<Marker::Marker> &markers)
{
    QJsonArray markersJson;
    for (const Marker::Marker marker : markers) {
        markersJson.append(QJsonObject {
            { MARKERS_TEMPLATE_JSON_NAME, marker.name },
            { MARKERS_TEMPLATE_JSON_COLOR, QString::number(marker.color.rgb()) },
            { MARKERS_TEMPLATE_JSON_SORTING, static_cast<int>(marker.sorting) } });
    }

    setValue(MARKERS_TEMPLATE_VERSION, VERSION_TEMPLATE_MARKERS);
    setValue(MARKERS_TEMPLATE_DATA,
             QJsonDocument(QJsonObject { { MARKERS_TEMPLATE_JSON_DATA, markersJson } })
                 .toJson(QJsonDocument::Compact));
}

// Markers template

bool Settings::parseMarkerData(int &id, const QJsonValue &data, QList<Marker::Marker> &markers)
{
    Marker::Marker marker;
    const QJsonObject markerData = data.toObject();
    QJsonValue value;

    value = markerData.value(MARKERS_TEMPLATE_JSON_NAME);
    if (value.type() != QJsonValue::String) {
        return false;
    } else {
        marker.name = value.toString();
    }

    value = markerData.value(MARKERS_TEMPLATE_JSON_COLOR);
    if (value.type() != QJsonValue::String) {
        return false;
    } else {
        marker.color = QColor::fromRgb(value.toString().toUInt());
    }

    value = markerData.value(MARKERS_TEMPLATE_JSON_SORTING);
    if (value.type() != QJsonValue::Double) {
        return false;
    } else {
        marker.sorting = static_cast<Marker::Sorting>(value.toInt());
    }

    if (marker.name.isEmpty()) {
        marker.id = 0;
    } else {
        marker.id = ++id;
    }
    markers.append(marker);

    return true;
}

void Settings::appendMarker(const QString &name, const QColor &color, Marker::Sorting sorting,
                            int id, QList<Marker::Marker> &markers) const
{
    Marker::Marker marker;
    marker.name = name;
    marker.color = color;
    marker.sorting = sorting;
    marker.id = id;
    markers.append(marker);
}

QList<Marker::Marker> Settings::defaultMarkersTemplate()
{
    QList<Marker::Marker> markers;
    appendMarker(tr("Rot"),  QColor(255, 0,   0,   255), Marker::Sorting::NewBlock, 1, markers);
    appendMarker(tr("Blau"), QColor(0,   0,   255, 255), Marker::Sorting::NewBlock, 2, markers);
    appendMarker(QString(),  QColor(0,   0,   0,   255), Marker::Sorting::NewBlock, 0, markers);
    appendMarker(tr("Grün"), QColor(0,   128, 0,   255), Marker::Sorting::Merged,   3, markers);
    return markers;
}

QList<Marker::Marker> Settings::markersTemplate()
{
    QList<Marker::Marker> markers;
    bool templateParsed = false;

    if (value(MARKERS_TEMPLATE_VERSION, 0).toInt() == VERSION_TEMPLATE_MARKERS) {
        // We have the correct version in the settings, so we look what we have :-)

        QJsonParseError error;
        const QJsonDocument json = QJsonDocument::fromJson(
            value(MARKERS_TEMPLATE_DATA, QByteArray()).toByteArray(), &error);

        if (error.error == QJsonParseError::NoError && json.isObject()) {
            // Try to parse the data we have to a markers list
            const QJsonValue data = json.object().value(MARKERS_TEMPLATE_JSON_DATA);
            if (data.isArray()) {
                templateParsed = true;
                int id = 0;
                for (const QJsonValue &value : data.toArray()) {
                    if (! parseMarkerData(id, value, markers)) {
                        templateParsed = false;
                        break;
                    }
                }
            }
        }

        // Be sure to have an "(unmarked)" marker with ID 0 and no name
        if (templateParsed) {
            templateParsed = false;
            for (const Marker::Marker &marker : markers) {
                if (marker.id == 0 && marker.name.isEmpty()) {
                    templateParsed = true;
                    break;
                }
            }
        }

    }

    // Return the default markers if we don't have a valid template
    if (! templateParsed) {
        return defaultMarkersTemplate();
    }

    return markers;
}

// Strings

void Settings::saveBoogerSymbol(const QString &text)
{
    setValue(STRINGS_BOOGER_SYMBOL, text);
}

QString Settings::boogerSymbol() const
{
    QString boogerSymbol = value(STRINGS_BOOGER_SYMBOL,
                                 QStringLiteral(DEFAULT_BOOGER_SYMBOL)).toString();
    if (boogerSymbol.left(1).simplified().isEmpty()) {
        return QStringLiteral(DEFAULT_BOOGER_SYMBOL);
    } else {
        return boogerSymbol;
    }
}

void Settings::saveNamesSeparator(const QString &text)
{
    setValue(STRINGS_NAMES_SEPARATOR, text);
}

QString Settings::namesSeparator() const
{
    QString namesSeparator = value(STRINGS_NAMES_SEPARATOR,
                                   QStringLiteral(DEFAULT_NAMES_SEPARATOR)).toString();
    return namesSeparator.remove(QLatin1Char('\t'));
}
