/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes

#include "SocketAddressWidget.h"

#include "shared/defaultValues.h"

// Qt includes
#include <QGridLayout>
#include <QNetworkInterface>
#include <QLabel>
#include <QLineEdit>
#include <QIntValidator>
#include <QToolTip>
#include <QTimer>

SocketAddressWidget::SocketAddressWidget(Mode mode, QWidget *parent) : QWidget(parent)
{
    setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);

    QGridLayout *layout = new QGridLayout(this);

    layout->addWidget(new QLabel(tr("IP-Adresse:")), 0, 0);

    m_ip = new QLineEdit;
    connect(m_ip, &QLineEdit::textEdited, this, &SocketAddressWidget::checkSettings);
    connect(m_ip, &QLineEdit::returnPressed, this, &SocketAddressWidget::returnPressed);
    layout->addWidget(m_ip, 0, 1);

    const QList<QHostAddress> allAddresses(QNetworkInterface::allAddresses());

    // Check for an IPv4 address
    for (const QHostAddress &address : allAddresses) {
        if (address != QHostAddress::LocalHost
            && address.protocol() == QAbstractSocket::IPv4Protocol) {

            m_hostAddress = address;
            break;
        }
    }

    // Check for an IPv6 address
    if (m_hostAddress.isNull()) {
        for (const QHostAddress &address : allAddresses) {
            if (address != QHostAddress::LocalHost
                && address.protocol() == QAbstractSocket::IPv6Protocol) {

                m_hostAddress = address;
                break;
            }
        }
    }

    // Check for the IPv4 localhost
    if (m_hostAddress.isNull()) {
        for (const QHostAddress &address : allAddresses) {
            if (address == QHostAddress::LocalHost
                && address.protocol() == QAbstractSocket::IPv4Protocol) {

                m_hostAddress = address;
                break;
            }
        }
    }

    // Check for an IPv6 localhost
    if (m_hostAddress.isNull()) {
        for (const QHostAddress &address : allAddresses) {
            if (address == QHostAddress::LocalHost
                && address.protocol() == QAbstractSocket::IPv6Protocol) {

                m_hostAddress = address;
                break;
            }
        }
    }

    switch (mode) {
    case Server:
        m_ip->setText(m_hostAddress.toString());
        break;
    case Client:
        if (m_hostAddress.toIPv4Address()) {
            m_ip->setText(m_hostAddress.toString().mid(
                        0, m_hostAddress.toString().lastIndexOf(QLatin1String(".")) + 1));
        }
        break;
    }

    layout->addWidget(new QLabel(tr("Port:")), 1, 0);
    m_port = new QLineEdit(QString::number(NETWORK_DEFAULT_PORT));
    m_port->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    m_port->setValidator(new QIntValidator(1024, 65535, this));
    connect(m_port, &QLineEdit::textEdited, this, &SocketAddressWidget::checkSettings);
    connect(m_port, &QLineEdit::returnPressed, this, &SocketAddressWidget::returnPressed);
    layout->addWidget(m_port, 1, 1);

    if (mode == Server) {
        // We have to do this via a QTimer call, otherwise it has no effect
        QTimer::singleShot(0, this, [this] { m_ip->deselect(); });
    }
}

void SocketAddressWidget::checkSettings()
{
    emit settingsOkay(m_hostAddress.setAddress(m_ip->text()) && m_port->hasAcceptableInput());

    if (m_ip->text().isEmpty() || m_hostAddress.isNull()) {
        QToolTip::hideText();
    } else {
        if (m_hostAddress.toString() != m_ip->text()) {
            QToolTip::showText(mapToGlobal(m_ip->geometry().bottomLeft()),
                                m_hostAddress.toString());
        } else {
            QToolTip::hideText();
        }
    }
}

QString SocketAddressWidget::ip() const
{
    return m_hostAddress.toString();
}

int SocketAddressWidget::port() const
{
   return m_port->text().toInt();
}

void SocketAddressWidget::setSocketAddress(const QString &ip, int port)
{
    m_ip->setText(ip);
    m_port->setText(QString::number(port));
    checkSettings();
}

void SocketAddressWidget::focusIp()
{
    QTimer::singleShot(0, this, [this] { m_ip->setFocus(); });
}
