/* Copyright (c) 2018-2019 Tobias Leupold <tobias.leupold@gmx.de>

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation in version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// Local includes
#include "StringTools.h"
#include "Characters.h"

// Qt includes
#include <QDebug>

StringTools::StringTools(QObject *parent) : QObject(parent)
{
    m_collator.setCaseSensitivity(Qt::CaseInsensitive);
}

void StringTools::sortList(QStringList &list) const
{
    // According to the Qt Wiki, one should use the following lambda function:
    // [&](const QString &a, const QString &b) { return m_sorter.compare(a, b) < 0; }
    // but referencing the QCollator directly works just fine!
    std::sort(list.begin(), list.end(), m_collator);
}

QString StringTools::toLower(const QString &text) const
{
    return m_locale.toLower(text);
}

void StringTools::appendIfNotExists(QStringList &list, const QString &text) const
{
    if (! list.contains(text)) {
        list.append(text);
    }
}

void StringTools::appendIfNotExists(QList<int> &list, const int value) const
{
    if (! list.contains(value)) {
        list.append(value);
    }
}

void StringTools::removeIntersection(QStringList &list1, QStringList &list2) const
{
    const QStringList workList(list1);
    for (const QString &entry : workList) {
        if (list2.removeOne(entry)) {
            list1.removeOne(entry);
        }
    }
}

int StringTools::levenshteinDistance(const QString &source, const QString &target) const
{
    // Mostly stolen from https://qgis.org/api/2.14/qgsstringutils_8cpp_source.html

    if (source == target) {
        return 0;
    }

    const int sourceCount = source.count();
    const int targetCount = target.count();

    if (source.isEmpty()) {
        return targetCount;
    }

    if (target.isEmpty()) {
        return sourceCount;
    }

    if (sourceCount > targetCount) {
        return levenshteinDistance(target, source);
    }

    QVector<int> column;
    column.fill(0, targetCount + 1);
    QVector<int> previousColumn;
    previousColumn.reserve(targetCount + 1);
    for (int i = 0; i < targetCount + 1; i++) {
        previousColumn.append(i);
    }

    for (int i = 0; i < sourceCount; i++) {
        column[0] = i + 1;
        for (int j = 0; j < targetCount; j++) {
            column[j + 1] = std::min( {
                                1 + column.at(j),
                                1 + previousColumn.at(1 + j),
                                previousColumn.at(j) + ((source.at(i) == target.at(j)) ? 0 : 1) } );
        }
        column.swap(previousColumn);
    }

    return previousColumn.at(targetCount);
}

QMap<int, int> StringTools::combinedLevenshteinDistance(const QString &text,
                                                        const QStringList &candidates) const
{
    QMap<int, int> combinedDistances;
    const QStringList words = text.split(QStringLiteral(" "));

    // Calculate a combinded distance for each candidate

    for (int i = 0; i < candidates.count(); i++) {
        QStringList textWords = words;
        QStringList candidateWords = candidates.at(i).split(QStringLiteral(" "));
        removeIntersection(textWords, candidateWords);

        const int textWordsCount = textWords.count();
        const int candidateWordsCount = candidateWords.count();
        int distance = 0;
        int biggestDistance = 0;
        QMap<int, QMap<int, int>> distances;

        // Calculate a distance matrix for each word pair
        for (int i = 0; i < textWordsCount; i++) {
            for (int j = 0; j < candidateWordsCount; j++) {
                const int currentDistance = levenshteinDistance(textWords.at(i),
                                                                candidateWords.at(j));
                distances[i][j] = currentDistance;
                if (biggestDistance < currentDistance) {
                    biggestDistance = currentDistance;
                }
            }
        }

        // Search for the smallest distance of each word and remove it from the candidate list

        while (! distances.isEmpty()) {
            int smallestDistance = biggestDistance + 1;
            int smallestDistanceWordIndex;
            int smallestDistanceCandidateIndex;
            for (const int textWordIndex : distances.keys()) {
                for (const int candidateWordIndex : distances[textWordIndex].keys()) {
                    if (distances[textWordIndex][candidateWordIndex] < smallestDistance) {
                        smallestDistance = distances[textWordIndex][candidateWordIndex];
                        smallestDistanceWordIndex = textWordIndex;
                        smallestDistanceCandidateIndex = candidateWordIndex;
                    }
                }
            }

            distance += smallestDistance;
            distances.remove(smallestDistanceWordIndex);
            for (const int textWordIndex : distances.keys()) {
                distances[textWordIndex].remove(smallestDistanceCandidateIndex);
            }
        }

        // This is the smallest possible distance of the respective name and the current candidate
        combinedDistances[i] = distance;
    }

    return combinedDistances;
}

QMap<int, int> StringTools::matchLists(QStringList &source, QStringList &target) const
{
    QMap<int, QMap<int, int>> allDistances;
    QList<int> remainingTargetIndices;
    QMap<int, int> assignment;

    // Calculate all distances for all possible pairings
    for (int i = 0; i < source.count(); i++) {
        allDistances[i] = combinedLevenshteinDistance(source.at(i), target);
        remainingTargetIndices.append(i);
    }

    // Calculate the best pairing
    while (! allDistances.isEmpty()) {
        int distance = -1;
        int bestSourceIndex;
        for (int sourceIndex : allDistances.keys()) {
            for (int targetIndex : allDistances[sourceIndex].keys()) {
                if (distance == -1 || allDistances[sourceIndex][targetIndex] < distance) {
                    distance = allDistances[sourceIndex][targetIndex];
                    bestSourceIndex = sourceIndex;
                }
            }
        }

        while (true) {
            int matchDistance = -1;
            int bestTargetIndex;
            for (int targetIndex : allDistances[bestSourceIndex].keys()) {
                if (matchDistance == -1 || allDistances[bestSourceIndex][targetIndex] < matchDistance) {
                    matchDistance = allDistances[bestSourceIndex][targetIndex];
                    bestTargetIndex = targetIndex;
                }
            }

            if (remainingTargetIndices.contains(bestTargetIndex)) {
                assignment[bestSourceIndex] = bestTargetIndex;
                remainingTargetIndices.removeOne(bestTargetIndex);
                break;
            } else {
                allDistances[bestSourceIndex].remove(bestTargetIndex);
            }
        }

        allDistances.remove(bestSourceIndex);
    }

    return assignment;
}

QString StringTools::longestCommonSubsequence(const QString &source, const QString &target) const
{
    // Mostly stolen from https://www.geeksforgeeks.org/printing-longest-common-subsequence/

    QMap<int, QMap<int, int>> l;
    for (int i = 0; i <= source.count(); i++) {
        for (int j = 0; j <= target.count(); j++) {
            if (i == 0 || j == 0) {
                l[i][j] = 0;
            } else if (source.at(i - 1) == target.at(j - 1)) {
                l[i][j] = l[i - 1][j - 1] + 1;
            } else {
                l[i][j] = std::max(l[i - 1][j], l[i][j - 1]);
            }
        }
    }

    int i = source.count();
    int j = target.count();
    int index = l[source.count()][target.count()];
    QString longestCommonSubsequence;
    while (i > 0 && j > 0) {
        if (source.at(i - 1) == target.at(j - 1)) {
            longestCommonSubsequence[index - 1] = source.at(i - 1);
            i--;
            j--;
            index--;
        } else if (l[i - 1][j] > l[i][j - 1]) {
            i--;
        } else {
            j--;
        }
    }

    return longestCommonSubsequence;
}

QStringList StringTools::diffList(QString text, QString longestCommonSubsequence) const
{
    if (text.isEmpty()) {
        return QStringList { text };
    }

    if (longestCommonSubsequence.isEmpty()) {
        return QStringList { QString(), text };
    }

    QStringList allParts;
    QString currentPart;
    bool matching = true;
    if (text.at(0) != longestCommonSubsequence.at(0)) {
        allParts.append(QString());
        matching = false;
    }

    while (! text.isEmpty()) {
        if (matching) {
            while (! text.isEmpty() && ! longestCommonSubsequence.isEmpty()
                   && text.at(0) == longestCommonSubsequence.at(0)) {

                currentPart.append(text.at(0));
                text.remove(0, 1);
                longestCommonSubsequence.remove(0, 1);
            }
            allParts.append(currentPart);
            currentPart.clear();
            matching = false;
        } else {
            while (! text.isEmpty() && text.at(0) != longestCommonSubsequence.at(0)) {
                currentPart.append(text.at(0));
                text.remove(0, 1);
            }
            allParts.append(currentPart);
            currentPart.clear();
            matching = true;
        }
    }

    return allParts;
}

QString StringTools::boogify(int score, int boogerScore) const
{
    return score == boogerScore ? QString::number(score)
                                : QStringLiteral("%1 %2").arg(QString::number(score),
                                                              m_boogerSymbol);
}

QString StringTools::locale() const
{
    return m_locale.bcp47Name();
}

QString StringTools::defaultLocale() const
{
    return QLocale().bcp47Name();
}

void StringTools::setLocale(const QString &name)
{
    m_locale = QLocale(name);
    m_collator.setLocale(m_locale);
    m_localeChanged = true;
}

void StringTools::resetLocale()
{
    if (! m_localeChanged) {
        return;
    }

    m_locale = QLocale();
    m_collator.setLocale(m_locale);
    m_localeChanged = false;
}

bool StringTools::localeChanged() const
{
    return m_localeChanged;
}

void StringTools::setBoogerSymbol(const QString &text)
{
    m_boogerSymbol = text;
}

const QString & StringTools::boogerSymbol() const
{
    return m_boogerSymbol;
}
